//================================================================
// nB (nano Base) - Copyright (c) 1996   Daniele Giacomini
//
// Permission to use, copy, and distribute this software and its
// documentation for any purpose with or without fee is hereby
// granted, provided that the above copyright notice appear in all
// copies and that both that copyright notice and this permission
// notice appear in supporting documentation.
//
// Permission to modify the software is granted, but not the right
// to distribute the modified code.
// Modifications are to be distributed as patches to released
// version.
//
// THIS SOFTWARE MAY BE EXTREMLY DANGEROUS! FOR THIS SOFTWARE
// THERE IS ABSOLUTELY NO WARRANTY! THIS SOFTWARE IS PROVIDED
// "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS OR
// IMPLIED! IN NO EVENT WILL THE AUTHOR BE LIABLE TO YOU FOR
// ANY DAMAGES, INCLUDING LOST PROFITS, LOST SAVINGS, OR OTHER
// INCIDENTAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF THE USE
// OR INABILITY TO USE THIS SOFTWARE. IF THE LAW OF A COUNTRY
// OR A REGION DO NOT PERMIT TO SAVE THE AUTHOR FROM ALL POSSIBLE
// RESPONSABILITIES ON THIS SOFTWARE, IT MUST BE ASSUMED THAT
// THERE IS NO PERMISSION TO USE THIS SOFTWARE THERE.
//
//================================================================
// Send your comments, suggestions, bug reports and english text
// corrections to:
//
// Daniele Giacomini
// Via Turati, 15
// I-31100 Treviso
// Italy
//
// e-mail:
// daniele@system.abacom.it
//
//================================================================
// STANDARD INCLUDE
//================================================================
//#include "std.ch"
#include "common.ch"
#include "box.ch"
#include "color.ch"
#include "directry.ch"
#include "Fileio.ch"
#include "inkey.ch"
#include "memoedit.ch"
#include "rddsys.ch"
#include "setcurs.ch"
#include "set.ch"
#include "error.ch"
#include "dbstruct.ch"

//================================================================
// REQUEST
//================================================================
// modify here if less or more RDDs (Replaceable Database Driver)
// are needed.
request dbfcdx
request dbfmdx
request dbfndx
request dbfntx // default

//================================================================
// ASCII characters
//================================================================
#define NUL chr(00)  // NULl
#define SOH chr(01)  // Start Of Hiteration
#define STX chr(02)  // Start TeXt
#define ETX chr(03)  // End TeXt
#define EOT chr(04)  // End Of Transmission
#define ENQ chr(05)  // ENQuery
#define ACK chr(06)  // ACKnowledge
#define BEL chr(07)  // BELl
#define BS  chr(08)  // Back Space
#define HT  chr(09)  // Horizontal Tab
#define LF  chr(10)  // Line Feed
#define VT  chr(11)  // Vertical Tab
#define FF  chr(12)  // Form Feed
#define CR  chr(13)  // Carriage Return
#define SO  chr(14)  // Shift Out
#define SI  chr(15)  // Shif In
#define DLE chr(16)  // Data Loop Exit
#define DC1 chr(17)  // Dispositive Check 1
#define DC2 chr(18)  // Dispositive Check 2
#define DC3 chr(19)  // Dispositive Check 3
#define DC4 chr(20)  // Dispositive Check 4
#define NAK chr(21)  // Not AKnowledge
#define SYN chr(22)  // SYNcronized
#define ETB chr(23)  // End Trasmission Block
#define CAN chr(24)  // CANcel
#define EM  chr(25)  // End Medium
#define SUB chr(26)  // SUBstitute
#define ESC chr(27)  // ESCape
#define FS  chr(28)  // File Separator
#define GS  chr(29)  // Group Separator
#define RS  chr(30)  // Record Separator
#define US  chr(31)  // Unit Separator

//================================================================
// ASCII STRINGS
//================================================================
#define NL( nTimes )    ( replicate ( CR + LF, nTimes ) )
#define ENDTEXT;
    NL(2) + "----------END OF TEXT----------"
#define SPACE_LEN       254 // max normal string, see memoline()
#define SPACE_STR       space(SPACE_LEN)

//================================================================
// COLORS
//================================================================
#define COLOR_BW0       "W/N, N/W, N, N, N/W"
#define COLOR_BW1       "N/W, W+/N, N, N, W/N"
#define COLOR_BW2       "W/N, N/W, N, N, W+/N"
#define COLOR_C0        "GB+/N, W+/G, N, N, N/W"
#define COLOR_C1        "W+/GR, G+/N, N, N, W/N"
#define COLOR_C2        "N/W, G+/N, N, N, B/W"
#define COLOR_DEFAULT0  ( iif( iscolor(), COLOR_C0, COLOR_BW0 ) )
#define COLOR_DEFAULT1  ( iif( iscolor(), COLOR_C1, COLOR_BW1 ) )
#define COLOR_DEFAULT2  ( iif( iscolor(), COLOR_C2, COLOR_BW2 ) )

//================================================================
// PROGRAM INFO/COPYRIGHT
//================================================================
#define PROGRAM_COPYRIGHT;
    "nB (nano Base) 1996.02.28 (c) 1996 Daniele Giacomini." +;
    + NL(1) +;
    "nB is freeware and comes with ABSOLUTELY NO WARRANTY!" +;
    + NL(1) +;
    "e-mail: daniele@system.abacom.it"

//================================================================
// GENERAL USE
//================================================================
#define _ERROR_EXCLUSIVE_REQUIRED;
    "For this action, the Alias must be opened in exclusive mode."
#define _ERROR_FLOCK_FAILURE    "The file is already in use."
#define _ERROR_NO_ALIAS         "There is no active Alias."
#define _ERROR_NO_INDEX;
    "An opened index and a selected order is needed before."
#define _ERROR_PROMPT_ALREADY_EXIST_OVERWRITE;
    "This file already exists. Overwrite?"

#define _EXTENTION_DBF      "DBF"
#define _EXTENTION_FORM     "FRM"
#define _EXTENTION_INDEXBAG substr( ordbagext(), 2 )
#define _EXTENTION_LABEL    "LBL"
#define _EXTENTION_TEXT     "TXT"

#define _KEY_CTRLY;
    "[Ctrl]+[Y] delete/recall"
#define _KEY_ENTER;
    "[Enter] select"
#define _KEY_ESC_EXIT_ENTER_UP_DOWN;
    "[Esc] exit  [Enter] select  []/[] move"
#define _KEY_ESC_CANCEL_F1;
    "[Esc] cancel  [F1] help"
#define _KEY_ESC_CANCEL_F1_F2_PGDN;
    "[Esc] cancel  [F1] help  [F2] list  [Pag] confirm"
#define _KEY_ESC_CANCEL_PGDN;
    "[Esc] cancel  [Pag] confirm"
#define _KEY_ESC_END_F1;
    "[Esc] end  [F1] help"
#define _KEY_ESC_EXIT_F1;
    "[Esc] exit  [F1] help"
#define _KEY_ESC_EXIT_F1_PGDN;
    "[Esc] exit  [F1] help  [Pag] continue"
#define _KEY_ESC_EXIT_F1_F2_PGDN;
    "[Esc] exit  [F1] help  [F2] list  [Pag] confirm"
#define _KEY_ESC_EXIT_F1_F2_CTRLY;
    "[Esc] exit  [F1] help  [F2] list  [Ctrl]+[Y] delete"
#define _KEY_ESC_EXIT_F1_F10;
    "[Esc] exit  [F1] help  [F10] menu"
#define _KEY_ESC_NORMAL_F1_PGDN;
    "[Esc] normal  [F1] help  [Pag] continue"
#define _KEY_F2_PGDN;
    "[F2] list/dir  [Pag] confirm"
#define _KEY_MENU_H;
    "[Esc] exit  [-]/[-] move cursor  [Enter] open menu"
#define _KEY_MENU_V;
    "[Esc] exit  [-]/[-]/[]/[] cursor movement  [Enter] select"
#define _KEY_MENU_P;
    "[Esc] exit  []/[] cursor movement  [Enter] select"
#define _KEY_UP_DOWN;
    "[]/[] move cursor"
#define _KEY_UP_DOWN_ENTER;
    "[]/[] move  [Enter] select"

#define _MAX_ORDER  250
#define _MAX_SELECT 250

#define _MENU_YES           " YES "
#define _MENU_NO            " NO "
#define _MENU_TENDINA_WIDTH 26

#define _MENU_HELP;
    "MENU" +;
    NL(2) +;
    " One Two Three Four Five" 			+ NL(1) +;
    "                     " 			+ NL(1) +;
    "                    Menu title"		+ NL(1) +;
    NL(2) +;
    "The cursor may be moved on a different menu title using " +;
    "arrow keys [-]/[-];" +;
    NL(2) +;
    "[Esc] terminates the menu;" + NL(1) +;
    "[Enter] opens a vertical menu." +;
    NL(4) +;
    " One Two Three Four Five" 			+ NL(1) +;
    "Ŀ        " 			+ NL(1) +;
    "First             Menu title" 		+ NL(1) +;
    "Second Menu function" 		+ NL(1) +;
    "Third      - Menu" 			+ NL(1) +;
    "" 				+ NL(1) +;
    NL(2) +;
    "The arrow keys UP and DOWN []/[] move the cursor "+;
    "to the previous or to the next menu function, inside " +;
    "the opened menu." + NL(1) +;
    "The arrow keys LEFT and RIGHT [-]/[-] change the menu." +;
    NL(2) +;
    "[Esc] closes the menu; " + NL(1) +;
    "[Enter] starts the selected menu function." +;
    NL(4) +;
    " One Two Three Four Five" 			+ NL(1) +;
    "Ŀ        " 			+ NL(1) +;
    "First             Menu title" 		+ NL(1) +;
    "Second     " 				+ NL(1) +;
    "Third     Ŀ"		+ NL(1) +;
    "ĳSub function 1 Menu" 	+ NL(1) +;
    "            Sub function 2      sub function" + NL(1) +;
    "            " 		+ NL(1) +;
    NL(2) +;
    "The arrow keys UP and DOWN []/[] move the cursor "+;
    "to the previous or to the next menu sub function, inside " +;
    "the opened pop up menu." +;
    NL(2) +;
    "[Esc] closes the pop up menu;" + NL(1) +;
    "[Enter] starts the selected sub menu function."

#define _UNTITLED    "UNTITLED"

//================================================================
// OTHERS DIVIDED PER FUNCTION GROUP
//================================================================
#define ACCEPT_FUNCTION_MANE;
    "Accept()"
#define ACCEPT_DIALOG_BOX_TOP;
    ACCEPT_FUNCTION_MANE

//----------------------------------------------------------------
#define ASSIST_FUNCTION_NAME;
    "Assist()"
#define ASSIST_PROMPT_EXIT;
    "Exit " + ASSIST_FUNCTION_NAME + " ?"
#define ASSIST_PROMPT_SET_ALTERNATE;
    "Set alternate to (define output device):"

#define ASSIST_MENU_FILE_CHOICE     "File"
#define ASSIST_MENU_EDIT_CHOICE     "Edit"
#define ASSIST_MENU_REPORT_CHOICE   "Report"
#define ASSIST_MENU_INFO_CHOICE     "Info"

#define ASSIST_MENU_FILE_MSG;
    "Main menu"
#define ASSIST_MENU_EDIT_MSG;
    "Editing menu"
#define ASSIST_MENU_REPORT_MSG;
    "Report menu"
#define ASSIST_MENU_INFO_MSG;
    "Information menu"

#define ASSIST_MENU_FILE_NEWDIR_CHOICE;
    padr( "Change Directory",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_DBF_CHOICE;
    padr( "File ." + _EXTENTION_DBF,;
    _MENU_TENDINA_WIDTH-3 ) + ""
#define ASSIST_MENU_FILE_NTX_CHOICE;
    padr( "File ." + _EXTENTION_INDEXBAG,;
    _MENU_TENDINA_WIDTH-3 ) + ""
#define ASSIST_MENU_FILE_ALIAS_CHOICE;
    padr( "Alias",;
    _MENU_TENDINA_WIDTH-3 ) + ""
#define ASSIST_MENU_FILE_INDEX_CHOICE;
    padr( "Order",;
    _MENU_TENDINA_WIDTH-3 ) + ""
#define ASSIST_MENU_FILE_RELATION_CHOICE;
    padr( "Relation",;
    _MENU_TENDINA_WIDTH-3 ) + ""
#define ASSIST_MENU_FILE_RDD_DEFAULT_CHOICE;
    padr( "RDD Default",;
    _MENU_TENDINA_WIDTH-3 ) + ""

#define ASSIST_MENU_FILE_NEWDIR_MSG;
    "Change the actual directory"
#define ASSIST_MENU_FILE_DBF_MSG;
    "." + _EXTENTION_DBF + " file operations"
#define ASSIST_MENU_FILE_NTX_MSG;
    "." + _EXTENTION_INDEXBAG + " file operations"
#define ASSIST_MENU_FILE_ALIAS_MSG;
    "Alias operations"
#define ASSIST_MENU_FILE_INDEX_MSG;
    "Index/Order operations"
#define ASSIST_MENU_FILE_RELATION_MSG;
    "Relation operations"
#define ASSIST_MENU_FILE_RDD_DEFAULT_MSG;
   "Replaceable Database Driver"

#define ASSIST_MENU_FILE_DBF_NEW_CHOICE;
    padr( "New ." + _EXTENTION_DBF,;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_DBF_MODSTRU_CHOICE;
    padr( "Modify ." + _EXTENTION_DBF + " structure",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_DBF_OPEN_CHOICE;
    padr( "Open ." + _EXTENTION_DBF,;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_FILE_DBF_NEW_MSG;
    "Create a new ." + _EXTENTION_DBF + " file"
#define ASSIST_MENU_FILE_DBF_OPEN_MSG;
    "Open a ." + _EXTENTION_DBF + " file"
#define ASSIST_MENU_FILE_DBF_MODSTRU_MSG;
    "Modify the structure of a ." + _EXTENTION_DBF + " file"

#define ASSIST_MENU_FILE_NTX_NEW_CHOICE;
    padr( "New ." + _EXTENTION_INDEXBAG + " / New Tag",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_NTX_OPEN_CHOICE;
    padr( "Open Index",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_FILE_NTX_NEW_MSG;
    "Create a new index for the actual Alias or add a Tag"
#define ASSIST_MENU_FILE_NTX_OPEN_MSG;
    "Open and index for the actual Alias"

#define ASSIST_MENU_FILE_ALIAS_SELECT_CHOICE;
    padr( "Select",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_ALIAS_DISPSTRU_CHOICE;
    padr( "Display structure",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_ALIAS_CLOSE_CHOICE;
    padr( "Close active Alias",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_ALIAS_CLOSEALL_CHOICE;
    padr( "Close all Aliases",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_FILE_ALIAS_SELECT_MSG;
    "Select another Alias or a new area"
#define ASSIST_MENU_FILE_ALIAS_DISPSTRU_MSG;
    "Show the actual Alias scructure"
#define ASSIST_MENU_FILE_ALIAS_CLOSE_MSG;
    "Close the actual Alias"
#define ASSIST_MENU_FILE_ALIAS_CLOSEALL_MSG;
    "Close all areas"

#define ASSIST_MENU_FILE_INDEX_REINDEX_CHOICE;
    padr( "Order List Rebuild",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_INDEX_SET_ORDER_CHOICE;
    padr( "Order Set Focus",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_INDEX_CLOSENTX_CHOICE;
    padr( "Order List Clear",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_FILE_INDEX_SET_ORDER_MSG;
    "Select another index order for the actual Alias"
#define ASSIST_MENU_FILE_INDEX_REINDEX_MSG;
    "Index rebuild"
#define ASSIST_MENU_FILE_INDEX_CLOSENTX_MSG;
    "Close indexes for the actual Alias"

#define ASSIST_MENU_FILE_RELATION_SET_CHOICE;
    padr( "Set Relation",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_RELATION_CLOSE_CHOICE;
    padr( "Clear Relation",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_FILE_RELATION_SET_MSG;
   "Define a relation"
#define ASSIST_MENU_FILE_RELATION_CLOSE_MSG;
   "Relise all relations for the actual Alias"

#define ASSIST_MENU_FILE_RDD_SHOW_CHOICE;
    padr( "Show actual RDD default",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_FILE_RDD_SET_DEFAULT_CHOICE;
    padr( "Set default RDD",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_FILE_RDD_SHOW_MSG;
    "Show the default RDD"
#define ASSIST_MENU_FILE_RDD_SET_DEFAULT_MSG;
    "Define a new default RDD"

#define ASSIST_MENU_EDIT_VIEW_CHOICE;
    padr( "View",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_EDIT_BROWSE_CHOICE;
    padr( "Edit/Browse",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_EDIT_REPLACE_CHOICE;
    padr( "Replace",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_EDIT_RECALL_CHOICE;
    padr( "Recall",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_EDIT_DELETE_CHOICE;
    padr( "Delete",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_EDIT_PACK_CHOICE;
    padr( "Pack",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_EDIT_BROWSE_MSG;
    "Browse the actual Alias"
#define ASSIST_MENU_EDIT_VIEW_MSG;
    "View the actual Alias"
#define ASSIST_MENU_EDIT_REPLACE_MSG;
    "Data replacement"
#define ASSIST_MENU_EDIT_RECALL_MSG;
    "Record undelete"
#define ASSIST_MENU_EDIT_DELETE_MSG;
    "Record delete"
#define ASSIST_MENU_EDIT_PACK_MSG;
    "Deleted record elimination"

#define ASSIST_MENU_REPORT_LABEL_NEW_CHOICE;
    padr( "New Label",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_REPORT_LABEL_MODIFY_CHOICE;
    padr( "Modify Label",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_REPORT_LABEL_FORM_CHOICE;
    padr( "Label Form",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_REPORT_REPORT_NEW_CHOICE;
    padr( "New Report",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_REPORT_REPORT_MODIFY_CHOICE;
    padr( "Modify Report",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_REPORT_REPORT_FORM_CHOICE;
    padr( "Report Form",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_REPORT_DOCUMENT_NEW_CHOICE;
    padr( "Create/Modify/Print Text",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_REPORT_LABEL_NEW_MSG;
    "Create a label file '.LBL'"
#define ASSIST_MENU_REPORT_LABEL_MODIFY_MSG;
    "Modify a label file '.LBL'"
#define ASSIST_MENU_REPORT_LABEL_FORM_MSG;
    "Print labels"
#define ASSIST_MENU_REPORT_REPORT_NEW_MSG;
    "Create a report file ." + _EXTENTION_FORM
#define ASSIST_MENU_REPORT_REPORT_MODIFY_MSG;
    "Modify a report file ." +_EXTENTION_FORM
#define ASSIST_MENU_REPORT_REPORT_FORM_MSG;
    "Print a report"
#define ASSIST_MENU_REPORT_DOCUMENT_NEW_MSG;
    "Create / Modify / Print a text or document file"

#define ASSIST_MENU_INFO_ABOUT_CHOICE;
    padr( "About",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_INFO_F1_CHOICE;
    padr( "[F1] help",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_INFO_F2_CHOICE;
    padr( "[F2] list",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_INFO_F3_CHOICE;
    padr( "[F3] Alias info",;
    _MENU_TENDINA_WIDTH-2 )
#define ASSIST_MENU_INFO_F5_CHOICE;
    padr( "[F5] Set default to ...",;
    _MENU_TENDINA_WIDTH-2 )

#define ASSIST_MENU_INFO_ABOUT_MSG;
    "Copyright information"
#define ASSIST_MENU_INFO_MANUAL_PRINT;
    "Print manual"

//----------------------------------------------------------------
#define DBI_STRUCTURE_TOP;
   "Name     Type Len Dec"

//----------------------------------------------------------------
#define DIR_WAIT_READING_DIRECTORY;
    "Reading Directory..."

#define DIR_HELP;
    "DIRECTORY" +;
    NL(2) +;
    "Select disk drive, directory or file from a list." +;
    NL(2) +;
    "Use arrow keys, []/[], to move cursor. " +; 
    "Press [Enter] to select a drive, a directory or " +;
    "a file." +;
    NL(2) +;
    "If a directory is selected, or a new drive, it will " +;
    "appear the new list." + NL(1) +;
    "If a file is selected, the " +;
    "selection window disappers automatically and the program " +;
    "continues." +;
    NL(2) +;
    "If the only purpose is to select a drive or directory, " +;
    "this operation is terminated pressing [Esc] after " +;
    "reaching the desired location."

//----------------------------------------------------------------
#define DOC_FUNCTION_NAME       "Doc()"

#define DOC_TOP      1
#define DOC_LEFT     0
#define DOC_BOTTOM   maxrow()-1
#define DOC_RIGHT    maxcol()
#define DOC_WIDTH    SPACE_STR
#define DOC_TABSIZE  4

#define DOC_OLD      1
#define DOC_NAME     2
#define DOC_TEXT     3
#define DOC_WRAP     4
#define DOC_INS      5
#define DOC_ROW      6
#define DOC_COL      7
#define DOC_RELROW   8
#define DOC_RELCOL   9
#define DOC_CHANGED  10
#define DOC_LASTKEY  11
#define DOC_EXIT     12
#define DOC_NEW      13
#define DOC_LENGTH   13  // array length

#define DOC_MENU_FILE_CHOICE "File"

#define DOC_MENU_FILE_MSG;
    "Main menu"

#define DOC_MENU_FILE_NEW_CHOICE;
    padr( "New",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_OPEN_CHOICE;
    padr( "Open ...",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_SAVE_CHOICE;
    padr( "Save",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_SAVE_AS_CHOICE;
    padr( "Save as ...",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_SET_DEFAULT_TO_CHOICE;
    padr( "Set default to ...",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_PRINT_AS_IS_CHOICE;
    padr( "Print as it is",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_PRINT_CHOICE;
    padr( "Print standard",;
    _MENU_TENDINA_WIDTH-2 )
#define DOC_MENU_FILE_EXIT_CHOICE;
    padr( "Exit",;
    _MENU_TENDINA_WIDTH-2 )

#define DOC_MENU_FILE_NEW_MSG;
    "Removes currently loaded file from memory"
#define DOC_MENU_FILE_OPEN_MSG;
    "Loads new file into memory"
#define DOC_MENU_FILE_SAVE_MSG;
    "Saves current file"
#define DOC_MENU_FILE_SAVE_AS_MSG;
    "Saves current file with specified name"
#define DOC_MENU_FILE_SET_DEFAULT_TO_MSG;
    "Determinate the output peripheral or file"
#define DOC_MENU_FILE_PRINT_AS_IS_MSG;
    "Prints the actual file as it appears on the screen"
#define DOC_MENU_FILE_PRINT_MSG;
    "Prints actual text/document using eventually " +;
    "Fields, Memvars and commands"
#define DOC_MENU_FILE_EXIT_MSG;
    "Exit " + DOC_FUNCTION_NAME

#define DOC_WRITING     "Writing..."
#define DOC_WRITE_OK    "Write OK."
#define DOC_WRITE_ERROR "Write error."

#define DOC_PROMPT_OPEN_FILE_DOC;
    "Text or document file name:"
#define DOC_PROMPT_EXIT;
    "Exit " + DOC_FUNCTION_NAME + " ?"
#define DOC_PROMPT_ABANDON_CHANGES;
    "Abandon changes?"
#define DOC_PROMPT_OPEN_FILE_TEXT_DOCUMENT;
    "Text/Document file name:"
#define DOC_PROMPT_WHILE_EXPRESSION;
    "WHILE expression condition:"
#define DOC_PROMPT_FOR_EXPRESSION;
    "FOR expression condition:"
#define DOC_OPEN_DOCUMENT;
    "Open Document for editing."
#define DOC_INSERT_NEW_DOCUMENT_NAME;
    "Insert new Document name."
#define DOC_FILE_NOT_CHANGED;
    "File has not been modified."
#define DOC_NEW_NAME;
    "Enter new file name."

#define DOC_DIALOG_BOX_TOP_OPEN;
    "Open a Text/Document"
#define DOC_DIALOG_BOX_TOP_PRINT;
    "Print a Text/Document"

#define DOC_ERROR_FILE_CREATION;
    "File creation error!"

#define DOC_HELP;
    "Function " + DOC_FUNCTION_NAME +;
    NL(2) +;
    "This function creates and modifies little ASCII " +;
    "files: max 64K." +;
    NL(2) +;
    DOC_FUNCTION_NAME + " permits to create also formatted " +;
    "text with variables identified by  delimiters: " +;
    "when an active Alias exists, [F2] gives a list of " +;
    "insertable fields." +;
    NL(2) +;
    "It follows the available command list." +;
    NL(2) +;
    "[Esc]                     Exit" + NL(1) +;
    "[F1]                      this help" + NL(1) +;
    "[F2]                      field list" + NL(1) +;
    "[] / [Ctrl]+[E]          up" + NL(1) +;
    "[] / [Ctrl]+[X]          down" + NL(1) +;
    "[-] / [Ctrl]+[S]         left" + NL(1) +;
    "[-] / [Ctrl]+[D]         right" + NL(1) +;
    "[Ctrl]+[-] / [Ctrl]+[A]  word left" + NL(1) +;
    "[Ctrl]+[-] / [Ctrl]+[F]  word right" + NL(1) +;
    "[Home]                    line start" + NL(1) +;
    "[End]                     line end" + NL(1) +;
    "[Ctrl]+[Home]             top window" + NL(1) +;
    "[Ctrl]+[End]              bottom window" + NL(1) +;
    "[Pag]                    previous window" + NL(1) +;
    "[Pag]                    next window" + NL(1) +;
    "[Ctrl]+[Pag]             document start" + NL(1) +;
    "[Ctrl]+[Pag]             end document" + NL(1) +;
    "[Del]                     delete character" + NL(1) +;
    "[Backspace] / [----]     delete character left" + NL(1) +;
    "[Tab] / [--|]            insert tab" + NL(1) +;
    "[Ins]                     insert/overwrite" + NL(1) +;
    "[Enter]                   next line" + NL(1) +;
    "[Ctrl]+[Y]                delete line" + NL(1) +;
    "[Ctrl]+[T]                delete word right" + NL(1) +;
    "[F10] / [Alt]+[M]         menu"

#define DOC_HELP_PRINT;
    "TEXT/DOCUMENT STANDARD PRINT" +;
    NL(2) +;
    "It prints the loaded text/document linking with data " +;
    "obtained from the active Alias." +;
    NL(2) +;
    "WHILE EXPRESSION - the WHILE condition for the Alias " +;
    "records;" + NL(1) +;
    "FOR EXPRESSION - the FOR condition for the Alias " +;
    "records." +;
    NL(2) +;
    "The condition .T. (point, T, point) is ever true."

//----------------------------------------------------------------
#define DOT_MESSAGE_ERROR               "Error"
#define DOT_PROMPT_EXIT                 "Exit nB ?"
#define DOT_KEY_REMINDER;
    "[Esc] exit  [Pag]/[Pag] previous/next  [Enter] execute"
#define DOT_HELP;
    "nB DOT" +;
    NL(2) +;
    "nB works " +;
    "like an interactive xbase where only function " +;
    "are allowed." +;
    NL(2) +;
    "[Esc] closes the program." +;
    NL(2) +;
    "[Enter] executes the function written into " +;
    "the prompt line." +;
    NL(2) +;
    "[Pag] recalls previous line." +;
    NL(2) +;
    "[Pag] go to next command line." +;
    NL(2) +;
    "The usable functions are all those of CA-Clipper " +;
    "5.2 and the others made available into this program." +;
    NL(2) +;
    "ASSIST() calls a menu driven function usefull for " +;
    "creating and modifying database files and more." +;
    NL(3) +;
    "CA-Clipper is a trademark of Computer Associates."

//----------------------------------------------------------------
#define ERROR_MENU_CHOICE_BREAK         "Break"
#define ERROR_MENU_CHOICE_RETRY         "Retry"
#define ERROR_MENU_CHOICE_DEFAULT       "Default"
#define ERROR_MENU_CHOICE_QUIT          "Quit"

//----------------------------------------------------------------
#define EX_EXIT_NORMAL          0
#define EX_EXIT_BREAK           1
#define EX_EXIT_NO_MACRO_FILE   2
#define EX_ERROR_END_ALONE      "END alone.;Missing IF/WHILE/..."
#define EX_ERROR_NO_CONDITION   "No condition!"
#define EX_ERROR_NO_END         "No END.;END not found while searching for it."
#define EX_ERROR_TOO_MUCH_END   "Too much END."
#define EX_ERROR_MACRO_FILE_NOT_FOUND   "Macro file not found!"
#define EX_MAX_NEST             256
#define EX_MENU_CHOICE_ABORT    "Abort"
#define EX_MENU_CHOICE_CONTINUE "Continue"
#define EX_PROMPT_MACRO_EXECUTION_ABORT "Abort or continue macro execution?"

//----------------------------------------------------------------

#define FIELD_WINDOW_TOP_FIELD_CHOICE;
   "Name     Type Length Decimal"

#define FIELD_DIALOG_BOX_TOP_ALIGN_CHARACTER;
   "Character alignment"
#define FIELD_DIALOG_BOX_TOP_ALIGN_NUMBER;
   "Number alignment"

#define FIELD_PROMPT_ALIGN_NORMAL;
   "Normal? (Y/T/N/F)"
#define FIELD_PROMPT_ALIGN_CHARACTER_DIMENSION;
   "Dimention"
#define FIELD_PROMPT_ALIGN_CHARACTER_ALIGNMENT;
   "Alignment (Left/Center/Right)"
#define FIELD_PROMPT_ALIGN_NUMBER_LENGTH;
   "Number Length"
#define FIELD_PROMPT_ALIGN_NUMBER_LENGTH_NOTE;
   "(included comma and decimal)"
#define FIELD_PROMPT_ALIGN_NUMBER_DECIMAL;
   "Decimal Length"
#define FIELD_PROMPT_MEMO_LENGTH;
   "Insert memo length."

#define FIELD_ERROR_MEMO_FILEDS_NOT_ALLOWED;
   "MEMO fields cannot be included into index keys."

#define FIELD_TRUE;
   "True "
#define FIELD_FALSE;
   "False"

#define FIELD_HELP_FIELD_CHOICE;
    "FIELD CHOICE" +;
    NL(2) +;
    "It permits to select a column " +;
    "or a field from a file. " +;
    NL(2) +;
    "The user have " +;
    "to move the cursor with the arrow keys, []/[], " +;
    "and select the name pressing [Enter]." +;
    NL(2) +;
    "Pressing [Esc], the function terminates."

#define FIELD_HELP_ALIGN_CHARACTER;
    "ALIGN CHARACTER FIELD" +;
    NL(2) +;
    "It helps to define the dimension " +;
    "and the alignment of a field." +;
    NL(2) +;
    "When normal is ture (.t.), the field is left with " +;
    "no transformations." +;
    NL(2) +;
    "Otherwise, the dimension and alignment must " +;
    "be spesified exactly." +;
    NL(2) +;
    "Pressing [Esc] will be returned a normal " +;
    "alignment. Pressing [Pag] will be confirmd the choice selected."

#define FIELD_HELP_ALIGN_NUMBER;
    "ALIGN NUMERIC FIELD" +;
    NL(2) +;
    "It helps to define the dimension " +;
    "and the alignment of a numeric field." +;
    NL(2) +;
    "When normal is ture (.t.), the number appears with " +;
    "a default transformation into string; " +;
    "otherwise, the dimension must " +;
    "be specified exactly. " +;
    NL(2) +;
    "Pressing [Esc] will be returned a normal " +;
    "alignment. Pressing [Pag] will be confirmd " +;
    "the choice selected." +;
    NL(3) +;
    "Examples:" +;
    NL(2) +;
    "Lenght, Decimals,  Number        Result    " + NL(1) +;
    "  10        0      12345,456     _____12345" + NL(1) +;
    "  10        2      12345,456     __12345.45" + NL(1) +;
    "  14        3      12345,456     _____12345.456" + NL(1) +;
    "   5        3      12345,456     *****" +;
    NL(2) +;
    "With the last case the result is an error."

//----------------------------------------------------------------
// xcFormVal vector
#define FRM_PAGE_HDR                 1
#define FRM_GRP_EXPR                 2
#define FRM_SUB_EXPR                 3
#define FRM_GRP_HDR                  4
#define FRM_SUB_HDR                  5
#define FRM_PAGE_WIDTH               6
#define FRM_LINES_PAGE               7
#define FRM_LEFT_MARG                8
#define FRM_RIGHT_MARG               9
#define FRM_COL_COUNT               10
#define FRM_DBL_SPACED              11
#define FRM_SUMMARY                 12
#define FRM_PE                      13
#define FRM_PEBP                    14
#define FRM_PEAP                    15
#define FRM_PLAINPAGE               16

// dimensioni dei buffer
#define FRM_SIZE_FILE_BUFF        1990   // dimensione del file
#define FRM_SIZE_LENGTHS_BUFF      110
#define FRM_SIZE_OFFSETS_BUFF      110
#define FRM_SIZE_EXPR_BUFF        1440
#define FRM_SIZE_FIELDS_BUFF       300
#define FRM_SIZE_PARAMS_BUFF        24

// buffer offsets (start points)
// contined into cBuffer, that is the file (.frm) loaded
// into a variable and seen from the point fo view
// of CA-Clipper and not from the one of Dos.
#define FRM_LENGTHS_OFFSET           5
#define FRM_OFFSETS_OFFSET         115
#define FRM_EXPR_OFFSET            225
#define FRM_FIELDS_OFFSET         1665
#define FRM_PARAMS_OFFSET         1965

// Offsets contained in acFormBuffer[ FRM_FIELDS_BUFF ]
#define FRM_FIELD_OFFSET               0
#define FRM_FIELD_WIDTH_OFFSET         1
#define FRM_FIELD_TOTALS_OFFSET        6
#define FRM_FIELD_DECIMALS_OFFSET      7
#define FRM_FIELD_CONTENT_EXPR_OFFSET  9
#define FRM_FIELD_HEADER_EXPR_OFFSET  11

// Offsets contained in cParamsBuff
#define FRM_PAGE_HDR_OFFSET         1
#define FRM_GRP_EXPR_OFFSET         3
#define FRM_SUB_EXPR_OFFSET         5
#define FRM_GRP_HDR_OFFSET          7
#define FRM_SUB_HDR_OFFSET          9
#define FRM_PAGE_WIDTH_OFFSET      11
#define FRM_LNS_PER_PAGE_OFFSET    13
#define FRM_LEFT_MRGN_OFFSET       15
#define FRM_RIGHT_MRGN_OFFSET      17
#define FRM_COL_COUNT_OFFSET       19
#define FRM_DBL_SPACE_OFFSET       21
#define FRM_SUMMARY_RPT_OFFSET     22
#define FRM_PE_OFFSET              23
#define FRM_PLNPG_PEAP_PEBP_OFFSET 24

// FormBuffers
#define FRM_LENGTHS_BUFF            1
#define FRM_OFFSETS_BUFF            2
#define FRM_EXPR_BUFF               3
#define FRM_PARAMS_BUFF             4
#define FRM_FIELDS_BUFF             5

// anNum[5]
#define FRM_PAGE_HDR_NUM            1
#define FRM_GRP_EXPR_NUM            2
#define FRM_SUB_EXPR_NUM            3
#define FRM_GRP_HDR_NUM             4
#define FRM_SUB_HDR_NUM             5

#define FRM_HEAD_REPORT_ORDER;
   "Order"
#define FRM_HEAD_REPORT_HEADER;
   "Column Header"
#define FRM_HEAD_REPORT_CONTENT;
   "Content"
#define FRM_HEAD_REPORT_WIDTH;
   "Width"
#define FRM_HEAD_REPORT_DECIMALS;
   "dec."
#define FRM_HEAD_REPORT_TOTAL;
   "Totals"

#define FRM_DIALOG_BOX_TOP_REPORT;
   "REPORT FORM ..."

#define FRM_PROMPT_COMPRESSED_PRINT;
   "Compressed Print ? (Y/T/N/F) "
#define FRM_PROMPT_OPEN_FILE_REPORT;
   "Report file name:"
#define FRM_PROMPT_WHILE_EXPRESSION;
   "WHILE expression condition:"
#define FRM_PROMPT_FOR_EXPRESSION;
   "FOR expression condition:"
#define FRM_PROMPT_TO_FILE;
   "Periferal or destination file:"
#define FRM_PROMPT_NEW_FILE_NAME;
   "Insert new name."
#define FRM_PROMPT_SAVE_REPORT;
   "Name to use to save the report:"
#define FRM_PROMPT_REPORT_PAGE_WIDTH;
   "Page Width....:"
#define FRM_PROMPT_REPORT_LINES_PER_PAGE;
   "Lines per Page:"
#define FRM_PROMPT_REPORT_LEFT_MARGIN;
   "Left Margin...:"
#define FRM_PROMPT_REPORT_RIGHT_MARGIN;
   "Right Margin..:"
#define FRM_PROMPT_REPORT_DOUBLE_SPACED;
   "Double Spaced ?"
#define FRM_PROMPT_REPORT_PAGE_EJECT_BEFORE;
   "Page Eject Before Print........?"
#define FRM_PROMPT_REPORT_PAGE_EJECT_AFTER;
   "Page Eject After Print.........?"
#define FRM_PROMPT_REPORT_PLAIN_PAGE;
   "Plain Page......?"
#define FRM_PROMPT_REPORT_PAGE_HEADER;
   "Page Header"
#define FRM_PROMPT_REPORT_GROUP_HEADER;
   "Group Header"
#define FRM_PROMPT_REPORT_GROUP_EXPRESSION;
   "Group Expression"
#define FRM_PROMPT_REPORT_SUMMARY_REPORT_ONLY;
   "Summary Report Only............?"
#define FRM_PROMPT_REPORT_PAGE_EJECT_AFTER_GROUP;
   "Page Eject After Group.........?"
#define FRM_PROMPT_REPORT_SUB_GROUP_HEADER;
   "Sub Group Header"
#define FRM_PROMPT_REPORT_SUB_GROUP_EXPRESSION;
   "Sub Group Expression"

#define FRM_HELP_REPORT_FORM;
    "REPORT FORM" +;
    NL(2) +;
    "It may ask for:" +;
    NL(2) +;
    "- the report file name;" +;
    NL(2) +;
    "- the WHILE condition;" +;
    NL(2) +;
    "- the FOR condition." +;
    NL(2) +;
    "The condition .T. (point, T, point) is ever true."

#define FRM_HELP_REPORT_EDIT;
    "MODIFY REPORT FORM" +;
    NL(2) +;
    "It opens a report form file and edits it."

#define FRM_HELP_REPORT_EDIT_HEAD;
    "STANDARD REPORT" +;
    NL(2) +;
    "It creates or modifies a report file, " +;
    "." + _EXTENTION_FORM + " under the dBaseIII+ standard." +;
    NL(2) +;
    "The informations are divided into two parts: " +;
    "the head and groups, the columns." +;
    NL(2) +;
    "PAGE WIDTH - the page width in characters;" +;
    NL(2) +;
    "LINES PER PAGE - the usable lines per per page;" +;
    NL(2) +;
    "LEFT MARGIN - the left margin in characters;" +;
    NL(2) +;
    "DOUBLE SPACED? - double spaced print, yes or no;" +;
    NL(2)+;
    "PAGE EJECT BEFORE PRINT? - form feed before print, " +;
    "yes or no;" +;
    NL(2) +;
    "PAGE EJECT AFTER PRINT? - form feed after print, " +;
    "yes or no;" +;
    NL(2) +;
    "PLAIN PAGE? - plain page, yes or no;" +;
    NL(2) +;
    "PAGE HEADER - the page header, max 4 lines of 60 characters " +;
    "(the separation between una line and the other is " +;
    "obtained writing a semicolon, ';');" +;
    NL(2)+;
    "GROUP HEADER - the group title;" +;
    NL(2) +;
    "GROUP EXPRESSION - the group expression (when it " +;
    "changes, the group changes);" +;
    NL(2) +;
    "SUMMARY REPORT ONLY? - only totals and no columns, yes or no;" +;
    NL(2)+;
    "PAGE EJECT AFTER GROUP? - form feed when the group changes, " +;
    "yes or no;" +;
    NL(2)+;
    "SUB GROUP HEADER - sub group title;" +;
    NL(2) +;
    "SUB GROUP EXPRESSION - the sub group expression."

#define FRM_HELP_REPORT_EDIT_COLUMNS;
    "STANDARD REPORT" +;
    NL(2) +;
    "It creates or modifies a report file, " +;
    "." + _EXTENTION_FORM + ", under the dBaseIII+ standard." +;
    NL(2) +;
    "The informations are divided into two parts: " +;
    "the head and groups, the columns." +;
    NL(2) +;
    "ORDER - used only to change the columns order;" +;
    NL(2) +;
    "COLUMN HEADER - column head description " +;
    "(it can contain 4 lines separated with a semicolon);" +;
    NL(2)+;
    "CONTENT - the column expression;" +;
    NL(2) +;
    "WIDTH - the column witdth;" +;
    NL(2) +;
    "DEC. - the decimal lengh for numeric columns;" +;
    NL(2)+;
    "TOTALS - totals to be calculated, yes or no " +;
    "(usefull only for numeric columns)."

#define FRM_HELP_REPORT_SAVE;
    "REPORT FORM SAVE" +;
    NL(2) +;
    "It specify the name to use to save this report."

#define FRM_ERROR_CANNOT_OPEN;
    "Cannot open this file. See FERROR() error n. "
#define FRM_ERROR_CANNOT_CREATE;
    "Cannot create this file. See FERROR() error n. "
#define FRM_ERROR_FILE_TOO_LITTLE;
    "File too little or empty."
#define FRM_ERROR_FILE_NOT_SAVED;
    "File not saved!"
#define FRM_ERROR_FILE_NOT_VALID;
    "File not valid."
#define FRM_ERROR_SOMETHING_GONE_WRONG;
    "Something is gone wrong trying to save."
#define FRM_ERROR_PRINT_FILE_NOT_FOUND;
    "The file do not exists."

#define FRM_WAIT_LOADING;
    "Loading..."
#define FRM_WAIT_SAVING;
    "Saving..."

//----------------------------------------------------------------
#define IDB_DIALOG_BOX_TOP_NTX_NEW;
   "New Index"
#define IDB_DIALOG_BOX_TOP_OPEN_DBF;
   "Open ." + _EXTENTION_DBF
#define IDB_DIALOG_BOX_TOP_OPEN_NTX;
   "Open ." + _EXTENTION_INDEXBAG
#define IDB_DIALOG_BOX_TOP_REPLACE;
   "Replace ALL"
#define IDB_DIALOG_BOX_TOP_RECALL;
   "idbRecall()"
#define IDB_DIALOG_BOX_TOP_DELETE;
   "idbDelete()"
#define IDB_DIALOG_BOX_TOP_RELATION_DEFINITION;
   "Relation definition"
#define IDB_DIALOG_BOX_DBF_FILE;
   "." + _EXTENTION_DBF + " File"

#define IDB_WINDOW_TOP_DBF_STRUCTURE;
   "Database file structure"

#define IDB_WAIT_PACK;
   "Removing deleted records..."

#define IDB_PROMPT_OPEN_FILE_TO_MODIFY;
   "." + _EXTENTION_DBF + " file - modify structure:"
#define IDB_PROMPT_DBF_CREATE_LAST_STRUCTURE;
   "Do you want to use last structure?"
#define IDB_PROMPT_DBF_MODIFY_STRUCTURE;
   "Confirm structure modification?"
#define IDB_PROMPT_DBF_REPLACE;
   "Confirm replacement on all records?"
#define IDB_PROMPT_DBF_RECALL;
   "Confirm deleted record recall?"
#define IDB_PROMPT_DBF_DELETE;
   "Confirm record deletion?"
#define IDB_PROMPT_DBF_PACK;
   "Confirm file pack?"
#define IDB_PROMPT_NEW_NTX_FILENAME;
   "Index file name to create:"
#define IDB_PROMPT_TAG_NAME;
   "Order name (Tag) to be created:"
#define IDB_PROMPT_NEW_NTX_KEYEXPR;
   "Key expression:"
#define IDB_PROMPT_OPEN_NTX_FILENAME;
   "Insert the index file name to open:"
#define IDB_PROMPT_WHILE_EXPRESSION;
   "WHILE expression condition (.t.=ALL):"
#define IDB_PROMPT_FOR_EXPRESSION;
   "FOR expression condition (.t.=ALL):"
#define IDB_PROMPT_OPEN_DBF_FILENAME;
   "Filename:"
#define IDB_PROMPT_OPEN_DBF_ALIASNAME;
   "Alias:"
#define IDB_PROMPT_OPEN_DBF_SHARED;
   "Shared ? (Y/T/N/F):"
#define IDB_PROMPT_OPEN_DBF_READ_ONLY;
   "Read Only ? (Y/T/N/F):"
#define IDB_PROMPT_REPLACE_FIELD;
   "Field to replace:"
#define IDB_PROMPT_REPLACE_NEW_VALUE;
   "New value expression:"
#define IDB_PROMPT_TEMP_DELETE;
   "Do you want to delete the temporary file " + ;
   "corresponding to the old one before the modification?"
#define IDB_PROMPT_RELATION_DEFINITION_CHILDNAME;
   "Child Alias:"
#define IDB_PROMPT_RELATION_DEFINITION_EXPRESSION;
   "Relation expression:"
#define IDB_PROMPT_NEW_DBF_SAVE;
   "." + _EXTENTION_DBF + " file save; insert the name:"

#define IDB_ERROR_FILE_NOT_EXIST;
   "The file do not exists."
#define IDB_ERROR_RELATION;
   "Relation creation error!"

#define IDB_HELP_OPEN_FILE_TO_MODIFY;
   "Insert the file name that you intend to modify."
#define IDB_HELP_DBF_CREATE;
    "." + _EXTENTION_DBF + " CREATION" +;
    NL(2)  // more help will be attached
#define IDB_HELP_DBF_MODIFY_STRUCTURE;
    "." + _EXTENTION_DBF + " STRUCTURE MODIFICATION" +;
    NL(2) +;
    "Only the content of columns with the same name " +;
    "are saved; the others are new empty columns." +;
    NL(2)  // more help will be attached
#define IDB_HELP_DBF_MODIFY_STRUCTURE_BODY;
    "ORDER - The column order may be changed modifying the " +;
    "serial number on the left. This information is " +;
    "temporary, that is, it is not part of the database file." +;
    NL(2) +;
    "FIELD NAME - The column name may be long 10 character max, it " +;
    "must begin with a letter and may contains also " +;
    "numbers and '_'; the space ' ' is not allowed." +;
    NL(2) +;
    "TYPE - The column type may be 'C' for character, 'N' for " +;
    "numeric, 'D' for data, 'L' for logic and 'M' for " +;
    "memo." +;
    NL(2) +;
    "LENGTH - The length is the column width in character." +;
    NL(2) +;
    "DECIMAL - Is the decimal portion of LENGTH when the " +;
    "column type identifies a number. The maximum decimal " +;
    "dimention is LENGTH -1 as the decimal point takes one " +;
    "place." +;
    NL(3) +;
    "The maximum column (field) width may be different " +;
    "depending on the active database driver." + NL(1) +;
    "For DBFNTX, the standard, the maximum dimentions are " +;
    "as follows:" +;
    NL(2) +;
    "C - Character - max length  = 255" + NL(1) +;
    "                max decimal = 250" + NL(1) +;
    "    Width = LENGTH + (DECIMAL * 256)" + NL(1) +;
    "    Max Width = 64000" +;
    NL(2) +;
    "N - Numeric - max length  = 16" + NL(1) +;
    "              max decimal = LENGTH-1" + NL(1) +;
    "    Width = LENGTH" +;
    NL(2) +;
    "D - Data - length  = leave 0" + NL(1) +;
    "           decimal = leave 0" +;
    NL(2) +;
    "L - Logic - length  = leave 0" + NL(1) +;
    "            decimal = leave 0" +;
    NL(2) +;
    "M - Memo - length  = leave 0" + NL(1) +;
    "           decimal = leave 0" +;
    NL(3) +;
    "After completing the column (field) description, " +;
    "press [Esc] to exit: the name to use to save " +;
    "the new structure will be asked."
#define IDB_HELP_NTX_CREATE;
    "INDEX/TAG CREATION" +;
    NL(2) +;
    "Create an index file (Order Bag) or add a Tag to " +;
    "an Order Bag." +;
    NL(2) +;
    "Depending on the active Database Driver, the following " +;
    "may be asked." +;
    NL(2) +;
    "- The index file (Order Bag) name." + NL(1) +;
    "- The key expression." + NL(1) +;
    "- The FOR condition."  + NL(1) +;
    "- The Tag name."
#define IDB_HELP_SET_INDEX;
    "." + _EXTENTION_INDEXBAG + " FILE OPEN" +;
    NL(2) +;
    "It activates an index previously created."
#define IDB_HELP_DBF_OPEN;
    "." + _EXTENTION_DBF + " FILE OPEN" +;
    NL(2) +;
    "Open (use) another ." +;
    _EXTENTION_DBF + " file. " +;
    NL(2) +;
    "FILENAME - The filename is the complete file " +;
    "pathname of what is going to be opened." +;
    NL(2) +;
    "ALIAS - The Alias is the name to use as Alias." +;
    NL(2) +;
    "SHARED - When 'Shared' is set to true (T), the file is " +;
    "opened in shared mode." +;
    NL(2) +;
    "READ ONLY - When 'Read only' is set to " +;
    "true, it will not be possible any modification to " +;
    "the file."
#define IDB_HELP_DBF_REPLACE;
    "FIELD REPLACE" +;
    NL(2) +;
    "It permits to modify the content of a " +;
    "column. The replace can be controlled by the " +;
    "WHILE and FOR condition."
#define IDB_HELP_DBF_RECALL;
    "DELETED RECORD RECALL" +;
    NL(2) +;
    "It permits to recall as many records as " +;
    "defined by the FOR condition."
#define IDB_HELP_DBF_DELETE;
    "DELETE RECORDS" +;
    NL(2) +;
    "It permits to delete as many records as " +;
    "defined by the FOR condition."
#define IDB_HELP_RELATION_DEFINITION;
    "RELATION DEFINITION" +;
    NL(2) +;
    "It makes a link between the active Alias and a 'child' " +;
    "one." +;
    NL(2) +;
    "Usually, the 'child' Alias is already associated " +;
    "to an index file." +;
    NL(2) +;
    "The expression defines the relation using elements " +;
    "from the active Alias."
//----------------------------------------------------------------
#define LBL_REMARK    1
#define LBL_HEIGHT    2
#define LBL_WIDTH     3
#define LBL_MARGIN    4
#define LBL_LINES     5
#define LBL_SPACES    6
#define LBL_ACROSS    7

#define LBL_SIZE      1034  // dimensione standard dei file di etichetta

#define LBL_DIALOG_BOX_LABEL_FORM_TOP;
   "LABEL FORM ..."

#define LBL_WINDOW_FILE_SEARCH_TOP;
   "Open File"
#define LBL_WINDOW_LABEL;
   "LABEL"

#define LBL_PROMPT_COMPRESSED_PRINT;
   "Compressed Print ? (Y/T/N/F) "
#define LBL_PROMPT_WHILE_EXPRESSION;
   "WHILE expression condition:"
#define LBL_PROMPT_FOR_EXPRESSION;
   "FOR expression condition:"
#define LBL_PROMPT_TO_FILE;
   "Periferal or destination file:"
#define LBL_PROMPT_SAVE_QUESTION;
   "Save?"
#define LBL_PROMPT_LABEL_REMARK;
   "Remark:"
#define LBL_PROMPT_LABEL_HEIGHT;
   "Height:"
#define LBL_PROMPT_LABEL_WIDTH;
   "Width: "
#define LBL_PROMPT_LABEL_MARGIN;
   "Margin:"
#define LBL_PROMPT_LABEL_LINES;
   "Lines: "
#define LBL_PROMPT_LABEL_SPACES;
   "Spaces:"
#define LBL_PROMPT_LABEL_ACROSS;
   "Across:"
#define LBL_PROMPT_OPEN_FILE_LABEL;
   "Label file name:"
#define LBL_PROMPT_SAVE_FILE_LABEL;
   "Name to use to save the label:"

#define LBL_HELP_LABEL;
    "STANDARD LABEL" +;
    NL(2) +;
    "It creates or modifies a label file, " +;
    "." + _EXTENTION_LABEL + ", under the dBaseIII+ standard." +;
    NL(2) +;
    "Labels may be printed in more than one column and can " +;
    "contain 16 lines maximum." +;
    NL(2) +;
    "REMARK - it will not be printed." +;
    NL(2) +;
    "HEIGHT - the label content; vertical dimension." +;
    NL(2) +;
    "WIDTH - the label content; horizzontal dimension." +;
    NL(2) +;
    "MARGIN - the left margin in character." +;
    NL(2) +;
    "LINES - the vertical spacing between labels." +;
    NL(2) +;
    "SPACES - the horizzontal spacing in character " +;
    "between label columns." +;
    NL(2) +;
    "ACROSS - the number of label columns." +;
    NL(2) +;
    "1-16 - the label content."

#define LBL_HELP_LABEL_FORM;
    "LABEL FORM PRINT" +;
    NL(2) +;
    "It may ask for:" +;
    NL(2) +;
    "- the label file name;" +;
    NL(2) +;
    "- the WHILE condition;" +;
    NL(2) +;
    "- the FOR condition." +;
    NL(2) +;
    "The condition .T. (point, T, point) is ever true."

#define LBL_HELP_LABEL_EDIT;
    "LABEL EDIT" +;
    NL(2) +;
    "It opens a label file and edit it."

#define LBL_HELP_LABEL_SAVE;
    "LABEL SAVE" +;
    NL(2) +;
    "It permit to specify the name to use to save this label."

#define LBL_ERROR_CANNOT_OPEN;
    "Cannot open this file. See FERROR() error n. "
#define LBL_ERROR_CANNOT_CREATE;
    "Cannot create this file. See FERROR() error n. "
#define LBL_ERROR_FILE_TOO_LITTLE;
    "File too little or empty."
#define LBL_ERROR_FILE_NOT_SAVED;
    "File not saved!"
#define LBL_ERROR_FILE_NOT_VALID;
    "File not valid."
#define LBL_ERROR_SOMETHING_GONE_WRONG;
    "Something is gone wrong during the label save."
#define LBL_ERROR_PRINT_SOMETHING_GONE_WRONG;
    "The label print is failed. Maybe the DBF file is not compatible."
#define LBL_ERROR_PRINT_FILE_NOT_FOUND;
    "The file do not exist."

#define LBL_WAIT_LOADING;
    "Loading..."
#define LBL_WAIT_SAVING;
    "Saving..."

//----------------------------------------------------------------
#define SELECT_MESSAGE_NEW_AREA;
    "New Area"
#define SELECT_MESSAGE_NATURAL_ORDER;
    "Natural"
#define SELECT_WINDOW_TOP_SELECT;
    "Alias list"
#define SELECT_WINDOW_TOP_RDD;
    "Database driver list"
#define SELECT_WINDOW_TOP_ORDER;
    "Order list for the actual Alias"
#define SELECT_HELP_RDD;
    "DATABASE DRIVER SELECTION" +;
    NL(2) +;
    "Select a database driver."
#define SELECT_HELP_SELECT;
    "ALIAS SELECTION" +;
    NL(2) +;
    "Select an Alias or a new area."
#define SELECT_HELP_ORDER;
    "ORDER SELECTION" +;
    NL(2) +;
    "Select an index order."

//----------------------------------------------------------------
#define RPT_DEFAULT_COMMAND_SYMBOL      "*"
#define RPT_DEFAULT_LPP                 60
#define RPT_ERROR_PRINT_COMMAND_UNCOMPLETED     "Print command uncompleted."
#define RPT_ERROR_PRINT_FILE_NOT_FOUND          "File not found."
#define RPT_ERROR_PRINT_ISOLATED_END;
    RPT_DEFAULT_COMMAND_SYMBOL + "END isolated."
#define RPT_ERROR_PARENTESES    "Error with   delimiters."

//----------------------------------------------------------------
#define TEXT_KEY_REMINDER;
   "[Esc] exit, []/[]/[Pag]/[Pag] move text."

//----------------------------------------------------------------
#define TB_FUNCTION_NAME        "Tb()"

#define TB_APP_MODE_ON( b )     ( b:cargo := TRUE  )
#define TB_APP_MODE_OFF( b )    ( b:cargo := FALSE )
#define TB_APP_MODE_ACTIVE( b ) ( b:cargo )
#define TB_ERROR_ARRAY_DIFFERENT_DIMENTIONS;    
                                "TB() - Different dimention arrays."
#define TB_ERROR_NO_COLUMNS     "TB() - No Columns."
#define TB_ERROR_NO_PICTURES    "TB() - No Pictures."
#define TB_ERROR_RECORD_LOCKED  "TB() - Record locked.;Try again later."
#define TB_PROMPT_DELETE_RECORD "Delete Record?"
#define TB_WINDOW_BOTTOM_MEMO   "[Esc] cancel  [Ctrl]+[W] save"

#define TB_HELP;
    "Function " + TB_FUNCTION_NAME +;
    NL(2) +;
    "This function permits to browse the active Alias, " +;
    "eventually with relations." +;
    NL(3) +;
    "Special keys:" +;
    NL(2) +;
    "[Pag]        previous page;" + NL(1) +;  
    "[Pag]        next page;" + NL(1) +; 
    "[Ctrl]+[Pag] top of file;" + NL(1) +;  
    "[Ctrl]+[Pag] bottom of file;" + NL(1) +;  
    "[Ctrl]+[Home] first column;" + NL(1) +; 
    "[Ctrl]+[End]  last column;" + NL(1) +;  
    "[Ctrl]+[Y]    delete/recall record." +;
    NL(3) +;
    "When a memo field is edited:" +;
    NL(2) +;
    "[Esc]         cancel and close window;" + NL(1) +;
    "[Ctrl]+[Y]    line delete;" + NL(1) +;
    "[Ctrl]+[W]    save and close window."
   
//================================================================
// STATIC GLOBAL
//================================================================
// used by Doc()
static aEdit

//================================================================
// nB - Main function
//================================================================
function nB( cPar0,;
             cPar1,;
             cPar2,;
             cPar3,;
             cPar4,;
             cPar5,;
             cPar6,;
             cPar7,;
             cPar8,;
             cPar9 )
//
// nB( [<cPar0>], [<cPar1>], [<cPar2>], [<cPar3>],
//     [<cPar4>], [<cPar5>], [<cPar6>], [<cPar7>],
//     [<cPar8>], [<cPar9>] ) --> NIL
//
// This is the main program function.
//
// <cPar0>      May contains the name of a macro program (&)
//              to be executed. If no paramenter is given,
//              nB() starts with the Dot() function.
//
// <cPar1> ... <cPar9>  Optional parameters to send to a macro
//                      program via public variables:
//                      c_Par1 ... c_Par9.
//

    local cOldScreen
    local cOldColor             := setcolor()
    local bOldErrorHandler      := errorblock( {|e| ErrorHandler(e)} )
    local nOldRow               := row()
    local nOldCol               := col()

    local Result

    // translate the calling parameters into public variables
    // for macro (&) use.
    public c_Par0
    public c_Par1
    public c_Par2
    public c_Par3
    public c_Par4
    public c_Par5
    public c_Par6
    public c_Par7
    public c_Par8
    public c_Par9

    Memvar->c_Par0 := cPar0
    Memvar->c_Par1 := cPar1
    Memvar->c_Par2 := cPar2
    Memvar->c_Par3 := cPar3
    Memvar->c_Par4 := cPar4
    Memvar->c_Par5 := cPar5
    Memvar->c_Par6 := cPar6
    Memvar->c_Par7 := cPar7
    Memvar->c_Par8 := cPar8
    Memvar->c_Par9 := cPar9

    begin sequence

        // Copyright
        QQOut( PROGRAM_COPYRIGHT )

        // Copyright notice is saved with savescreen()
        cOldScreen      := savescreen()

        // Default
        set( _SET_DATEFORMAT, "dd/mm/yyyy" )
        set( _SET_DEFAULT, "")
        set( _SET_DELETED, "OFF" )
        set( _SET_CONSOLE, "ON" )
        set( _SET_CONFIRM, "ON" )
        set( _SET_SCOREBOARD, "OFF" )
        set( _SET_INSERT, "ON" )
        setcolor( COLOR_DEFAULT0 )
        rddsetdefault( "dbfntx" )

        // Macro execution or Command line activation
        if !( empty( cPar0 ) )
            // The function EX() will check alone the file
            // existance.
            EX( cPar0 )
        else
            // command line
            dot()
        end

        // end of program
        dbcloseall()
        dbcommitall()
        scroll()
        setpos(0,0)

    recover
        // nil
    end sequence
    errorblock(bOldErrorHandler)

    // restore old values
    setcolor( cOldColor )
    restscreen( NIL, NIL, NIL, NIL, cOldScreen )
    setpos( nOldRow, nOldCol )

    return NIL

//----------------------------------------------------------------
// DOT - COMMAND LINE
//----------------------------------------------------------------
static function Dot()
//
// Dot () --> NIL
//
// Interactive command line. Only functions and calculations
// allowed.
//

   local getlist := {}
   local bOld_F1 := setkey( K_F1, { || Text( DOT_HELP )} )
   local bOldErrorHandler
   local cOldScreen     := savescreen()

   local lMore     := .T.
   local cCommand  := ""
   local acCommand := {}
   local nMax      := 0
   local nInd      := 0

   // prepare command array and indexes
   aadd( acCommand, SPACE_STR )
   nMax := 1
   nInd := 1
      
   // clear screen
   scroll()
   
   // loop
   while .T.
      
      scroll( maxrow()-1,00,maxrow(),maxcol() )
      devpos( maxrow(), 00 )
      
      devpos( maxrow(), 00 )
      devout( padc( DOT_KEY_REMINDER, maxcol()+1 ) )
      devpos( maxrow(), 0 )
      devout( ltrim(str(nInd)) + "/" + ltrim(str(nMax)) )
      
      @maxrow()-1,00 get acCommand[nMax] picture "@s"+ltrim(str(maxcol()+1))+"@"
      
      readmodal(getlist)
      getlist := {}
      
      do case
      case lastkey() = K_PGUP  // previous
         if nInd > 1
            nInd--
            acCommand[nMax] := acCommand[nInd]
         end
      case lastkey() = K_PGDN  // next
         if nInd < nMax
            nInd++
            acCommand[nMax] := acCommand[nInd]
         end
      case lastkey() = K_ESC
         if alert( DOT_PROMPT_EXIT,;
                     { _MENU_YES, _MENU_NO } ) == 1
            exit
         end
      case lastkey() = K_ENTER
         cCommand := alltrim( acCommand[nMax] )
         if !( cCommand == "" )
            
            scroll(00,00,maxrow()-2,maxcol(), +2)
            devpos( maxrow()-3, 00 )
            devout( cCommand )
            
            bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
            begin sequence
               devpos( maxrow()-2, 00 )
               devout( &(cCommand) )
            recover
               devpos( maxrow()-2, 00 )
               devout( DOT_MESSAGE_ERROR )
            end
            errorblock(bOldErrorHandler)
            
            // array increment
            aadd( acCommand, SPACE_STR )
            nMax := len( acCommand )
            nInd := nMax
         end
      otherwise
         // do nothing
      end
   end

   setkey( K_F1, bOld_F1 )
   restscreen( NIL, NIL, NIL, NIL, cOldScreen )

   return NIL

//----------------------------------------------------------------
// ERROR - ERROR HANDLING
//----------------------------------------------------------------
static function ErrorHandler(e)
//
// errorblock( {|e| ErrorHandler(e)} )
//

    local cMessage
    local aOptions
    local nChoice
   
    // zero division => 0
    if ( e:genCode == EG_ZERODIV )
        return (0)
    end

    // network open error
    if ( e:genCode == EG_OPEN .and. e:osCode == 32 .and. e:canDefault )
        neterr(.T.)
        return (.F.)
    end

    // lock error during APPEND BLANK
    if ( e:genCode == EG_APPENDLOCK .and. e:canDefault )
        neterr(.T.)
        return (.F.)
    end

    // build error message
    cMessage := ErrorMessage(e)

    // build options array
    aOptions := {ERROR_MENU_CHOICE_BREAK}
    if (e:canRetry)
        aadd(aOptions, ERROR_MENU_CHOICE_RETRY)
    end
    if (e:canDefault)
        aadd(aOptions, ERROR_MENU_CHOICE_DEFAULT)
    end
    aadd(aOptions, ERROR_MENU_CHOICE_QUIT)
   
    // alert box
    nChoice := 0
    while ( nChoice == 0 )
        nChoice := alert( cMessage, aOptions )
        if ( nChoice == NIL )
            exit
        end
    end

    // do as instructed
    do case
    case ( aOptions[nChoice] == ERROR_MENU_CHOICE_BREAK )
        break(e)
    case ( aOptions[nChoice] == ERROR_MENU_CHOICE_RETRY )
        return (.T.)
    case ( aOptions[nChoice] == ERROR_MENU_CHOICE_DEFAULT )
        return (.F.)
    otherwise
        // exit
    end

    cMessage += ErrorMoreInfo(e)

    Text( cMessage )

    // end of program
    errorlevel(1)
    quit

    return (.F.)

//----------------------------------------------------------------
static function ErrorMessage(e)

    local cMessage

    // start error message
    cMessage := if( e:severity > ES_WARNING, "Error ", "Warning " )

    // add subsystem name if available
    if ( ValType(e:subsystem) == "C" )
        cMessage += e:subsystem()
    else
        cMessage += "???"
    end

    // add subsystem's error code if available
    if ( ValType(e:subCode) == "N" )
        cMessage += ("/" + alltrim( str(e:subCode) ) )
    else
        cMessage += "/???"
    end

    // add error description if available
    if ( ValType(e:description) == "C" )
        cMessage += ("  " + e:description)
    end

    // add either filename or operation
    if ( !Empty(e:filename) )
        cMessage += (": " + e:filename)
    elseif ( !Empty(e:operation) )
        cMessage += (": " + e:operation)
    end

    // add dos error
    if !( Empty(e:osCode) )
        cMessage += ";(DOS Error " + alltrim( str(e:osCode) ) + ")"
    end

    return cMessage

//----------------------------------------------------------------
static function ErrorMoreInfo(e)

   local cMessage := ""
   local i := 0

   cMessage += NL(2) 
   
   // procedure names
   i := 2 // to jump ErrorHandler() and ErrorMoreInfo()
   while ( !Empty(procname(i)) )
      cMessage += rtrim( ProcName(i) ) +;
                  "(" + alltrim( str( ProcLine(i) ) ) + ")" +;
                  NL(1)
      i++
   end

   // actual Alias
   if len( alias() ) == 0
      cMessage += NL(1)
      cMessage += "Actual area n. 0 - New Area"
   else
      cMessage += "Actual area n. " +; 
                  alltrim( str( select() ) ) +;
                  " - Alias: " + alias() + NL(1) +;
                  "----------"
   end
   
   // all alias info
   for i := 1 to _MAX_SELECT
      if !( alias(i) == "" )
         cMessage += NL(1)
         cMessage += (i)->(dbiStatus()) // alla fine c' un NL(1)        
         cMessage += "----------"
      end
   next

   return cMessage

//----------------------------------------------------------------
// DBI - DB INFO
//----------------------------------------------------------------
static function dbiStatus(lRecursion)

   local cFileInfo := ""
   local cOrdKey := ""
   local cLinkExpr := ""
   local nOrder    := 0
   local nRelation := 0
   local cFilter   := ""
   local nRelSelect := 0

   // default
   if lRecursion == NIL
      lRecursion := .F.
   end

   if len( alias() ) == 0
      // No file is available now.
   else
      cFileInfo := "Select n. " +;
          alltrim( str( select() ) ) + NL(1)
      cFileInfo += "Alias: " + alias() + NL(1)

      nOrder := 1
      while .T.
         cOrdKey := ordkey( nOrder )
         if cOrdKey == ""
            exit
         else
            cFileInfo += "Order n. " +;
                alltrim( str( nOrder ) ) +;
                ": " + ordname() + " " + cOrdKey + NL(1)
         end
         nOrder++
      end

      cFilter := dbfilter()
      if cFilter == ""
         // exit
      else
         cFileInfo += "Filter : " + cFilter + NL(1)
      end
      
      nRelation := 0
      while .T.
         cLinkExpr := dbrelation( ++nRelation )
         if cLinkExpr == ""
            exit
         else
            nRelSelect := dbrselect( nRelation )
            cFileInfo += "Relation n. " +;
                alltrim( str( nRelation ) ) +;
                " " + alias(nRelSelect) +;
                ": " + cLinkExpr + NL(1) +;
                "--------------------" + NL(1) +;
                (nRelSelect)->(dbiStatus(.T.)) + NL(1) +;
                "--------------------" + NL(1) +;
                "end Relation n. " +;
                alltrim( str( nRelation ) ) +;
                " " + alias(nRelSelect)
         end
      end

   end

   return cFileInfo

//----------------------------------------------------------------
static function dbiStructure()
//
// dbiStructure() --> cTextStructure
//                --> NIL
//
// It returns a text containing the active Alias structure

    local cStructure := ""
    local cFileName  := ""

    if alias() == ""  // null string = nessun file aperto
        alert( _ERROR_NO_ALIAS )
        return NIL
    end

    // alias save
    cFileName := alias()

    copy structure extended to &(strTempPath()+"\TMP_STRU")

    // nuovo select per non dover chiudere il file aperto
    dbusearea(.T., NIL, strTempPath()+"\TMP_STRU",;
        "TMP_STRU", .F., .F.)
    if neterr()
        select(cFileName)
        return NIL
    end

    TMP_STRU->(dbgotop())
    cStructure := "Alias: " + cFileName +;
         NL(2) +;
         DBI_STRUCTURE_TOP +;
         NL(2)
    while .T.
        cStructure +=;
            TMP_STRU->Field_Name + " " +;
            TMP_STRU->Field_Type + " " +;
            str( TMP_STRU->Field_Len, 3, 0 ) + " " +;
            str( TMP_STRU->Field_Dec, 4, 0 ) + NL(1)
        skip +1
        if TMP_STRU->(eof())
            exit
        end
    end

    // Close temp file.
    TMP_STRU->(dbclosearea())

    // restore previous Alias
    select(cFileName)

    return cStructure

//----------------------------------------------------------------
// MENU - K_LEFT/K_RIGHT
//----------------------------------------------------------------
static procedure MenuLeft()

   // redirect left key for menu
   keyboard chr(K_ESC)+chr(K_LEFT)+chr(K_ENTER)

   return

//----------------------------------------------------------------
static procedure MenuRight()

   // redirect right key for menu
   keyboard chr(K_ESC)+chr(K_RIGHT)+chr(K_ENTER)

   return

//----------------------------------------------------------------
// DIR - CHANGE DRIVE / DIRECTORY - SELECT FILE
//----------------------------------------------------------------
static function dir( cFileSpec, lDrives, lDirs, lFiles,;
    nSortColumn )
//
// Dir( [<cFileSpec>], [<lDrives>], [<lDirs>], [<lFiles>],
//      [<cBottomMessage>], [<nSortColumn>] ) --> cPathname
//
// <cFileSpec>  Filename, Pathname, also with wildcards.
//
// <lDrives>    true (.t.) means: include drives letters.
//
// <lDirs>      true (.t.) means: include directory names.
//
// <lFiles>     true (.t.) means: include file names.
//
// <cBottomMessage>     message to show to the bottom.
//
// <nSortColumn>        the column number to use to sort the
//                      list.
//                      The columns are:
//                      F_NAME = 1,
//                      F_SIZE = 2,
//                      F_DATE = 3,
//                      F_TIME = 4,
//                      F_ATTR = 5.
//                      It is not possible to sort for extention.
//
//

    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NONE )
    local nOldRow       := row()
    local nOldCol       := col()
    local bOld_F1       :=;
        setkey( K_F1, {|| Text( DIR_HELP ) } ) 
    local bOld_F2       := setkey( K_F2, NIL ) 
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
    
    local aDir := {}
    local nI := 0
    local cDrive
    local cPath
    local cWild
    local cFile
    
    local cNewDir
    local cNewDrive

    local cReturn

    default cFileSpec      to "*.*"
    default lDrives        to .t.
    default lDirs          to .t.
    default lFiles         to .t.
    default nSortColumn    to F_NAME

    begin sequence

        // correct eventual error on nSortColumn
        if nSortColumn > F_ATTR
            nSortColumn := F_NAME
        end

        // name division
        cDrive := strDrive( cFileSpec )
        cPath  := strPath( cFileSpec )

        do case
        case cDrive == "" .and.;
             cPath == "" .and.;
             curdir() == ""
            cPath := "\"
        case cDrive == "" .and.;
             cPath == "" .and.;
             !empty( curdir() )
            cPath := "\" + curdir() + "\"
        case cDrive == "" .and.;
             !empty( cPath )
            // ok, do nothing
        case !empty( cDrive ) .and.;
             cPath == "" .and.;
             curdir( cDrive ) == ""
            cPath := "\"
        case !empty( cDrive ) .and.;
             cPath == "" .and.;
             curdir( cDrive ) <> ""
            cPath := "\" + curdir( cDrive ) + "\"
        case !empty( cDrive ) .and.;
             !empty( cPath )
            // ok, do nothing
        end

        cWild  := StrFile( cFileSpec )

        // read the directory
        aDir := dirArray( cDrive+cPath+cWild, lDrives, lDirs, lFiles, nSortColumn )
   
        // Create a kind of window.
        nTop    := 0
        nLeft   := maxcol()-46
        nBottom := maxrow()
        nRight  := maxcol()
        nWidth   := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        dispbox( nTop, nLeft, nBottom, nLeft, B_DOUBLE )
        dispbox( nTop, nRight, nBottom, nRight, B_DOUBLE )

        while .t.

            setcolor( COLOR_DEFAULT1 )
            @nTop,nLeft+1 say padc( cDrive+cPath+cWild, nWidth-2 )
            @nBottom-1,nLeft+1 say padc( _KEY_ESC_EXIT_F1, nWidth-2 )
            @nBottom,nLeft+1 say padc( _KEY_UP_DOWN_ENTER, nWidth-2 )
            setcolor( COLOR_DEFAULT0 )

            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft+1, nBottom-2, nRight-1 )
            nI := achoice ( nTop+1, nLeft+1, nBottom-2, nRight-1, aDir )
       
            do case
            case nI == 0  // [Esc]
                exit
            case left( aDir[nI], 1 ) == "["  // directory
                // the at() function cannot work on "[" parentesis !!!
          
                cNewDir := left( aDir[nI], 15 )                    // "[EXAMPLE.XX]   "
                cNewDir := alltrim( cNewDir )                      // "[EXAMPLE.XX]"  
                cNewDir := substr( cNewDir, 2 )                    // "EXAMPLE.XX]
                cNewDir := substr( cNewDir, 1, len( cNewDir ) -1 ) // "EXAMPLE.XX"
          
                do case
                case cNewDir == "."
                    // don't move
                case cNewDir == ".."
                    // go back
                    cPath := strParent( cPath )
                    aDir := dirArray( cDrive+cPath+cWild, lDrives, lDirs, lFiles, nSortColumn ) 
                otherwise
                    cPath := cPath+cNewDir+"\"
                    aDir := dirArray( cDrive+cPath+cWild, lDrives, lDirs, lFiles, nSortColumn ) 
                end

            case left( aDir[nI], 1 ) == "<"  // drive
                cNewDrive := substr( aDir[nI], 2, 1 )
                
                cDrive := cNewDrive+":"
                if curdir(cDrive) == ""
                    cPath := "\"
                else
                    cPath := "\" + curdir(cDrive) + "\"
                end
                aDir := dirArray( cDrive+cPath+cWild, lDrives, lDirs, lFiles, nSortColumn ) 
      
            otherwise
                cFile := left( aDir[nI], 16 )
                cFile := alltrim( cFile )
                exit
            end
        end
   
    end sequence    
   
    // restore
    setkey( K_F1, bOld_F1 ) 
    setkey( K_F2, bOld_F2 ) 
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
   
    // exit
    if nI == 0  // [Esc]
        cReturn := cDrive+cPath
    else
        cReturn := cDrive+cPath+cFile
    end

    return cReturn
   
//----------------------------------------------------------------
static function dirArray( cFileSpec, lDrives, lDirs, lFiles, nSortColumn )

    local aDirectory
    local aDir := {}
    local nI := 0

    local cDrive := strDrive( cFileSpec )
    local cPath  := strPath( cFileSpec )
    
    if len( cPath ) > 0 .and.;
       right( cPath, 1 ) <> "\"
        cPath := cPath + "\"
    end

    WaitFor( DIR_WAIT_READING_DIRECTORY )
   
    // drives
    if lDrives
        aDirectory := dirArrayDrives()
        for nI := 1 to len( aDirectory )
            aadd( adir,; 
                left( "<" + aDirectory[nI] + ":>" + space(60), 60 ) )
        next
    end

    // directories
    if lDirs
        aDirectory := asort( directory( cDrive+cPath+"*.*", "D" ), NIL, NIL,;  // all included directories
                    { |x, y| x[nSortColumn] < y[nSortColumn] } )
    
        for nI := 1 to len( aDirectory )
            if aDirectory[nI,F_ATTR] <> "D"
                loop
            end
            aadd( adir, left( "[" + aDirectory[nI, F_NAME] + "]" + space(20), 15) +;
                str(aDirectory[nI,F_SIZE], 10, 0) + " " +;
                dtoc(aDirectory[nI,F_DATE]) + " " +;
                aDirectory[nI,F_TIME] + " " +;
                aDirectory[nI,F_ATTR]  )
        next
    end
   
    // files
    if lFiles
        aDirectory := asort( directory( cFileSpec, "" ), NIL, NIL,;  // only files
                    { |x, y| x[nSortColumn] < y[nSortColumn] } )
        for nI := 1 to len( aDirectory )
            aadd( aDir, left( aDirectory[nI,F_NAME] + space(20), 15)  +;
                str(aDirectory[nI,F_SIZE],10,0) + " " +;
                dtoc(aDirectory[nI,F_DATE]) + " " +;
                aDirectory[nI,F_TIME] + " " +;
                aDirectory[nI,F_ATTR] )
        next
    end
   
    WaitFor( NIL )
   
    return aDir

//----------------------------------------------------------------
static function dirArrayDrives()

    local aLogicalDisks := {}
    local nChar
   
    // chr(65) == "A"; chr(90) == "Z"
    for nChar := 65 to 90
        aadd( aLogicalDisks, chr( nChar ) )
    next
   
    return aLogicalDisks

//----------------------------------------------------------------
// TRUE - FUNCTIONS THAT RETURNS EVER .T.
//----------------------------------------------------------------
static function true( p1, p2, p3, p4, p5, p6, p7, p8, p9, p10 )
//
// true( , , , , , , , , , ) --> true
//
// It does nothing, it returns true.
//

    return .t.

//----------------------------------------------------------------
static function trueSetKey( nInkeyCode, bAction )
//
// trueSetKey( <nInkeyKode>, [<bAction>] ) --> true
//
// Just like SetKey() but returns ever true (.t.) and not the
// previous bAction.
//

    return true( setkey( nInkeyCode, bAction ) )

//----------------------------------------------------------------
// FIELD - FIELD SELECTION
//----------------------------------------------------------------
static function fieldNormal()
//
// this function calls fieldArray() and returns the selected 
// field name or "" if no choice was made.
//
   
    local aMyField := {}
    local cReturnString := ""
    
    begin sequence
   
        aMyField := fieldArray()
   
        do case
        case len( aMyField ) == 0
            cReturnString := ""
        otherwise
            cReturnString := aMyField[DBS_NAME]
        end
  
    end sequence
   
    return cReturnString

//----------------------------------------------------------------
static function fieldIndex()
//
// this function calls fieldArray() and returns a character
// string usefull for indexes creation or "" if no choice
// was made.
//
   
    local aMyField := {}
    local cReturnString := ""
   
    begin sequence
   
        aMyField := fieldArray()
    
        do case
        case len( aMyField ) == 0
            cReturnString := ""
        case aMyField[DBS_TYPE] == "C"
            cReturnString := "+upper(" + aMyField[DBS_NAME] + ")"
        case aMyField[DBS_TYPE] == "N" 
            cReturnString := "+str(" +;
                             aMyField[DBS_NAME] +;
                             "," +;
                             alltrim(str(aMyField[DBS_LEN])) +;
                             "," +;
                             alltrim(str(aMyField[DBS_DEC])) +;
                             ")"
        case aMyField[DBS_TYPE] == "D"
            cReturnString := "+dtos(" + aMyField[DBS_NAME] + ")"
        case aMyField[DBS_TYPE] == "L"
            cReturnString := "+transform(" + aMyField[DBS_NAME] + ",'L')"
        case aMyField[DBS_TYPE] == "M"
            alert( FIELD_ERROR_MEMO_FILEDS_NOT_ALLOWED )
        end

     end sequence
    
     return cReturnString
 
//----------------------------------------------------------------
static function fieldRelation()
//
// this function calls fieldArray() and returns a character
// string usefull for relation creation or "" if no choice
// was made.
//
   
    local aMyField := {}
    local cAlias   := alias()
    local cReturnString := ""
    
    begin sequence
   
        aMyField := fieldArray()
    
        do case
        case len( aMyField ) == 0
            cReturnString := ""
        case aMyField[DBS_TYPE] == "C"
            cReturnString := "+upper(" + cAlias + "->" + aMyField[DBS_NAME] + ")"
        case aMyField[DBS_TYPE] == "N" 
            cReturnString := "+str(" +;
                             cAlias + "->" + aMyField[DBS_NAME] +;
                             "," +;
                             alltrim(str(aMyField[DBS_LEN])) +;
                             "," +;  
                             alltrim(str(aMyField[DBS_DEC])) +;
                             ")"
        case aMyField[DBS_TYPE] == "D"
            cReturnString := "+dtos(" + cAlias + "->" + aMyField[DBS_NAME] + ")"
        case aMyField[DBS_TYPE] == "L"
            cReturnString := "+transform(" + cAlias + "->" + aMyField[DBS_NAME] + ",'L')"
        case aMyField[DBS_TYPE] == "M"
            alert( FIELD_ERROR_MEMO_FILEDS_NOT_ALLOWED )
        end
  
     end sequence
   
     return cReturnString

//----------------------------------------------------------------
static function fieldPicture()
//
// This function calls fieldArray() and asks for a picture.
// It returns a character string usefull for text substitution 
// or "" if no choice was made.
//
   
    local aMyField := {}
    local cReturnString := ""
    local lNormal := .F.
    local nDim := 20
    local cAlign := "L"
    local nNumLen := 10
    local nDecLen := 0
    local nMemoLen := 0

    begin sequence
   
        aMyField := fieldArray()
   
        do case
        case len( aMyField ) == 0
            cReturnString := ""
        case aMyField[DBS_TYPE] == "C"
            // alignment prompt
            fieldAlCharacter( @lNormal, @nDim, @cAlign, aMyField[DBS_NAME] )
            do case
            case lNormal
                cReturnString := aMyField[DBS_NAME]
            case cAlign == "L"
                cReturnString := [padr(alltrim(] + aMyField[DBS_NAME] +;
                                 [),] + alltrim(str(nDim)) + [)]
            case cAlign == "C"
                cReturnString := [padc(alltrim(] + aMyField[DBS_NAME] +;
                                 [),] + alltrim(str(nDim)) + [)]
            case cAlign == "R"
                cReturnString := [padl(alltrim(] + aMyField[DBS_NAME] +;
                                 [),] + alltrim(str(nDim)) + [)]
            otherwise
                cReturnString := aMyField[DBS_NAME]
            end
        case aMyField[DBS_TYPE] == "N" 
            // alignment prompt
            nNumLen := aMyField[DBS_LEN]
            nDecLen := aMyField[DBS_DEC]
            fieldAlNumber( @lNormal, @nNumLen, @nDecLen )
            do case
            case lNormal
                cReturnString := [str(] + aMyField[DBS_NAME] + [)]
            case nNumLen > nDecLen
                cReturnString := [str(] + aMyField[DBS_NAME] +;
                                 [,] + alltrim(str(nNumLen)) + [,] + alltrim(str(nDecLen)) + [)]
            otherwise
                cReturnString := [str(] + aMyField[DBS_NAME] + [)]
            end
        case aMyField[DBS_TYPE] == "D"
            cReturnString := [dtoc(] + aMyField[DBS_NAME] + [)]
        case aMyField[DBS_TYPE] == "L"
            cReturnString := [iif(] + aMyField[DBS_NAME] + [,"<] + FIELD_TRUE +[>","<] + FIELD_FALSE +[>" )]
        case aMyField[DBS_TYPE] == "M"
            nMemoLen := Accept( nMemoLen, FIELD_PROMPT_MEMO_LENGTH )
            if valtype(nMemoLen) == "N"
                cReturnString := [left(]  + aMyField[DBS_NAME] + [,] +alltrim(str(nMemoLen)) + [)]
            end
        end

   end sequence
   
   return cReturnString

//----------------------------------------------------------------
static function fieldAlCharacter( lNormal, nDim, cAlign )
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor()
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local bOld_F1               := setkey( K_F1, {|| Text( FIELD_HELP_ALIGN_CHARACTER ) } ) 
    local bOld_F2               := setkey( K_F2, NIL ) 
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
    
    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 7
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( FIELD_DIALOG_BOX_TOP_ALIGN_CHARACTER, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_ESC_NORMAL_F1_PGDN, nWidth ) )
   
        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-1, nRight )

            setpos( nTop, nLeft )
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_NORMAL
            @row()+1,nLeft get lNormal picture "L";
                when trueSetkey( K_F2, NIL );
                valid ( .t. )
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_CHARACTER_DIMENSION
            @row()+1,nLeft get nDim picture "999";
                when trueSetkey( K_F2, NIL ) .and.;
                    !lNormal;
                valid ( .t. )
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_CHARACTER_ALIGNMENT 
            @row()+1,nLeft get cAlign picture "!";
                when trueSetkey( K_F2, NIL ) .and.;
                    !lNormal;
                valid cAlign == "L" .or.;
                          cAlign == "C" .or.;
                          cAlign == "R"  
            
            read
            
            do case
            case lastkey() = K_ESC  // exit
                lNormal := .T.
                exit
            case lastkey() = K_PGDN // ok
                exit
            otherwise
                // loop
            end
        end

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor ) 
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function fieldAlNumber( lNormal, nNumLen, nDecLen )
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor()
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local bOld_F1               := setkey( K_F1, {|| Text( FIELD_HELP_ALIGN_NUMBER ) } ) 
    local bOld_F2               := setkey( K_F2, NIL ) 
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
    
    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 8
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( FIELD_DIALOG_BOX_TOP_ALIGN_NUMBER, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_ESC_NORMAL_F1_PGDN, nWidth ) )
   
        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-1, nRight )

            setpos( nTop, nLeft )
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_NORMAL
            @row()+1,nLeft get lNormal picture "L";
                when trueSetkey( K_F2, NIL );
                valid ( .t. )
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_NUMBER_LENGTH
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_NUMBER_LENGTH_NOTE
            @row()+1,nLeft get nNumLen picture "999"; 
                when trueSetkey( K_F2, NIL ) .and.;
                    !lNormal;
                valid ( .t. )
            @row()+1,nLeft say FIELD_PROMPT_ALIGN_NUMBER_DECIMAL
            @row()+1,nLeft get nDecLen picture "999";   
                when trueSetkey( K_F2, NIL ) .and.;
                    !lNormal;
                valid ( .t. )
            
            read
            
            do case
            case lastkey() = K_ESC  // exit
                lNormal := .T.
                exit
            case lastkey() = K_PGDN // ok
                exit
            otherwise
                // loop
            end
        end

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor ) 
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function fieldArray()
//   
// fieldArray() --> aFieldData
//              --> {}
//
// This function refers to the active alias and may be used as
// usually so:
//      (cAlias)->(fieldArray())
//
// This function returns an empty array when no selections are
// made; otherewise, an array containing the field data is
// returned.
//   

    local bOldErrorHandler
    local cOldScreen            := ""
    local cOldColor             := setcolor() 
    local nOldCursor            := setcursor( SC_NONE ) 
    local bOld_F1               := setkey( K_F1, {|| Text( FIELD_HELP_FIELD_CHOICE )} )
    local bOld_F2               := setkey( K_F2, NIL ) 
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
   
    local aField        := {}
    local aFieldEasy    := {}
    local nI            := 0

    begin sequence
    
        if alias() == ""
             alert( _ERROR_NO_ALIAS )
             break
        end
    
        aField := dbstruct()
   
        for nI := 1 to len( aField )
            aadd( aFieldEasy, left( aField[nI, DBS_NAME] + space(20), 11) +;
                aField[nI, DBS_TYPE] + "    " +;
                str( aField[nI, DBS_LEN], 4, 0) + "   " +;
                str( aField[nI, DBS_DEC], 4, 0)  )
        next
   
        // Create a kind of window.
        nTop    := 0
        nLeft   := maxcol()-30
        nBottom := maxrow()
        nRight  := maxcol()
        nWidth  := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        dispbox( nTop, nLeft, nBottom, nLeft, B_DOUBLE )
        dispbox( nTop, nRight, nBottom, nRight, B_DOUBLE )
   
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft+1, nTop, nRight-1 )
        scroll( nBottom-1, nLeft+1, nBottom, nRight-1 )
        @nTop,nLeft+1;
            say padr( FIELD_WINDOW_TOP_FIELD_CHOICE, nWidth-2 )
            // must be on the left
        @nBottom-2,nLeft+1;
            say padc( _KEY_ESC_CANCEL_F1, nWidth-2 )
        @nBottom-1,nLeft+1 say padc( _KEY_UP_DOWN, nWidth-2 )
        @nBottom,nLeft+1 say padc( _KEY_ENTER, nWidth-2 )

        setcolor( COLOR_DEFAULT2 )
        scroll( nTop+1, nLeft+1, nBottom-3, nRight-1 )
        nI := achoice ( nTop+1, nLeft+1, nBottom-2, nRight-1, aFieldEasy )

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
    setpos( nOldRow, nOldCol )
        
    if nI <> 0
        return { aField[nI, DBS_NAME],; 
                 aField[nI, DBS_TYPE],; 
                 aField[nI, DBS_LEN],; 
                 aField[nI, DBS_DEC] }
    end
   
    // otherwise
    return {}

//----------------------------------------------------------------
// SELECT - SELECTION LIST
//----------------------------------------------------------------
static function selectAlias()
//
// selectAlias()    --> cAlias
//                  --> ""
//

    local nSelect

    nSelect := selectArray()[1]

    if nSelect <> NIL;
        .and. nSelect <> 0
        return alias(nSelect)
    end

    return ""

//----------------------------------------------------------------
static function selectArray()
//
// selectChoice()  --> aSelect
//                 --> NIL
// selectChoice()[1] == Area number
// selectChoice()[2] == Alias
// selectChoice()[3] == Ordsetfocus
// selectChoice()[4] == OrdKey
//

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(SELECT_HELP_SELECT) } )
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
   
    local aSelect := {}
    local aSelectEasy := {}
    local nI := 0
    local aSelected             :=;
        { NIL, NIL, NIL, NIL }

    begin sequence
    
        // legge gli alias ed in particolare crea
        // il primo con il numero 0
        aadd( aSelect, { 0, SELECT_MESSAGE_NEW_AREA, "", "" } )
        for nI := 1 to _MAX_SELECT
            if alias(nI) == "" // non si pu usare il diverso (<>)
                // nix
            else
                aadd( aSelect, { nI, alias(nI), (alias(nI))->(ordsetfocus()), (alias(nI))->(ordkey(ordsetfocus()))  } )
            end
        next

        // travasa il vettore
        for nI := 1 to len( aSelect )
            aadd( aSelectEasy, str( aSelect[nI,1], 3 ) + " " +; 
                         padr( aSelect[nI,2], 10 ) + " " +; 
                         aSelect[nI,3] + " " +;
                         aSelect[nI,4] )
        next

        // Create a kind of window.
        nTop    := 0
        nLeft   := maxcol()-50
        // It is a big window as it must contain also the index
        // information when an index is associated.
        nBottom := maxrow()
        nRight  := maxcol()
        nWidth   := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        dispbox( nTop, nLeft, nBottom, nLeft, B_DOUBLE )
        dispbox( nTop, nRight, nBottom, nRight, B_DOUBLE )

        setcolor( COLOR_DEFAULT1 )
        @nTop,nLeft+1 say;
            padc( SELECT_WINDOW_TOP_SELECT, nWidth-2 )
        @nBottom-1,nLeft+1;
            say padc( _KEY_ESC_CANCEL_F1, nWidth-2 )
        @nBottom,nLeft+1;
            say padc( _KEY_UP_DOWN_ENTER, nWidth-2 )
        setcolor( COLOR_DEFAULT2 )
        scroll( nTop+1, nLeft+1, nBottom-2, nRight-1 )
        nI := achoice ( nTop+1, nLeft+1, nBottom-2, nRight-1, aSelectEasy )

        if nI <> 0
            aSelected := aSelect[nI]
        end

    end sequence

    // restore
    setcursor( nOldCursor )
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return aSelected

//----------------------------------------------------------------
static function selectOrder()
//
// selectOrder() --> nOrder
//                --> NIL
//
// It returns the order number selectable form the order list.
//

    local getlist               := {}
    local bOldErrorHandler
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NONE )
    local bOld_F1               :=;
        setkey( K_F1, { || Text(SELECT_HELP_ORDER) } )
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local aOrder := {}
    local aOrderEasy := {}
    local nI := 0

    begin sequence

        if alias() == ""  // nell'area attuale non c' un alias
            alert( _ERROR_NO_ALIAS )
            break
        end

        // Read the indexes including Zero as the natural order
        aadd( aOrder, { 0, SELECT_MESSAGE_NATURAL_ORDER, "" } )
        for nI := 1 to _MAX_ORDER
            if ordkey(nI) == "" // non si pu usare il diverso (<>)
                // nix
            else
                aadd( aOrder, { nI, ordname(nI), ordkey(nI) } )
            end
        next

        // Array transfer
        for nI := 1 to len( aOrder )
            aadd( aOrderEasy, str( aOrder[nI,1], 3 ) + " " +;
                aOrder[nI,2] + " " +;
                aOrder[nI,3] )
        next

        // Create a kind of window.
        nBottom     := maxrow()
        nRight      := maxcol()
        nTop        := 0
        nLeft       := nRight - 30
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
	@nTop,nLeft to nBottom,nLeft double
	@nTop,nRight to nBottom,nRight double
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft+1, nTop, nRight-1 )
        scroll( nBottom, nLeft+1, nBottom, nRight-1 )
        devpos( nTop, nLeft+1 )
        devout( padc( SELECT_WINDOW_TOP_ORDER, nWidth-2 ) )
        @nBottom-1,nLeft+1;
            say padc( _KEY_ESC_CANCEL_F1, nWidth-2 )
        @nBottom,nLeft+1;
            say padc( _KEY_UP_DOWN_ENTER, nWidth-2 )
        setcolor( COLOR_DEFAULT2 )
        scroll( nTop+1, nLeft+1, nBottom-2, nRight-1 )
        nI := achoice ( nTop+1, nLeft+1, nBottom-2, nRight-1, aOrderEasy )

    end sequence

    // restore
    setcursor( nOldCursor )
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    if nI <> 0
        return aOrder[nI,1]
    end

    // termina la procedura
    return NIL  // nel caso si esca con esc

//----------------------------------------------------------------
static function selectRdd()
//
// selectRDD() --> cRDD
//             --> NIL
//
// It returns a RDD name from the available ones.
//

    local getlist               := {}
    local bOldErrorHandler
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NONE )
    local bOld_F1               :=;
        setkey( K_F1, { || Text(SELECT_HELP_RDD) } )
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local aRDD := {}
    local aRDDX := {}
    local nI := 0

    begin sequence

        aRDD := rddlist( RDT_FULL )

        // Delete the driver names than do not contains index
        nI := 1
        for nI = 1 to len( aRDD )
            if right( aRDD[nI], 1 ) == "X"  // driver with index
                aadd( aRDDX, aRDD[nI] )
            end
        next

        // Create a kind of window.
        nBottom     := maxrow()
        nRight      := maxcol()
        nTop        := 0
        nLeft       := nRight - 30
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
	@nTop,nLeft to nBottom,nLeft double
	@nTop,nRight to nBottom,nRight double
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft+1, nTop, nRight-1 )
        scroll( nBottom, nLeft+1, nBottom, nRight-1 )
        devpos( nTop, nLeft+1 )
        devout( padc( SELECT_WINDOW_TOP_RDD, nWidth-2 ) )
        @nBottom-1,nLeft+1;
            say padc( _KEY_ESC_CANCEL_F1, nWidth-2 )
        @nBottom,nLeft+1;
            say padc( _KEY_UP_DOWN_ENTER, nWidth-2 )
        setcolor( COLOR_DEFAULT2 )
        scroll( nTop+1, nLeft+1, nBottom-2, nRight-1 )
        nI := achoice ( nTop+1, nLeft+1, nBottom-2, nRight-1, aRDDX )

    end sequence

    // restore
    setcursor( nOldCursor )
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    if nI <> 0
        return aRDDX[nI]
    end

    // if nI == 0, [Esc]
    return NIL

//----------------------------------------------------------------
// IDB - INTERACTIVE DB FUNCTIONS
//----------------------------------------------------------------
static function idbModStructure()
//
// idbModStructure() --> NIL
//
// Permette di modificare un file DBF chiedendo l'inserimento
// del nome del file stesso.

    local getlist               := {}
    local bOldErrorHandler
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL )
    local bOld_F1               :=;
        setkey( K_F1, {|| Text( IDB_HELP_OPEN_FILE_TO_MODIFY ) } )
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local cFileName := padr( "*." + _EXTENTION_DBF, SPACE_LEN )
    local lGoOn := .F.

    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 4
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_DBF_FILE, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )

            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_OPEN_FILE_TO_MODIFY
            @row()+1,nLeft get cFileName;
                picture "@s"+ltrim(str(nWidth))+"@";
                when trueSetkey( K_F2,;
                     {|| cFileName :=;
                     padr( Dir( cFileName ), SPACE_LEN ) } );
                valid file( cFileName :=;
                      padr( strAddExtention( cFileName,;
                      _EXTENTION_DBF ), SPACE_LEN) )
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                if file( strAddExtention( cFileName, _EXTENTION_DBF ) )
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
        end

        // close window
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )

        if lGoOn
           idbStructure( cFileName )
        end

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function idbStructure(cName)
//
// idbStructure(<cName>) --> NIL
//
// <cName>  database file name.
//
// It modifies or creates the file (.dbf) <cName>
//

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local nOldSelect            := select()
    local bOld_F1               := setkey( K_F1, NIL ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
    local lOldDeleted           := set( _SET_DELETED, .T. )

    local aStruct := {}
    local acCol 
    local acColSayPic
    local acColHead
    local alColCalc
    local acColValid
    local acColMsg
    local acColSep
    local nOrder := 0
    local lCreate := .F.
    local lGoOn := .F.
    local cNewName
    local nI
    local cTempPath := strTempPath()
   
    begin sequence

        if cName == NIL 
            // the file must be created
            lCreate := .T.
            cName := SPACE_STR
        else
            // eventual extention addition
            cName := strAddExtention( cName, _EXTENTION_DBF )
            if !file( cName )
                alert( cName + ";" + IDB_ERROR_FILE_NOT_EXIST )
                break
            end
        end
   
        if lCreate
            setkey( K_F1, { || Text(IDB_HELP_DBF_CREATE +;
                IDB_HELP_DBF_MODIFY_STRUCTURE_BODY) } )
        else
            setkey( K_F1, { || Text(IDB_HELP_DBF_MODIFY_STRUCTURE +;
                IDB_HELP_DBF_MODIFY_STRUCTURE_BODY) } )
        end

        // a structure file is created
        aStruct := {}
        aadd( aStruct, { "Order",       "N",  6, 0 } )
        aadd( aStruct, { "Field_Name",  "C", 10, 0 } )
        aadd( aStruct, { "Field_Type",  "C",  1, 0 } )
        aadd( aStruct, { "Field_Len",   "N",  3, 0 } )
        aadd( aStruct, { "Field_Dec",   "N",  4, 0 } )
        dbcreate( cTempPath + "\TMPKSTRU", aStruct )

        // the structure file is opened
        dbusearea( .T., NIL, cTempPath + "\TMPKSTRU",;
            "TMPKSTRU", .F., .F. )
        if neterr()
            // if fail, break
            break
        end

        // create structure index file
        dbcreateindex( cTempPath+"\TMPKSTRU", "Field->Order",;
            { || Field->Order } )
        // test if modify o creation is on
        if !lCreate 
            // try to open the existing file
            bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                // If the file has a structure defect,
                // an error occurs and the files TMP_STRU and
                // DBF_MODIFY must be closed.
                dbusearea( .T., NIL, cName,; 
                    "DBF_MODIFY", .F., .F. )
            recover
                TMPKSTRU->(dbclosearea())
                DBF_MODIFY->(dbclosearea())
                break
            end sequence
            errorblock( bOldErrorHandler )
            if neterr()    
                // if fail, break
                alert( alltrim( cName ) + ";" +;
                    _ERROR_FLOCK_FAILURE )
                TMPKSTRU->(dbclosearea())
                break
            end
            // copy structure
            copy structure extended to;
                &(cTempPath+"\TMP_STRU")
            // open structure file
            dbusearea( .T., NIL, cTempPath + "\TMP_STRU",;
                "TMP_STRU", .F., .F. )
            if neterr()    
                break
            end

            // if everything goes ok
            TMP_STRU->(dbgotop())
            nOrder := 0
            while .T.
                TMPKSTRU->(dbappend())
                nOrder += 600
                TMPKSTRU->Order := nOrder
                TMPKSTRU->Field_Name := TMP_STRU->Field_Name
                TMPKSTRU->Field_Type := TMP_STRU->Field_Type
                TMPKSTRU->Field_Len  := TMP_STRU->Field_Len 
                TMPKSTRU->Field_Dec  := TMP_STRU->Field_Dec
                TMP_STRU->(dbskip())
                if TMP_STRU->(eof())
                    exit
                end
            end
         
            // close
            TMP_STRU->(dbclosearea())

            // make a temporary copy.
            // rename is not possible as a memo (.dbt) file
            // can be present.
            
            select( "DBF_MODIFY" )
            __dbCopy( cTempPath+"\TEMP", NIL,;
                { || WaitFileEval() } )

            // close
            DBF_MODIFY->( dbCloseArea() )

        end

        // select
        select( "TMPKSTRU" )
        dbgotop()
   
        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := 0
        nRight      := maxcol()
        nLeft       := nRight - 41
        nWidth      := nRight - nLeft +1
        cOldScreen  := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_WINDOW_TOP_DBF_STRUCTURE, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_END_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_CTRLY, nWidth ) )

        setcolor( COLOR_DEFAULT2 )
        @nTop,nLeft to nBottom,nLeft double
        @nTop,nRight to nBottom,nRight double
        scroll( nTop+1, nLeft+1, nBottom-2, nRight-1 )
        
        acCol        := { "if( deleted(), '*', ' ' )",;
                        "Order",;
                        "Field_Name",;
                        "Field_Type",;
                        "Field_Len",;
                        "Field_Dec" }
                        
        acColSayPic  := { "!",;
                        "999999",;
                        "!!!!!!!!!!",;
                        "!",;
                        "999",;
                        "9999" }
                        
        acColHead    := { "*",;
                        "Order",;
                        "Field Name",;
                        "Type",;
                        "Length",;
                        "Decimal" }
                        
        alColCalc    := { .T.,;
                        .F.,;
                        .F.,;
                        .F.,;
                        .F.,;
                        .F. }
   
        // TB( <nTop>, <nLeft>, <nBottom>, <nRight>,
        //      <acCol>, <acColSayPic>, [<acColHead>], 
        //      [<alColCalc>, [<nFreeze>],
        //      [<lModify>], [<lAppend>], [<lDelete>],
        //      [<lAutosort>] )  --> NIL

        TB( nTop+1, nLeft+1, nBottom-2, nRight-1,;
            acCol, acColSayPic, acColHead, alColCalc,;
            1, .T., .T., .T., .T. )
        pack
        
        sort to (cTempPath+"\TMPSSTRU") on Order 
        TMPKSTRU->(dbclosearea())

        // closes window
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        cNewName := padr( strAddExtention( cName, _EXTENTION_DBF ), SPACE_LEN )

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 4
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_DBF_FILE, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )
      
        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_NEW_DBF_SAVE
            @row()+1,nLeft get cNewName picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cNewName := padr( Dir( cNewName ), SPACE_LEN ) } );      
                valid ( ( cNewName := padr(strAddExtention(cNewName, _EXTENTION_DBF), SPACE_LEN) ) <> NIL )    
            read

            do case
            case lastkey() = K_ESC  // exit
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN // ok
                do case
                case strAddExtention(cNewName, _EXTENTION_DBF);
                    <> strAddExtention( cName, _EXTENTION_DBF );
                    .and. file( strAddExtention( cNewName, _EXTENTION_DBF ) );
                    .and. ( alert( alltrim(cNewName) + ";" +;
                    _ERROR_PROMPT_ALREADY_EXIST_OVERWRITE,;
                    { _MENU_NO, _MENU_YES } ) == 1 )
                    //
                    // loop
                case strCutExtention( cNewName ) == ""
                        // loop
                otherwise
                    lGoOn := .T.
                    exit
                end
            
            otherwise
                // loop
            end
        end
        
        if lGoOn
            // ok save
            cNewName := strAddExtention( cNewname, _EXTENTION_DBF )
        else
            break
        end
      
        // the file is created
        bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
        begin sequence
            // the file is created deleting before the
            // eventual memo file
            cNewName := strCutExtention( cNewName )
            ferase( cNewName+".DBT" )
            create (cNewName) FROM &(cTempPath+"\TMPSSTRU") NEW
            select( strFile(cNewName) )

            if !lCreate
                // the content is copied
                append FROM &(cTempPath+"\TEMP") FOR WaitFileEval()
                // temporary file deletion
                if ( alert( IDB_PROMPT_TEMP_DELETE,;
                    { _MENU_NO, _MENU_YES } ) == 2 )
                    //
                    ferase( cTempPath+"\TEMP." + _EXTENTION_DBF )
                    ferase( cTempPath+"\TEMP.DBT" )
                end
            end
        
            // close
            dbclosearea()
            // temporary file deletion
            ferase( cTempPath+"\TMP_STRU." + _EXTENTION_DBF )
            ferase( cTempPath+"\TMPKSTRU." + _EXTENTION_DBF )
            ferase( cTempPath+"\TMPKSTRU." + _EXTENTION_INDEXBAG )
            ferase( cTempPath+"\TMPSSTRU." + _EXTENTION_DBF )
      
        recover
            (cNewName)->( dbCloseArea() )
            dbusearea( .T., NIL, cTempPath+"\TEMP", "Temp", .F. )
            select( "Temp" )
            copy TO &(cNewName) FOR WaitFileEval()
            ("Temp")->( dbCloseArea() )
            WaitProgress(2)
        end sequence
        errorblock(bOldErrorHandler)
   
    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
    set( _SET_DELETED, lOldDeleted )
    dbselectarea( nOldSelect )

    return NIL

//----------------------------------------------------------------
static function idbUseArea()
//
// idbUseArea() --> NIL
//
// It Opens a (.dbf) file.
//   

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_DBF_OPEN) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
    
    local cDbfName              := padr( "*." + _EXTENTION_DBF, SPACE_LEN )
    local cAliasName            := space(10) // max alias name length
    local lShared               := .T.
    local lReadOnly             := .T.
    local lGoOn

    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 10
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_OPEN_DBF, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )

            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_OPEN_DBF_FILENAME
            @row()+1,nLeft get cDbfName;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cDbfName :=;
                    padr( Dir( cDbfName ), SPACE_LEN ) } );
                valid ( file ( cDbfName :=;
                    padr(strAddExtention(cDbfName, _EXTENTION_DBF),;
                    SPACE_LEN) ) );
                    .and. !( "*" $ cDbfName );
                    .and. !( "?" $ cDbfName )
            @row()+1,nLeft say IDB_PROMPT_OPEN_DBF_ALIASNAME
            @row()+1,nLeft get cAliasName;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when ( truesetkey( K_F2, NIL );
                    .and. (cAliasName:=;
                    padr( strCutExtention( strFile(cDbfName) ),;
                    10) ) <> NIL );
                    valid ( cAliasName <> SPACE_STR )
            @row()+1,nLeft say IDB_PROMPT_OPEN_DBF_SHARED
            @row()+1,nLeft get lShared picture "@L";
                when truesetkey( K_F2, NIL )  
            @row()+1,nLeft say IDB_PROMPT_OPEN_DBF_READ_ONLY
            @row()+1,nLeft get lReadOnly;
                picture "@L";  
                when truesetkey( K_F2, NIL )
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                // chack for valid data
                if file( strAddExtention( cDbfName, _EXTENTION_DBF ) );
                    .and. cAliasName <> SPACE_STR
                    //
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
      
        end

        restscreen( nTop, nLeft,;
            nBottom, nRight, cOldScreen )
    
        if lGoOn
            bOldErrorHandler :=;
                errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                // the file is opened
                dbusearea( .T., NIL,;
                    alltrim(cDbfName), alltrim(cAliasName),;
                    lShared, lReadOnly )
            recover
                // nix
            end sequence
            errorblock(bOldErrorHandler)
        end
   
    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL
//----------------------------------------------------------------
static function idbReplace()
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_DBF_REPLACE) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local lGoOn := .F.

    static cField
    static cExpr
    static cWhileCondition
    static cForCondition

    if cWhileCondition == NIL .or.;
        cWhileCondition = SPACE_STR
        cWhileCondition := padr( ".T.", SPACE_LEN ) 
    end
   
    if cForCondition == NIL .or.;
        cForCondition = SPACE_STR
        cForCondition := padr( ".T.", SPACE_LEN ) 
    end

    if cField = NIL
        cField := space(10)
    end

    if cExpr == NIL
        cExpr := SPACE_STR
    end

    begin sequence

        if alias() == ""
            alert( _ERROR_NO_ALIAS )
            break
        end

        // try to lock
        if flock()
            // ok
        else
            alert( _ERROR_FLOCK_FAILURE )
            break
        end
   
        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 10
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_REPLACE, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_REPLACE_FIELD
            @row()+1,nLeft get cField;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cField :=;
                    padr( fieldNormal(),;
                    10 ) } );
                valid ( .t. )
            @row()+1,nLeft say IDB_PROMPT_REPLACE_NEW_VALUE
            @row()+1,nLeft get cExpr;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cExpr :=;
                    padr( rtrim(cExpr) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( .t. )
            @row()+1,nLeft say IDB_PROMPT_WHILE_EXPRESSION
            @row()+1,nLeft get cWhileCondition;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cWhileCondition :=;
                    padr( rtrim(cWhileCondition) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( iif( cWhileCondition == SPACE_STR,;
                    ".T.", cWhileCondition ) <> NIL )
            @row()+1,nLeft say IDB_PROMPT_FOR_EXPRESSION
            @row()+1,nLeft get cForCondition;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cForCondition :=;
                    padr( rtrim(cForCondition) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( iif( cForCondition == SPACE_STR,;
                    ".T.", cForCondition ) <> NIL )
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                if alert( IDB_PROMPT_DBF_REPLACE,;
                    { _MENU_NO, _MENU_YES } ) == 2
                    //
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
        end
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        if lGoOn
            bOldErrorHandler :=;
                errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                cField := padr( alltrim(cField), 10 )
                dbgotop()
                dbeval( {|| _FIELD->&(cField) :=;
                    &(rtrim(cExpr))},;
                    {|| &(alltrim(cForCondition) );
                    .AND. !deleted();
                    .AND. WaitFileEval()},;
                    {|| &(alltrim(cWhileCondition))},,, .F. )
                dbgotop()
               
            recover
                dbgotop()
            end sequence
            WaitProgress(2)
            errorblock(bOldErrorHandler)
        end

        dbunlock()

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function idbRecall()
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_DBF_RECALL) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local lGoOn := .F.

    static cWhileCondition
    static cForCondition

    if cWhileCondition == NIL .or.;
        cWhileCondition = SPACE_STR
        cWhileCondition := padr( ".T.", SPACE_LEN ) 
    end
   
    if cForCondition == NIL .or.;
        cForCondition = SPACE_STR
        cForCondition := padr( ".T.", SPACE_LEN ) 
    end

    begin sequence

        if alias() == ""
            alert( _ERROR_NO_ALIAS )
            break
        end

        // try to lock
        if flock()
            // ok
        else
            alert( _ERROR_FLOCK_FAILURE )
            break
        end
   
        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 6
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_RECALL, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )

            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_WHILE_EXPRESSION
            @row()+1,nLeft get cWhileCondition;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cWhileCondition :=;
                    padr( rtrim(cWhileCondition) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( iif( cWhileCondition == SPACE_STR,;
                    ".T.", cWhileCondition ) <> NIL )
            @row()+1,nLeft say IDB_PROMPT_FOR_EXPRESSION
            @row()+1,nLeft get cForCondition;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cForCondition :=;
                    padr( rtrim(cForCondition) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( iif( cForCondition == SPACE_STR,;
                    ".T.", cForCondition ) <> NIL )
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                if alert( IDB_PROMPT_DBF_RECALL,;
                    { _MENU_NO, _MENU_YES } ) == 2
                    //
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
        end
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        if lGoOn
            bOldErrorHandler :=;
                errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                dbgotop()
                dbeval( {|| dbrecall()},;
                    {|| deleted();
                    .and. &(alltrim(cForCondition));
                    .and. WaitFileEval() },;  
                    {|| &(alltrim(cWhileCondition))},,, .F. )
                dbgotop()
               
            recover
                dbgotop()
            end sequence
            WaitProgress(2)
            errorblock(bOldErrorHandler)
        end

        dbunlock()

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function idbDelete()
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_DBF_DELETE) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local lGoOn := .F.

    static cWhileCondition
    static cForCondition

    if cWhileCondition == NIL .or.;
        cWhileCondition = SPACE_STR
        cWhileCondition := padr( ".F.", SPACE_LEN ) 
    end
   
    if cForCondition == NIL .or.;
        cForCondition = SPACE_STR
        cForCondition := padr( ".F.", SPACE_LEN ) 
    end

    begin sequence

        if alias() == ""
            alert( _ERROR_NO_ALIAS )
            break
        end

        // try to lock
        if flock()
            // ok
        else
            alert( _ERROR_FLOCK_FAILURE )
            break
        end
   
        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 6
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_DELETE, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_WHILE_EXPRESSION
            @row()+1,nLeft get cWhileCondition;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cWhileCondition :=;
                    padr( rtrim(cWhileCondition) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( iif( cWhileCondition == SPACE_STR,;
                    ".T.", cWhileCondition ) <> NIL )
            @row()+1,nLeft say IDB_PROMPT_FOR_EXPRESSION
            @row()+1,nLeft get cForCondition;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cForCondition :=;
                    padr( rtrim(cForCondition) + fieldNormal(),;
                    SPACE_LEN ) } );
                valid ( iif( cForCondition == SPACE_STR,;
                    ".T.", cForCondition ) <> NIL )
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                if alert( IDB_PROMPT_DBF_DELETE,;
                    { _MENU_NO, _MENU_YES } ) == 2
                    //
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
        end
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        if lGoOn
            bOldErrorHandler :=;
                errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                dbgotop()
                dbeval( {|| dbdelete()},;
                    {|| !deleted();
                    .and. &(alltrim(cForCondition));
                    .and. WaitFileEval() },;  
                    {|| &(alltrim(cWhileCondition))},,, .F. )
                dbgotop()
               
            recover
                dbgotop()
            end sequence
            WaitProgress(2)
            errorblock(bOldErrorHandler)
        end

        dbunlock()

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function idbPack()
//
// It tries to make a PACK, and if it fails, there is no program
// interruption.
//

    local bOldErrorHandler         
   
    if alias() == ""  // no alias
        alert( _ERROR_NO_ALIAS )
        return NIL
    end

    if alert( IDB_PROMPT_DBF_PACK,;
        { _MENU_NO, _MENU_YES } ) == 1
        return NIL
    end
            
    WaitfOR( IDB_WAIT_PACK )

    bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
    begin sequence
        pack
    recover
        // nix
    end sequence
    errorblock(bOldErrorHandler)
   
    WaitFor()

    return NIL

//----------------------------------------------------------------
static function idbOrdCreate()
//
// idbOrdCreate() --> NIL
//
// Create an index for the active Alias.
//

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_NTX_CREATE) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
    local lOldDeleted           := set( _SET_DELETED, .T. )

    local cFileName := SPACE_STR
    local lExit     := .F.
    local lGoOn     := .T.

    static cTagName
    static cForCondition
    static cOrdBagName
    static cOrdExpr

    if cForCondition == NIL .or.;
        cForCondition = SPACE_STR    // apparentemente vuoto
        cForCondition := padr( ".T.", SPACE_LEN )
    else
        cForCondition := padr( cForCondition, SPACE_LEN )
    end
   
    if cTagName == NIL 
        cTagName := SPACE_STR    
    else
        cTagName := padr( cTagName, SPACE_LEN )
    end
   
    if cOrdBagName == NIL 
        cOrdBagName := padr( "*." + _EXTENTION_INDEXBAG,;
            SPACE_LEN )    
    else
        cOrdBagName := padr( cOrdBagName, SPACE_LEN )
    end

    if cOrdExpr == NIL 
        cOrdExpr := SPACE_STR    
    else
        cOrdExpr := padr( cOrdExpr, SPACE_LEN )
    end

    begin sequence

        if alias() == "" // no alias
            alert( _ERROR_NO_ALIAS )
            break
        end

   
        // Create a kind of window.
        nBottom     := maxrow()
        do case
        case rddname() == "DBFNTX"
            nTop    := nBottom - 8
        case rddname() == "DBFNDX"
            nTop    := nBottom - (8 -2)
        case rddname() == "DBPX"
            nTop    := nBottom - (8 -2)
        case rddname() == "DBFCDX"
            nTop    := nBottom - (8 +2)
        case rddname() == "DBFMDX"
            nTop    := nBottom - (8 +2)
        otherwise
            nTop    := nBottom - 8 +2
        end    
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_NTX_NEW, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_NEW_NTX_FILENAME
            @row()+1,nLeft get cOrdBagName;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cOrdBagName :=;
                    padr( Dir( cOrdBagName ), SPACE_LEN ) } );
                valid true( cOrdBagName :=;
                    padr(strAddExtention( cOrdBagName,;
                    _EXTENTION_INDEXBAG ), SPACE_LEN ) );
                    .and. !( "*" $ cOrdBagName );
                    .and. !( "?" $ cOrdBagName )
            @row()+1,nLeft say IDB_PROMPT_NEW_NTX_KEYEXPR
            @row()+1,nLeft get cOrdExpr;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cOrdExpr :=;
                    padr( rtrim(cOrdExpr) + fieldIndex(),;
                    SPACE_LEN ) } );
                valid ( cOrdExpr <> SPACE_STR )
            if rddname() == "DBFMDX";
                .or. rddname() == "DBFCDX"
                //
                @row()+1,nLeft say IDB_PROMPT_TAG_NAME
                @row()+1,nLeft get cTagName;
                    picture "@s"+ltrim(str(nWidth))+"@";
                    when truesetkey( K_F2, NIL );
                    valid true( iif( cTagName == SPACE_STR,;
                        cTagName := cOrdBagName,;
                        .t. ) )
            end
            if rddname() == "DBFNDX";
                .or. rddname() == "DBPX"
                //
                // there is no FOR condition
            else
                @row()+1,nLeft say IDB_PROMPT_FOR_EXPRESSION
                @row()+1,nLeft get cForCondition;
                    picture "@s"+ltrim(str(nWidth))+"@";
                    when trueSetkey( K_F2, {|| cForCondition :=;
                        padr( rtrim(cForCondition) + fieldNormal(),;
                        SPACE_LEN ) } );
                    valid true( iif( cForCondition == SPACE_STR,;
                        ".t.", cForCondition ) )
            end
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                if file( strAddExtention( cOrdBagName,;
                    _EXTENTION_INDEXBAG ) )
                    //
                    if alert( strAddExtention( cOrdBagName,;
                        _EXTENTION_INDEXBAG ) + ";" +;
                        _ERROR_PROMPT_ALREADY_EXIST_OVERWRITE,;
                        {_MENU_NO, _MENU_YES} ) == 2
                        //
                        lGoOn := .T.
                        exit
                    else
                        // loop
                    end
                else
                    if !( "*" $ cOrdBagName );
                        .and. !( "?" $ cOrdBagName )
                        //
                        lGoOn := .T.
                        exit
                    else
                        // loop
                    end    
                end
            otherwise
                // loop
            end
      
        end
   
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )

        if lGoOn
            
            bOldErrorHandler :=;
                errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                lExit := .T.
                do case
                case rddname() == "DBFCDX"
                    ordCondSet( (cForCondition),;
                        {|| &(cForCondition)},,,; 
                        {|| WaitFileEval()},, RECNO(),,,, .F. )
                    ordCreate((alltrim(cOrdBagName)),;
                        (alltrim(cTagName)),;
                        (alltrim(cOrdExpr)),;
                        {|| &(alltrim(cOrdExpr))}, .F. )
                case rddname() == "DBFMDX"
                    ordCondSet( (cForCondition),;
                        {|| &(cForCondition)},,,; 
                        {|| WaitFileEval()},, RECNO(),,,, .F. ) 
                    ordCreate((alltrim(cOrdBagName)),;
                        (alltrim(cTagName)),;
                        (alltrim(cOrdExpr)),;
                        {|| &(alltrim(cOrdExpr))}, .F. )
                case rddname() == "DBFNTX"
                    ordCondSet( (cForCondition),;
                        {|| &(cForCondition)},,,; 
                        {|| WaitFileEval()},, RECNO(),,,, .F. ) 
                    ordCreate((alltrim(cOrdBagName)),,;
                        (alltrim(cOrdExpr)),;
                        {|| &(alltrim(cOrdExpr))}, .F. )
                case rddname() == "DBFNDX"
                    ordCreate((alltrim(cOrdBagName)),,;
                        (alltrim(cOrdExpr)),;
                        {|| &(alltrim(cOrdExpr))}, .F. )
                end

                dbgotop()
            recover
                set index to
                dbgotop()
                lExit := .F.
            end sequence
            errorblock(bOldErrorHandler)
            WaitProgress(2)
        end

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
    set( _SET_DELETED, lOldDeleted )

    return NIL

//----------------------------------------------------------------
static function idbSetindex()
//
// idbNtxOpen() --> NIL
//
// It Opens an index bag file.
//

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_SET_INDEX) } )
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local cOrdBagName           :=;
        padr( "*." + _EXTENTION_INDEXBAG, SPACE_LEN )
    local lGoOn

    begin sequence

        if alias() == ""
            alert( _ERROR_NO_ALIAS )
            break
        end

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 3
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_OPEN_NTX, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say IDB_PROMPT_OPEN_NTX_FILENAME
            @row()+1,nLeft get cOrdBagName;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cOrdBagName :=;
                    padr( Dir( cOrdBagName ), SPACE_LEN ) } );
                valid file( cOrdBagName :=;
                    padr(strAddExtention(cOrdBagName, _EXTENTION_INDEXBAG),;
                    SPACE_LEN) )
            read
            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                // chack for valid data
                if file( strAddExtention( cOrdBagName,;
                    _EXTENTION_INDEXBAG ) );
                    .and. ( cOrdBagName <> SPACE_STR )
                    //
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end

        end

        restscreen( nTop, nLeft,;
            nBottom, nRight, cOldScreen )

        if lGoOn
            bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                dbsetindex( alltrim( cOrdBagName ) )
            recover
                set index to
            end sequence
            errorblock(bOldErrorHandler)
        end

    end sequence

    // restore
    setcursor( nOldCursor )
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL

//----------------------------------------------------------------
static function idbClearIndex()

    if alias() == ""
        // there is no need
    else
        dbclearindex()
    end

    return NIL

//----------------------------------------------------------------
static function idbReindex()

    local bOldErrorHandler

    do case
    case ( alias() == "" )  // no alias
        alert( _ERROR_NO_ALIAS )
    case ordbagname( ordsetfocus() ) == ""
        alert( _ERROR_NO_INDEX )
    otherwise
        // Try to flock, but it works only on NTX files.
        // With other index bag types an exclusive,
        // open is required.
        if !flock()
            return NIL
        end
      
        bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
        begin sequence
         
            reindex eval WaitFileEval()
      
        recover
            alert( _ERROR_EXCLUSIVE_REQUIRED )
        end sequence
        errorblock(bOldErrorHandler)
        dbunlock()

    end

    return NIL

//----------------------------------------------------------------
static function idbRelation()
//
// idbRelation() --> NIL
//
// Establish a relation in a guided form.
//

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local cOldSelect            := select()
    local bOld_F1               :=;
        setkey( K_F1, { || Text(IDB_HELP_RELATION_DEFINITION) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0

    local cChildName := space(10)
    local lGoOn := .F.

    static cRelExpression

    if cRelExpression == NIL
        cRelExpression := SPACE_STR
    end
   
    begin sequence

        if alias() == ""
            alert( _ERROR_NO_ALIAS )
            break
        end

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 6
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( IDB_DIALOG_BOX_TOP_RELATION_DEFINITION, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft;
                say IDB_PROMPT_RELATION_DEFINITION_CHILDNAME
            @row()+1,nLeft get cChildName;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when truesetkey( K_F2, {|| cChildName :=;
                    padr( selectAlias(), 10 ) } );
                valid ( .T. )
            @row()+1,nLeft say IDB_PROMPT_RELATION_DEFINITION_EXPRESSION
            @row()+1,nLeft get cRelExpression;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when truesetkey( K_F2, {|| cRelExpression :=;
                    padr( rtrim(cRelExpression) + fieldRelation(),;
                    SPACE_LEN ) } );
                valid ( .T. )
            read

            do case
            case lastkey() = K_ESC
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN
                lGoOn := .T.
                exit
            otherwise
                // loop
            end
      
        end
   
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )

        if lGoOn
            
            bOldErrorHandler :=;
                errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                dbSetRelation( alltrim(cChildName),;
                    {|| &(alltrim(cRelExpression))},;
                    (alltrim(cRelExpression)) )
            recover
                alert( IDB_ERROR_RELATION )
            end sequence
            errorblock(bOldErrorHandler)
      
        end
   
    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL

//----------------------------------------------------------------
// FRM - FORM REPORT CREATION AND EDITING
//----------------------------------------------------------------
static function frmPrint( cReportFile, bWhileCondition, bForCondition )
//
// frmPrint( [<cReportFile>], [<bWhileCondition>],
//      [<bForCondition>] ) --> NIL
//
// This function asks for informations when not given.
//
// <cReportFile>        the report (form) file name
//
// <bWhileCondition>    the WHILE code block condition
//
// <bForCondition>      the FOR code block condition
//


    local getlist   := {}
    local bOldErrorHandler
    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local bOld_F1       := setkey( K_F1, { || Text( FRM_HELP_REPORT_FORM ) } )
    local bOld_F2       := setkey( K_F2, NIL )
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
    
    local lGoPrint := .T.
    local cWhileCondition := padr( ".T.", SPACE_LEN )
    local cForCondition   := padr( ".T.", SPACE_LEN )
   
    local lReportFile     := .F.
    local lWhileCondition := .F. 
    local lForCondition   := .F.

    begin sequence
   
        if ( alias() == "" ) // no alias
            alert( _ERROR_NO_ALIAS )
            break
        end
   
        // analisi dei parametri forniti   
        do case 
        case valtype(cReportFile) == "C";
            .and. !file( cReportFile :=;
            padr(strAddExtention( cReportFile,;
            _EXTENTION_FORM ), SPACE_LEN) )
            //   
            lReportFile := .T.
        case valtype(cReportFile) <> "C" 
            lReportFile := .T.
            cReportFile :=;
                padr( "*." + _EXTENTION_FORM, SPACE_LEN )
        end

        do case
        case valtype(bWhileCondition) <> "B" 
            lWhileCondition := .T.
            cWhileCondition := padr( ".T.", SPACE_LEN )
        end
      
        do case
        case valtype(bForCondition) <> "B" 
            lForCondition := .T.
            cForCondition := padr( ".T.", SPACE_LEN )
        end
      
        // check for parameters
        if lReportFile .or.;
           lWhileCondition .or.;
           lForCondition
            
            // Create a kind of window.
            nBottom     := maxrow()
            nTop        := nBottom - 8
            nLeft       := 0
            nRight      := maxcol()
            nWidth      := nRight - nLeft +1
            cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
            setcolor( COLOR_DEFAULT1 )
            scroll( nTop, nLeft, nTop, nRight )
            scroll( nBottom-1, nLeft, nBottom, nRight )
            devpos( nTop, nLeft )
            devout( padc( FRM_DIALOG_BOX_TOP_REPORT, nWidth ) )
            devpos( nBottom-1, nLeft )
            devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
            devpos( nBottom, nLeft )
            devout( padc( _KEY_F2_PGDN, nWidth ) )

            while .t.
                setcolor( COLOR_DEFAULT2 )
                scroll( nTop+1, nLeft, nBottom-2, nRight )
                setpos( nTop, nLeft )
                if lReportFile
                    @row()+1,nLeft;
                        say FRM_PROMPT_OPEN_FILE_REPORT
                    @row()+1,nLeft;
                        get cReportFile;
                        picture "@s"+ltrim(str(nWidth))+"@";    
                        when trueSetkey( K_F2, {|| cReportFile :=;
                            padr( Dir( cReportFile ),;
                            SPACE_LEN ) } );      
                        valid ( file( cReportFile :=;
                            padr(strAddExtention( cReportFile,;
                            _EXTENTION_FORM ), SPACE_LEN) ) );
                            .and. !( "*" $ cReportFile );
                            .and. !( "?" $ cReportFile )  
                end
                if lWhileCondition
                    @row()+1,nLeft;
                        say FRM_PROMPT_WHILE_EXPRESSION  
                    @row()+1,nLeft;
                        get cWhileCondition;
                        picture "@s"+ltrim(str(nWidth))+"@";
                        when trueSetkey( K_F2,;
                            {|| cWhileCondition :=;
                            padr( fieldNormal(),;
                            SPACE_LEN ) } );     
                        valid ( ( cWhileCondition :=;
                            iif( cWhileCondition <> SPACE_STR,;
                            cWhileCondition,;
                            padr( ".T.", SPACE_LEN ) ) ) <> NIL )
                end
                if lForCondition
                    @row()+1,nLeft;
                        say FRM_PROMPT_FOR_EXPRESSION
                    @row()+1,nLeft;
                        get cForCondition;
                        picture "@s"+ltrim(str(nWidth))+"@";
                        when trueSetkey( K_F2,;
                            {|| cForCondition :=;
                            padr( fieldNormal(),;
                            SPACE_LEN ) } );     
                        valid ( ( cForCondition :=;
                            iif( cForCondition <> SPACE_STR,;
                            cForCondition, padr( ".T.",;
                            SPACE_LEN ) ) ) <> NIL )
                end
                read

                do case
                case lastkey() = K_ESC  // exit
                    lGoPrint := .F.
                    exit
                case lastkey() = K_PGDN // confirm
                    // check for correct data
                    if file( strAddExtention( cReportFile,;
                        _EXTENTION_FORM ) );
                        .and. cWhileCondition <> SPACE_STR;
                        .and. cForCondition <> SPACE_STR
                        //
                        lGoPrint := .T.
                        exit
                    else
                        // loop
                    end
                    // transform character conditions
                    if lWhileCondition
                        bWhileCondition := { || &cWhileCondition }
                    end
                    if lForCondition
                        bForCondition := { || &(cForCondition) .and. WaitFileEval() }
                    end
                otherwise
                    // loop
                end
            end
   
            // close window
            restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        end

        if lGoPrint
            bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                __ReportForm( alltrim(cReportFile), .f., ,;
                    .f., bForCondition, bWhileCondition,; 
                    , , .f.,; 
                    .f., , .t. , .f. )
                dbgotop() // otherwise it stays on eof()
            recover
                dbgotop()
            end sequence
            errorblock(bOldErrorHandler)
            WaitProgress(2)
        end

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL

//----------------------------------------------------------------
static function frmModify()
//
// frmModify() --> NIL
//
// It executes the frm() function asking for the file name to
// modify.
//   

    local getlist   := {}
    local bOldErrorHandler
    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local bOld_F1       := setkey( K_F1, { || Text( FRM_HELP_REPORT_EDIT ) } )
    local bOld_F2       := setkey( K_F2, NIL )
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
                  
    local lGoOn         := .F.
    local cReportFile   :=;
        padr( "*." + _EXTENTION_FORM, SPACE_LEN )

    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 4
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( FRM_DIALOG_BOX_TOP_REPORT, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )
      
        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say FRM_PROMPT_OPEN_FILE_REPORT
            @row()+1,nLeft get cReportFile;
                picture "@s"+ltrim(str(nWidth))+"@";    
                when trueSetkey( K_F2, {|| cReportFile :=;
                    padr( Dir( cReportFile ), SPACE_LEN ) } );
                valid ( file( cReportFile :=;
                    padr(strAddExtention( cReportFile,;
                    _EXTENTION_FORM ), SPACE_LEN) ) )  
            read

            do case
            case lastkey() = K_ESC  // exit
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN // ok
                if file( strAddExtention( cReportFile,;
                    _EXTENTION_FORM ) )
                    //
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
        end
   
        // close window
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        if lGoOn
            frm( cReportFile )
        end
   
    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL

//----------------------------------------------------------------
static function frm( cFileName )
//
// frm( <cFileName> ) --> NIL
//
// <cFileName>  the report file name to modify.
//              If not present, it will be created.
//
// Note sugli offset:
//    quando vengono usati puntatori a variabili di clipper
//    occorre tenere presente che i puntatori di queste
//    si riferiscono allo standard utilizzato da clipper,
//    cio il primo elemento viene indirizzato con il
//    numero 1, mentre quando si fa riferimento ad indirizzi
//    scritti all'interno del file di Report Form, questi
//    sono scritti utilizzando il sistema per cui il primo
//    elemento ha l'indirizzo 0.
//    Quando si parler di "DOS offsets" si far riferimento
//    a quest'ultimo caso.
//
// Note originali del programma "RL".
//    Report file length is 7C6h (1990d) bytes.
//    Expression length array starts at 04h (4d) and can
//       contain upto 55 short (2 byte) numbers.
//    Expression offset index array starts at 72h (114d) and
//       can contain upto 55 short (2 byte) numbers.
//    Expression area starts at offset E0h (224d).
//    Expression area length is 5A0h (1440d).
//    Expressions in expression area are null terminated.
//    Field expression area starts at offset 680h (1664d).
//    Field expressions (column definition) are null terminated.
//    Field expression area can contain upto 25 12-byte blocks.
//
   
    local getlist   := {}
    local bOldErrorHandler
    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local bOld_F1       := setkey( K_F1, { || Text( FRM_HELP_REPORT_EDIT ) } )
    local bOld_F2       := setkey( K_F2, NIL )
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
    local lOldDeleted   := set( _SET_DELETED, .T. )
    
    local axFormVal[16]
    local nLastSelect   := select() // save last select
    local aStruct := {}

    local cOldName
    local lSave
    local lGoOn

    begin sequence

        // Viene inizializzato il vettore dei valori e delle intestazioni      
        axFormVal[FRM_PAGE_WIDTH]     := 80
        axFormVal[FRM_LINES_PAGE]     := 58
        axFormVal[FRM_LEFT_MARG]      := 8
        axFormVal[FRM_RIGHT_MARG]     := 0
        axFormVal[FRM_COL_COUNT]      := 0
        axFormVal[FRM_DBL_SPACED]     := "N"
        axFormVal[FRM_SUMMARY]        := "N"
        axFormVal[FRM_PE]             := "N"
        axFormVal[FRM_PLAINPAGE]      := "N"
        axFormVal[FRM_PEAP]           := "Y"
        axFormVal[FRM_PEBP]           := "N"
        axFormVal[FRM_PAGE_HDR]       := space(240)
        axFormVal[FRM_GRP_EXPR]       := space(200)
        axFormVal[FRM_SUB_EXPR]       := space(200)
        axFormVal[FRM_GRP_HDR]        := space(50)
        axFormVal[FRM_SUB_HDR]        := space(50)
   
        // creazione del file di struttura del corpo del report form
        aadd( aStruct, { "ORDER", "N", 6, 0 } )
        aadd( aStruct, { "WIDTH", "N", 3, 0 } )
        aadd( aStruct, { "TOTALS", "C", 1, 0 } )
        aadd( aStruct, { "DECIMALS", "N", 2, 0 } )
        aadd( aStruct, { "CONTENTS", "C", 254, 0 } )    // 254
        aadd( aStruct, { "HEADER", "C", 254, 0 } )      // 260
   
        dbcreate( strTempPath()+"\TEMPFORM", aStruct )
        dbusearea( .T., NIL, strTempPath()+"\TEMPFORM", "FormStru" )
        index on Field->Order to &(strTempPath()+"\TEMPFORM")
   
        if valtype( cFileName ) <> "C"
            cFileName :=;
                strAddExtention( _UNTITLED, _EXTENTION_FORM )
                //
            cOldName := NIL
            // verr creato
        else
            if file( cFileName )
                cOldName := alltrim(cFileName)
                // viene caricato
                frmLoad( cFileName, @axFormVal )
            else
                cOldName := NIL
                // verr creato
            end
        end

        lSave := frmEdit( cFileName, @axFormVal, nLastSelect )

        if lSave
         
            cFileName := padr( cFileName, SPACE_LEN )

            // Create a kind of window.
            nBottom     := maxrow()
            nTop        := nBottom - 4
            nLeft       := 0
            nRight      := maxcol()
            nWidth      := nRight - nLeft +1
            cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
            setcolor( COLOR_DEFAULT1 )
            scroll( nTop, nLeft, nTop, nRight )
            scroll( nBottom-1, nLeft, nBottom, nRight )
            devpos( nTop, nLeft )
            devout( padc( FRM_DIALOG_BOX_TOP_REPORT, nWidth ) )
            devpos( nBottom-1, nLeft )
            devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
            devpos( nBottom, nLeft )
            devout( padc( _KEY_F2_PGDN, nWidth ) )
      
            while .t.
                setcolor( COLOR_DEFAULT2 )
                scroll( nTop+1, nLeft, nBottom-2, nRight )
                setpos( nTop, nLeft )
                @row()+1,nLeft say FRM_PROMPT_SAVE_REPORT
                @row()+1,nLeft get cFileName;
                    picture "@s"+ltrim(str(nWidth))+"@";    
                    when trueSetkey( K_F2, {|| cFileName :=;
                        padr( Dir( cFileName ), SPACE_LEN ) } );      
                    valid ( ( cFileName :=;
                        padr(strAddExtention(cFileName,;
                        _EXTENTION_FORM), SPACE_LEN) ) <> NIL )    
                read

                do case
                case lastkey() = K_ESC  // exit
                    lGoOn := .F.
                    exit
                case lastkey() = K_PGDN // ok
                    do case
                    case alltrim(cFileName) <> cOldName;
                        .and. file( strAddExtention( cFileName,;
                        _EXTENTION_FORM ) );
                        .and. ( alert( alltrim(cFileName) + ";" +;
                        _ERROR_PROMPT_ALREADY_EXIST_OVERWRITE,;
                        { _MENU_NO, _MENU_YES } ) == 1 )
                        //
                        // loop
                    case strCutExtention( cFileName ) == ""
                        // loop
               
                    otherwise
                        lGoOn := .T.
                        exit
                    end
            
                otherwise
                    // loop
                end
            end
   
            if lGoOn
                // ok save
                if FrmSave( alltrim(cFileName), axFormVal )
                    // ok
                else
                    alert( FRM_ERROR_FILE_NOT_SAVED )
                end
            else
                // nix
            end

        else
            // nix
        end

        // chiude il file di struttura del Report Form
        close FormStru

        // viene riselezionato il file precedentemente aperto
        dbselectarea( nLastSelect )

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
    set( _SET_DELETED, lOldDeleted )

    return NIL

//----------------------------------------------------------------
static function frmLoad( cFileName, axFormVal )
                  
    local cBuffer      := space( FRM_SIZE_FILE_BUFF )
    local acFormBuffer[5]
    local nPlusByte    := 0
    local nPointer     := 0
    local nHandle      := fopen( cFileName )
    local nByteRead    := 0
    local nFieldOffset := 0
    local nI           := 0
    local lFileError   := .F.

    if !ferror() == 0
        alert( cFileName + ";" +;
            FRM_ERROR_CANNOT_OPEN + str(ferror()) )
        return .F.
    end

    // legge il file
    nByteRead := fread( nHandle, @cBuffer, FRM_SIZE_FILE_BUFF )
   
    // controlla la validit apparente del file
    if nByteRead < FRM_SIZE_FILE_BUFF 
        alert( cFileName + ";" +;
            FRM_ERROR_FILE_TOO_LITTLE )
        return .F.
    elseif !( bin2i( substr( cBuffer, 1, 2 ) ) == 2 .and.;
              bin2i( substr( cBuffer, FRM_SIZE_FILE_BUFF - 1, 2 ) ) == 2 )
        // se non inizia e finisce con il numero 2 non  un report form
        alert( cFileName + ";" +;
            FRM_ERROR_FILE_NOT_VALID )
        return .F.
    end

    Waitfor( FRM_WAIT_LOADING )
   
    // estrae i valori dal file appena caricato
    acFormBuffer[ FRM_LENGTHS_BUFF ] := ;
        substr( cBuffer, FRM_LENGTHS_OFFSET, FRM_SIZE_LENGTHS_BUFF )
    acFormBuffer[ FRM_OFFSETS_BUFF ] := ;
        substr( cBuffer, FRM_OFFSETS_OFFSET, FRM_SIZE_OFFSETS_BUFF )
    acFormBuffer[ FRM_EXPR_BUFF ] := ;
        substr( cBuffer, FRM_EXPR_OFFSET, FRM_SIZE_EXPR_BUFF )
    acFormBuffer[ FRM_FIELDS_BUFF ] := ;
        substr( cBuffer, FRM_FIELDS_OFFSET, FRM_SIZE_FIELDS_BUFF )
    acFormBuffer[ FRM_PARAMS_BUFF ] := ;
        substr( cBuffer, FRM_PARAMS_OFFSET, FRM_SIZE_PARAMS_BUFF )

    axFormVal[FRM_PAGE_WIDTH] := ;
        bin2i( substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_PAGE_WIDTH_OFFSET, 2 ) )
    axFormVal[FRM_LINES_PAGE] := ;
        bin2i( substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_LNS_PER_PAGE_OFFSET, 2 ) )
    axFormVal[FRM_LEFT_MARG] := ;
        bin2i( substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_LEFT_MRGN_OFFSET, 2 ) )
    axFormVal[FRM_RIGHT_MARG] := ;
        bin2i( substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_RIGHT_MRGN_OFFSET, 2 ) )
    axFormVal[FRM_COL_COUNT] := ;
        bin2i( substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_COL_COUNT_OFFSET, 2 ) )

    axFormVal[FRM_DBL_SPACED] := ;
        substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_DBL_SPACE_OFFSET, 1 )
    axFormVal[FRM_SUMMARY] := ;
        substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_SUMMARY_RPT_OFFSET, 1 )
    axFormVal[FRM_PE] := ;
        substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_PE_OFFSET, 1 )

    nPlusByte := ;
        asc( substr( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_PLNPG_PEAP_PEBP_OFFSET, 1) )
    if int( nPlusByte/4 ) == 1
        axFormVal[FRM_PLAINPAGE] := "Y"
        nPlusByte -= 4
    end
    if int( nPlusByte/2 ) == 1
        axFormVal[FRM_PEAP] := "Y"
        nPlusByte -= 2
    end
    if int( nPlusByte/1 ) == 1
        axFormVal[FRM_PEBP] := "N"
        nPlusByte -= 1
    end

    // estrazioni di espressioni (stringhe) puntate tramite puntatori
   
    // Page Header
    nPointer := ;
        bin2i( substr ( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_PAGE_HDR_OFFSET, 2 ) )
    axFormVal[FRM_PAGE_HDR] := ;
        padr( frmGetExpr( @acFormBuffer, nPointer ), 240 )
    // Grouping Expression
    nPointer := ;
        bin2i( substr ( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_GRP_EXPR_OFFSET, 2 ) )
    axFormVal[FRM_GRP_EXPR] := ;
        padr( frmGetExpr( @acFormBuffer, nPointer ), 200 )
    // Sub-grouping Expression
    nPointer := ;
        bin2i( substr ( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_SUB_EXPR_OFFSET, 2 ) )
    axFormVal[FRM_SUB_EXPR] := ;
        padr( frmGetExpr( @acFormBuffer, nPointer ), 200 )
    // Group Header
    nPointer := ;
        bin2i( substr ( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_GRP_HDR_OFFSET, 2 ) )
    axFormVal[FRM_GRP_HDR] := ;
        padr( frmGetExpr( @acFormBuffer, nPointer ), 50 )
    // Sub-group header
    nPointer := ;
        bin2i( substr ( acFormBuffer[ FRM_PARAMS_BUFF ], FRM_SUB_HDR_OFFSET, 2 ) )
    axFormVal[FRM_SUB_HDR] := ;
        padr( frmGetExpr( @acFormBuffer, nPointer ), 50 )

    // estrazione dei campi in un file di passaggio
    // che conterr tutti i dati necessari.

    // trasferimento campi
    nFieldOffset := 12
    for nI = 1 to axFormVal[FRM_COL_COUNT]
        nFieldOffset := ;
         frmGetField( @acFormBuffer, nFieldOffset, @lFileError )
    next

    WaitFor()

    return !lFileError

//----------------------------------------------------------------
static function frmGetExpr( acFormBuffer, nPointer )
//
// The expression is empty if:
//    - passed pointer is equal to 65535;
//    - character following character pointed to by
//      pointer is CHR(0) (NULL).
//
   
    local nExprOffset   := 0
    local nExprLength   := 0
    local nOffsetOffset := 0
    local cString       := ""

    if nPointer != 65535
        // Convert DOS FILE offset to CLIPPER string offset
        nPointer += 1
        // Calculate offset into OFFSETS_BUFF
        if nPointer > 1
           nOffsetOffset = ( nPointer * 2 ) - 1
        end

        nExprOffset := ;
            bin2i( substr( acFormBuffer[ FRM_OFFSETS_BUFF ], nOffsetOffset, 2) )
        nExprLength := ;
            bin2i( substr( acFormBuffer[ FRM_LENGTHS_BUFF ], nOffsetOffset, 2) )

        // EXPR_OFFSET points to a NULL, so add one (+1) to get the string
        // and subtract one (-1) from EXPR_LENGTH for correct length
        nExprOffset += 1
        nExprLength -= 1

        // Extract string
        cString := substr( acFormBuff[ FRM_EXPR_BUFF ], nExprOffset, nExprLength )

        // dBASE does this so we must do it too
        // Character following character pointed to by pointer is NULL
        if chr(0) == substr( cString, 1, 1) .and.; 
              len( substr( cString, 1, 1 ) ) == 1
            cString = ""
        end
    end

    return cString

//----------------------------------------------------------------
static function frmGetField( acFormBuffer, nFieldOffset, lFileError )
//
// The Header or Contents expressions are empty if:
//    - passed pointer is equal to 65535;
//    - character following character pointed to by
//      pointer is CHR(0) (NULL).
//

    local nPointer := 0
    local nNumber  := 0

    // aggiunge un record vuoto
    append blank

    // Field sequence number
    FormStru->Order := nFieldOffset * 50
    // Column width
    FormStru->Width := ;
        bin2i( substr( acFormBuffer[ FRM_FIELDS_BUFF ], nFieldOffset + FRM_FIELD_WIDTH_OFFSET, 2) )
    // Total
    FormStru->Totals := substr( acFormBuffer[ FRM_FIELDS_BUFF ], nFieldOffset + FRM_FIELD_TOTALS_OFFSET, 1)
    // Decimals width
    FormStru->Decimals := ;
       bin2i( substr( acFormBuffer[ FRM_FIELDS_BUFF ], nFieldOffset +;
           FRM_FIELD_DECIMALS_OFFSET, 2 ) )
    // Content expression
    nPointer := bin2i( substr( acFormBuffer[ FRM_FIELDS_BUFF ], nFieldOffset +;
                  FRM_FIELD_CONTENT_EXPR_OFFSET, 2 ) )
    FormStru->Contents := frmGetExpr( @acFormBuffer, nPointer )
    // Header expression
    nPointer := bin2i( substr( acFormBuffer[ FRM_FIELDS_BUFF ], nFieldOffset +;
                  FRM_FIELD_HEADER_EXPR_OFFSET, 2))
    FormStru->Header := frmGetExpr( @acFormBuffer, nPointer )

    return nFieldOffset + 12

//----------------------------------------------------------------
static function frmEdit( cFileName, axFormVal, nLastSelect )
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local bOld_F1               := setkey( K_F1, NIL ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
   
    local cLastAlias := alias(nLastSelect)
    local lSave  := .F.

    local acCol
    local acColSayPic
    local acColHead
    local alColCalc
    local acColValid
    local acColMsg
    local acColSep
   
    begin sequence
   
        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 21
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( FRM_DIALOG_BOX_TOP_REPORT, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_ESC_EXIT_F1_F2_PGDN, nWidth ) )
      
        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-1, nRight )

            setkey( K_F1, { || Text( FRM_HELP_REPORT_EDIT_HEAD)} )
            setpos( nTop, nLeft )
            @row()+1,nLeft say FRM_PROMPT_REPORT_PAGE_WIDTH;
                get axFormVal[ FRM_PAGE_WIDTH ]  picture "999";     
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_LINES_PER_PAGE;
                get axFormVal[ FRM_LINES_PAGE ]  picture "999";
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_LEFT_MARGIN;
                get axFormVal[ FRM_LEFT_MARG ]   picture "999";   
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_RIGHT_MARGIN;
                get axFormVal[ FRM_RIGHT_MARG ]  picture "999";   
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_DOUBLE_SPACED;  
                get axFormVal[ FRM_DBL_SPACED ]  picture "!";
                when trueSetkey( K_F2, NIL );
                valid axFormVal[FRM_DBL_SPACED ] = "Y" .or.;
                    axFormVal[FRM_DBL_SPACED ] = "N"
            @row()+1,nLeft say FRM_PROMPT_REPORT_PAGE_EJECT_BEFORE;
                get axFormVal[ FRM_PEBP ]  picture "!";  
                when trueSetkey( K_F2, NIL );
                valid axFormVal[FRM_PEBP ] = "Y" .or.;
                      axFormVal[FRM_PEBP ] = "N"
            @row()+1,nLeft say FRM_PROMPT_REPORT_PAGE_EJECT_AFTER;
                get axFormVal[ FRM_PEAP ]  picture "!";  
                when trueSetkey( K_F2, NIL );
                valid axFormVal[FRM_PEAP ] = "Y" .or.;
                      axFormVal[FRM_PEAP ] = "N"
            @row()+1,nLeft say FRM_PROMPT_REPORT_PLAIN_PAGE;
                get axFormVal[ FRM_PLAINPAGE ]  picture "!";  
                when trueSetkey( K_F2, NIL );
                valid axFormVal[FRM_PLAINPAGE ] = "Y" .or.;
                    axFormVal[FRM_PLAINPAGE ] = "N"
            @row()+1,nLeft say FRM_PROMPT_REPORT_PAGE_HEADER
            @row()+1,nLeft get axFormVal[ FRM_PAGE_HDR ]  picture "@s"+ltrim(str(nWidth))+"@";
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_GROUP_HEADER
            @row()+1,nLeft get axFormVal[ FRM_GRP_HDR ]   picture "@s"+ltrim(str(nWidth))+"@";
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_GROUP_EXPRESSION
            @row()+1,nLeft get axFormVal[ FRM_GRP_EXPR ]  picture "@s"+ltrim(str(nWidth))+"@";  
                when trueSetkey( K_F2, { || (cLastAlias)->( keyboard( fieldNormal() ) ) } )
            @row()+1,nLeft say FRM_PROMPT_REPORT_SUMMARY_REPORT_ONLY;
                get axFormVal[ FRM_SUMMARY ]  picture "!";  
                when trueSetkey( K_F2, NIL );
                valid axFormVal[FRM_SUMMARY ] = "Y" .or.;
                      axFormVal[FRM_SUMMARY ] = "N"
            @row()+1,nLeft say FRM_PROMPT_REPORT_PAGE_EJECT_AFTER_GROUP;  
                get axFormVal[ FRM_PE ]  picture "!";  
                when trueSetkey( K_F2, NIL );
                valid axFormVal[FRM_PE ] = "Y" .or.;
                      axFormVal[FRM_PE ] = "N"
            @row()+1,nLeft say FRM_PROMPT_REPORT_SUB_GROUP_HEADER
            @row()+1,nLeft get axFormVal[ FRM_SUB_HDR ]  picture "@s"+ltrim(str(nWidth))+"@";  
                when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say FRM_PROMPT_REPORT_SUB_GROUP_EXPRESSION
            @row()+1,nLeft get axFormVal[ FRM_SUB_EXPR ]  picture "@s"+ltrim(str(nWidth))+"@";  
                when trueSetkey( K_F2, { || (cLastAlias)->( keyboard( fieldNormal() ) ) } )
            read
         
            do case
            case lastkey() = K_ESC
                lSave := .F.
            exit
            case lastkey() = K_PGDN // conferma la scelta
                // close last and open next window
                restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
                nBottom     := maxrow()
                nTop        := nBottom - 22
                nLeft       := 0
                nRight      := maxcol()
                nWidth      := nRight - nLeft +1
                cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
                setcolor( COLOR_DEFAULT1 )
                scroll( nTop, nLeft, nTop, nRight )
                scroll( nBottom-1, nLeft, nBottom, nRight )
                devpos( nTop, nLeft )
                devout( padc( FRM_DIALOG_BOX_TOP_REPORT, nWidth ) )
                devpos( nBottom, nLeft )
                devout( padc( _KEY_ESC_EXIT_F1_F2_CTRLY, nWidth ) )
                setcolor( COLOR_DEFAULT0 )
                scroll( nTop+1, nLeft, nBottom-1, nRight )

                setkey( K_F2, { || (cLastAlias)->( keyboard( fieldNormal() ) ) } )

                dbgotop() // si posiziona sulla prima colonna
                setkey( K_F1, { || Text( FRM_HELP_REPORT_EDIT_COLUMNS ) } ) 

                acCol := { "if( deleted(), '*', ' ' )",;
                           "Order",; 
                           "Header",; 
                           "Contents",; 
                           "Width",; 
                           "Decimals",; 
                           "Totals" }
                acColSayPic := { "X",; 
                                 "999999",; 
                                 "@s25",; 
                                 "@s25",; 
                                 "999",; 
                                 "99",; 
                                 "!" }
                acColHead := { "*",; 
                                FRM_HEAD_REPORT_ORDER,; 
                                FRM_HEAD_REPORT_HEADER,; 
                                FRM_HEAD_REPORT_CONTENT,; 
                                FRM_HEAD_REPORT_WIDTH,; 
                                FRM_HEAD_REPORT_DECIMALS,; 
                                FRM_HEAD_REPORT_TOTAL}
                alColCalc := { .T.,; 
                               .F.,; 
                               .F.,; 
                               .F.,; 
                               .F.,; 
                               .F.,; 
                               .F. }
            
                /*
                TB( <nTop>, <nLeft>, <nBottom>, <nRight>,
                     <acCol>, <acColSayPic>, [<acColHead>], 
                     [<alColCalc>, [<nFreeze>],
                     [<lModify>], [<lAppend>], [<lDelete>],
                     [<lAutosort>] )  --> NIL
                */
                TB( nTop+1, nLeft, nBottom-1, nRight,;
                     acCol, acColSayPic, acColHead,; 
                     alColCalc, 1,;
                     .t., .t., .t., .t.)
            
                // close window
                restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )

                lSave := .T.
                exit
      
            otherwise
                // nix
            end
            // loop
        end
   
   
    end sequence
    
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    // termina la procedura
    return lSave

//---------------------------------------------------------------
static function frmSave( cFileName, axFormVal )

    local nHandle   := 0
    local nByteWritten := 0
    local cBuffer      := ""
   
    local nLastExpr    := 0  // valore iniziale
    local nExprCount   := -1 // valore iniziale
    local nLengthOffset := FRM_LENGTHS_OFFSET - 1 // DOS offsets
    local nOffsetsOffset := FRM_OFFSETS_OFFSET - 1
    local nFieldsOffset := FRM_FIELDS_OFFSET - 1
    local nLengthsOffset := FRM_LENGTHS_OFFSET - 1

    local anNum[5]

    local i := 0
    local j := 0

    local lOk := .F.

    nHandle := fcreate( cFileName )

    if !ferror() == 0
        alert( cFileName + ";" +;
            FRM_ERROR_CANNOT_CREATE + str(ferror()) )
        return .F.
    end

    // scrive lo scheletro del Report Form
    cBuffer := chr(2) + chr(0) +; 
                 replicate( chr(0), ( FRM_SIZE_FILE_BUFF - 4 ) ) +;
                 chr(2) + chr(0)
    nByteWritten = fwrite( nHandle, cBuffer, FRM_SIZE_FILE_BUFF )

    if nByteWritten != FRM_SIZE_FILE_BUFF
        alert( FRM_ERROR_SOMETHING_GONE_WRONG )
        return .F.
    end
   
    // Write Page Heading info
    if ( anNum[ FRM_PAGE_HDR_NUM ] :=; 
         frmWriteExpr( nHandle, axFormVal[ FRM_PAGE_HDR ], .T.,; 
            @nLastExpr, @nExprCount,; 
            @nLengthsOffset, @nOffsetsOffset ) ) = -1
        return .F.
    end

    // Write Grouping expression info
    if ( anNum[ FRM_GRP_EXPR_NUM ] :=; 
         frmWriteExpr( nHandle, axFormVal[ FRM_GRP_EXPR ], .T.,; 
            @nLastExpr, @nExprCount,; 
            @nLengthsOffset, @nOffsetsOffset ) ) = -1
        return .F.
    end
         
    if ( anNum[ FRM_SUB_EXPR_NUM ] :=;
         frmWriteExpr( nHandle, axFormVal[ FRM_SUB_EXPR ], .T.,; 
            @nLastExpr, @nExprCount,; 
            @nLengthsOffset, @nOffsetsOffset ) ) = -1
        return .F.
    end
            
    if ( anNum[ FRM_GRP_HDR_NUM ] :=;
         frmWriteExpr( nHandle, axFormVal[ FRM_GRP_HDR ], .T.,; 
            @nLastExpr, @nExprCount,; 
            @nLengthsOffset, @nOffsetsOffset ) ) = -1
        return .F.
    end

    if ( anNum[ FRM_SUB_HDR_NUM ] :=;
         frmWriteExpr( nHandle, axFormVal[ FRM_SUB_HDR ], .T.,; 
            @nLastExpr, @nExprCount,; 
            @nLengthsOffset, @nOffsetsOffset ) ) = -1
        return .F.
    end

    // se siamo qui ok
    lOk := .T.
   
    // in questo momento  aperto il file di
    // struttura del Report Form
    dbgotop()
    j := lastrec()
    axFormVal[ FRM_COL_COUNT ] := j
    for i = 1 to j
        if i == j .and.; 
           len( trim( Field->Contents ) ) == 0
            lOk := .T.
            axFormVal[ FRM_COL_COUNT ]-- // decrementa
        else 
            if !frmWriteField( nHandle, @nFieldsOffset, @nLengthsOffsets, @nOffsetsOffset, @nLastExpr, @nExprCount ) 
                lOk := .F.
                exit // se non ha funzionato un inserimento di colonna
            end
        end
        // se non  ancora uscito
        dbskip() 
    next
                  
    // Column info written ok?
    if lOk
        // Write last 24 bytes of report and update next_free_offset
        lOk := frmWriteParams( nHandle, anNum, axFormVal, nLastExpr )
    end
   
    // close 
    if !fclose( nHandle )
        lOk := .F.
    end

    return lOk

//---------------------------------------------------------------
static function frmWriteExpr( nHandle, cString, lBlank, nLastExpr,; 
                            nExprCount, nLengthsOffset,; 
                            nOffsetsOffset )
//
// cString    stringa da scrivere nella area dell'espressione
//
// lBlank     test for dBASE like blank expression handling
//            and return a 65535 if expression to write is blank.
//
// -->
//    numeric, expression count (0 to 55 inclusive) or
//    65535 (if blank = .T. and EMPTY(string) = .T.) or
//    -1 (if WRITE/SEEK error).
//

    local lStatus     := .F.
    local cWriteItem  := ""
    local nWriteLen   := 0
    local nWriteCount := 0
    local nReturnCount := 0        // expression count/65535 if empty/-1 error.

    // For dBASE compatability
    if lBlank .and. len(cString) = 0
        // espressione vuota
        return 65535 
    end

    cWriteItem := cString + chr(0)
    nWriteLen  := len(cWriteItem)
   
    // Move to the next free area (DOS offsets)
    fseek( nHandle, FRM_EXPR_OFFSET - 1 + nLastExpr )
    if ferror() <> 0
        return -1
    end

    // Write the expression 
    nWriteCount := fwrite( nHandle, cWriteItem, nWriteLen )
    if nWriteCount == 0 .or. ferror() <> 0
        return -1
    end
         
    fseek( nHandle, nOffsetsOffset)
    if ferror() <> 0
        return -1
    end

    // Add an offset to the offsets array. 
    nWriteCount := fwrite( nHandle, i2Bin( nLastExpr ), 2 )
    if nWriteCount == 0 .or. ferror() <> 0
        return -1
    end

    fseek( nHandle, nLengthsOffset )
    if ferror() <> 0
        return -1
    end

    // Add the expression length to the lengths array 
    nWriteCount := fwrite( nHandle, i2bin( nWriteLen ), 2)
    if nWriteCount == 0 .or. ferror() <> 0
        return -1
    end

    // Move offsets to next position
    nLastExpr += nWriteLen
    nLengthsOffset += 2                  
    nOffsetsOffset += 2

    nExprCount += 1         // global increment.

    return nExprCount

//---------------------------------------------------------------
static function frmWriteField( nHandle, nFieldsOffset, nLengthsOffset, nOffsetsOffset, nLastExpr, nExprCount )
//
//
// -->
//    logical, success or fail of write operation.
//   

    local cWriteItem  := ""
    local nWriteLen   := 0
    local nWriteCount := 0
    local nCntsOffset := 65535
    local nHdrOffset  := 65535
 
    // Write Contents 
    nCntsOffset := frmWriteExpr( nHandle, trim( Field->Contents ), .F.,; 
                     @nLastExpr, @nExprCount,; 
                     @nLengthsOffset, @nOffsetsOffset ) 
    // WRITE ok? 
    if nCntsOffset != -1
        // Write Header
        nHdrOffset := frmWriteExpr( nHandle,; 
                       trim( Field->Header ),; 
                       .T.,;
                       @nLastExpr,; 
                       @nExprCount,; 
                       @nLengthsOffset,; 
                       @nOffsetsOffset )
        // WRITE ok? 
        if nHdrOffset != -1
            // Seek to the next free FIELDS area 
            nFieldsOffset := nFieldsOffset + 12
            fseek( nHandle, nFieldsOffset )

            // SEEK ok? 
            if ferror() == 0
                cWriteItem := i2bin( Field->Width ) +;
                          replicate( chr(0), 3) +;
                          Field->Totals +;
                          i2bin( Field->Decimals ) +;
                          i2bin( nCntsOffset ) +;
                          i2bin( nHdrOffset )
                nWriteLen := len( cWriteItem )

                //  Write the FIELDS info 
                nWriteCount := fwrite( nHandle, cWriteItem, nWriteLen )

                if nWriteCount = 0 .or.;
                    ferror() <> 0
                    return .F.
                else
                    return .T.
                end
            end
        end
    end

    return .T.

//---------------------------------------------------------------
static function frmWriteParams( nHandle, anNum, axFormVal, nLastExpr )
//
// 
//  Writes the last 24 bytes of the report file plus
//   updates the first un-used offset. (last_offset)
//
// -->
//    logical, success or fail of write operation.
//

    local status       := .F.
    local cWriteItem   := ""
    local nWriteLen    := 0
    local nWriteCount  := 0 
    local nPlusByte    := 0

    // Calculate plus byte
    if axFormVal[ FRM_PLAINPAGE ] == "Y"
        nPlusByte += 4
    end
    if axFormVal[ FRM_PEAP ] == "Y"
        nPlusByte += 2
    end
    if axFormVal[ FRM_PEBP ] == "N"
        nPlusByte += 1
    end

    // Prepare miscellaneous data area string for write ops 
    cWriteItem = i2bin( anNum[ FRM_PAGE_HDR_NUM ] ) +;
                i2bin( anNum[ FRM_GRP_EXPR_NUM ] ) +;
                i2bin( anNum[ FRM_SUB_EXPR_NUM ] ) +;
                i2bin( anNum[ FRM_GRP_HDR_NUM ] ) +;
                i2bin( anNum[ FRM_SUB_HDR_NUM ] ) +;
                i2bin( axFormVal[ FRM_PAGE_WIDTH ] ) +;
                i2bin( axFormVal[ FRM_LINES_PAGE ] ) +;
                i2bin( axFormVal[ FRM_LEFT_MARG ] ) +;
                i2bin( axFormVal[ FRM_RIGHT_MARG ] ) +;
                i2bin( axFormVal[ FRM_COL_COUNT ] ) +;
                axFormVal[ FRM_DBL_SPACED ] +;
                axFormVal[ FRM_SUMMARY ] +;
                axFormVal[ FRM_PE ] +;
                chr( nPlusByte )
    nWriteLen := len( cWriteItem )

    // Seek to first parameters area (DOS offsets)
    fseek( nHandle, FRM_PARAMS_OFFSET -1  )

    // SEEK ok? 
    if ferror() == 0
        nWriteCount := fwrite( nHandle, cWriteItem, nWriteLen )
         
        if nWriteCount = 0 .or.;
            ferror() <> 0
            return .F.
        end
    else 
        return .F.
    end

    fseek( nHandle, 2 )  // next_free_offset

    if ferror() == 0   
        // Update the next free expression offset 
        nWriteCount := fwrite( nHandle, i2bin( nLastExpr ), 2)
        // write error
        if nWriteCount = 0 .or.;
            ferror() <> 0
            return .F.
        end
    end

    // se  arrivato fino a qui ok
    return .T.

//----------------------------------------------------------------
// LBL - LABELS CREATION / EDITING
//----------------------------------------------------------------
static function lblPrint( cLabelFile, bWhileCondition, bForCondition )
//
// lblPrint( [<cLabelFile>], [<bWhileCondition>], 
//      [<bForCondition>] ) --> NIL
//
// This function asks for informations when not given.
//
// <cReportFile>        the report (label) file name
//
// <bWhileCondition>    the WHILE code block condition
//
// <bForCondition>      the FOR code block condition
//


    local getlist   := {}
    local bOldErrorHandler
    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local bOld_F1       :=;
        setkey( K_F1, { || Text( LBL_HELP_LABEL_FORM ) } )
    local bOld_F2       := setkey( K_F2, NIL )
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
    
    local lGoPrint := .T.
    local cWhileCondition := padr( ".T.", SPACE_LEN )
    local cForCondition   := padr( ".T.", SPACE_LEN )
   
    local lLabelFile      := .F.
    local lWhileCondition := .F. 
    local lForCondition   := .F.

    begin sequence
   
        if ( alias() == "" ) // no alias
            alert( _ERROR_NO_ALIAS )
            break
        end
   
        // analisi dei parametri forniti   
        do case 
        case valtype(cLabelFile) == "C" .and.;
             !file( cLabelFile :=;
             padr( strAddExtention( cLabelFile,;
                 _EXTENTION_LABEL ),;
             SPACE_LEN) )
            lLabelFile := .T.
        case valtype(cLabelFile) <> "C" 
            lLabelFile := .T.
            cLabelFile := padr( "*." + _EXTENTION_LABEL,;
                SPACE_LEN )
        end

        do case
        case valtype(bWhileCondition) <> "B" 
            lWhileCondition := .T.
            cWhileCondition := padr( ".T.", SPACE_LEN )
        end
      
        do case
        case valtype(bForCondition) <> "B" 
            lForCondition := .T.
            cForCondition := padr( ".T.", SPACE_LEN )
        end
      
        // check for parameters
        if lLabelFile .or.;
           lWhileCondition .or.;
           lForCondition
            
            // Create a kind of window.
            nBottom    := maxrow()
            nTop       := nBottom - 8
            nLeft      := 0
            nRight     := maxcol()
            nWidth     := nRight - nLeft +1
            cOldScreen := savescreen( nTop, nLeft,;
                nBottom, nRight )
            setcolor( COLOR_DEFAULT1 )
            scroll( nTop, nLeft, nTop, nRight )
            scroll( nBottom-1, nLeft, nBottom, nRight )
            devpos( nTop, nLeft )
            devout( padc( LBL_DIALOG_BOX_LABEL_FORM_TOP, nWidth ) )
            devpos( nBottom-1, nLeft )
            devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
            devpos( nBottom, nLeft )
            devout( padc( _KEY_F2_PGDN, nWidth ) )
   
            while .t.
                setcolor( COLOR_DEFAULT2 )
                scroll( nTop+1, nLeft, nBottom-2, nRight )
                setpos( nTop, nLeft )
                if lLabelFile
                    @row()+1,nLeft say LBL_PROMPT_OPEN_FILE_LABEL
                    @row()+1,nLeft get cLabelFile;
                        picture "@s"+ltrim(str(nWidth))+"@";
                        when trueSetkey( K_F2,;
                            {|| cLabelFile :=;
                            padr( Dir( cLabelFile ),;
                            SPACE_LEN ) } );
                        valid ( file( cLabelFile :=;
                            padr( strAddExtention( cLabelFile,;
                            _EXTENTION_LABEL ), SPACE_LEN) ) );
                            .and. !( "*" $ cLabelFile );
                            .and. !( "?" $ cLabelFile )
                end
                if lWhileCondition
                    @row()+1,nLeft say LBL_PROMPT_WHILE_EXPRESSION
                    @row()+1,nLeft get cWhileCondition;
                        picture "@s"+ltrim(str(nWidth))+"@";
                        when trueSetkey( K_F2,;
                            {|| cWhileCondition :=;
                            padr( fieldNormal(),;
                            SPACE_LEN ) } );
                        valid ( ( cWhileCondition :=;
                            iif( cWhileCondition <> SPACE_STR,;
                            cWhileCondition, padr( ".T.",;
                            SPACE_LEN ) ) ) <> NIL )
                end
                if lForCondition
                    @row()+1,nLeft say LBL_PROMPT_FOR_EXPRESSION
                    @row()+1,nLeft get cForCondition;
                        picture "@s"+ltrim(str(nWidth))+"@";
                        when trueSetkey( K_F2,;
                            {|| cForCondition :=;
                            padr( fieldNormal(),;
                            SPACE_LEN ) } );
                        valid ( ( cForCondition :=;
                            iif( cForCondition <> SPACE_STR,;
                            cForCondition, padr( ".T.",;
                            SPACE_LEN ) ) ) <> NIL )
                end
                read

                do case
                case lastkey() = K_ESC  // exit
                    lGoPrint := .F.
                    exit
                case lastkey() = K_PGDN // confirm
                    // check for correct data
                    if file( strAddExtention( cLabelFile,;
                        _EXTENTION_LABEL ) );
                        .and. cWhileCondition <> SPACE_STR;
                        .and. cForCondition <> SPACE_STR
                        //
                        lGoPrint := .T.
                        exit
                    else
                        // loop
                    end
                    // transform character conditions
                    if lWhileCondition
                        bWhileCondition := { || &cWhileCondition }
                    end
                    if lForCondition
                        bForCondition := { || &(cForCondition) .and. WaitFileEval() }
                    end
                otherwise
                    // loop
                end
            end
   
            // close window
            restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        end

        if lGoPrint
            bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
            begin sequence
                 __LabelForm( cLabelFile, .f.,, .f.,;
                     bForCondition, bWhileCondition, , , .f., .f. )
                dbgotop() // otherwise it stays on eof()
         
            recover
                dbgotop()
            end sequence
            errorblock(bOldErrorHandler)
            WaitProgress(2)
        end

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL

//----------------------------------------------------------------
static function lblModify()
//
// lblModify() --> NIL
//
// It executes the lbl() function asking for the file name to
// modify.
//   

    local getlist   := {}
    local bOldErrorHandler
    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local bOld_F1       :=;
        setkey( K_F1, { || Text( LBL_HELP_LABEL_EDIT ) } )
    local bOld_F2       := setkey( K_F2, NIL )
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
                  
    local lGoOn         := .F.
    local cLabelFile   := padr( "*." + _EXTENTION_LABEL,;
        SPACE_LEN )

    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 4
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( LBL_WINDOW_LABEL, nWidth ) )
        devpos( nBottom-1, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_F2_PGDN, nWidth ) )

        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-2, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say LBL_PROMPT_OPEN_FILE_LABEL
            @row()+1,nLeft get cLabelFile;
                picture "@s"+ltrim(str(nWidth))+"@";
                when trueSetkey( K_F2,;
                     {|| cLabelFile :=;
                     padr( Dir( cLabelFile ), SPACE_LEN ) } );
                valid file( cLabelFile :=;
                      padr( strAddExtention( cLabelFile,;
                      _EXTENTION_LABEL ), SPACE_LEN) )
            read

            do case
            case lastkey() = K_ESC  // exit
                lGoOn := .F.
                exit
            case lastkey() = K_PGDN // ok
                if file( strAddExtention( cLabelFile, _EXTENTION_LABEL ) ) 
                    lGoOn := .T.
                    exit
                else
                    // loop
                end
            otherwise
                // loop
            end
        end
   
        // close window
        restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
   
        if lGoOn
            lbl( cLabelFile )
        end
   
    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL

//----------------------------------------------------------------
static function lbl( cFileName )  // la parola Label  corretta
//
// lbl ( [<cFileName>] ) --> NIL
//
// <cFileName>  the label file name to modify.
//              If the name is not given, it is created.
//
   
    local getlist   := {}
    local bOldErrorHandler
    local cOldScreen    := ""
    local cOldColor     := setcolor()
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local bOld_F1       :=;
        setkey( K_F1, { || Text( LBL_HELP_LABEL_SAVE ) } )
    local bOld_F2       := setkey( K_F2, NIL )
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
   
    local axLblDim := { space(60), 5, 35, 0, 1, 0, 1 }  // default
    local acLblCont[16]
    local nI
    local cOldName
    local lSave := .F.
    local lGoOn := .F.

    begin sequence
   
        if valtype( cFileName ) <> "C"
             cFileName := strAddExtention( _UNTITLED,;
                _EXTENTION_LABEL )
             cOldName := NIL
             // it must be created
             // the array acLblCont[] is prepared
             for nI = 1 to 16
                 acLblCont[nI] := space(60)
             next
         else
            if file( cFileName )
                cOldName := alltrim(cFileName)
                // the label file is created
                lblLoad( cFileName, @axLblDim, @acLblCont )
            else
                cOldName := NIL
                // it must be created
                // the array acLblCont[] is prepared
                for nI = 1 to 16
                    acLblCont[nI] := space(60)
                next
            end
        end

        lSave := lblEdit( @axLblDim, @acLblCont )
      
        if lSave
         
            cFileName := padr( cFileName, SPACE_LEN )

            // Create a kind of window.
            nBottom     := maxrow()
            nTop        := nBottom - 4
            nLeft       := 0
            nRight      := maxcol()
            nWidth      := nRight - nLeft +1
            cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
            setcolor( COLOR_DEFAULT1 )
            scroll( nTop, nLeft, nTop, nRight )
            scroll( nBottom-1, nLeft, nBottom, nRight )
            devpos( nTop, nLeft )
            devout( padc( LBL_WINDOW_LABEL, nWidth ) )
            devpos( nBottom-1, nLeft )
            devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
            devpos( nBottom, nLeft )
            devout( padc( _KEY_F2_PGDN, nWidth ) )
      
            while .t.
                setcolor( COLOR_DEFAULT2 )
                scroll( nTop+1, nLeft, nBottom-2, nRight )
                setpos( nTop, nLeft )
                @row()+1,nLeft say LBL_PROMPT_SAVE_FILE_LABEL
                @row()+1,nLeft get cFileName;
                    picture "@s"+ltrim(str(nWidth))+"@";    
                    when trueSetkey( K_F2,;
                        {|| cFileName :=;
                        padr( Dir( cFileName ), SPACE_LEN ) } );
                    valid ( ( cFileName :=;
                        padr(strAddExtention(cFileName,;
                        _EXTENTION_LABEL), SPACE_LEN) ) <> NIL )
                read

                do case
                case lastkey() = K_ESC  // exit
                    lGoOn := .F.
                    exit
                case lastkey() = K_PGDN // ok
                    do case
                    case alltrim(cFileName) <> cOldName;
                        .and. file( strAddExtention( cFileName,;
                        _EXTENTION_LABEL ) );
                        .and. ( alert( alltrim(cFileName) + ";" +;
                        _ERROR_PROMPT_ALREADY_EXIST_OVERWRITE,;
                        { _MENU_NO, _MENU_YES } ) == 1 )
                        //
                        // loop
                    case strCutExtention( cFileName ) == ""
                        // loop
               
                    otherwise
                        lGoOn := .T.
                        exit
                    end
            
                otherwise
                    // loop
                end
            end
   
            if lGoOn
                // ok save
                if lblWrite( alltrim(cFileName), axLblDim, acLblCont ) 
                    // ok
                else
                    alert( LBL_ERROR_FILE_NOT_SAVED )
                end
            else
               // nix
            end

        else
            // nix
        end

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )

    return NIL
 
//----------------------------------------------------------------
static function lblLoad( cFileName, axLblDim, acLblCont )

    local cBuffer   := space( LBL_SIZE )
    local nHandle   := 0
    local nByteRead := 0
    local nOffset   := 0
    local nI

    nHandle := fopen( cFileName )

    if !ferror() == 0
        alert( cFileName + ";" +;
            LBL_ERROR_CANNOT_OPEN + str(ferror()) )
        return .F.
    end

    // legge il file
    nByteRead := fread( nHandle, @cBuffer, LBL_SIZE )
    if nByteRead < LBL_SIZE
        alert( cFileName + ";" +; 
            LBL_ERROR_FILE_TOO_LITTLE )
        return .F.
    end

    WaitFor( LBL_WAIT_LOADING )
   
    // legge le dimensioni
    axLblDim[LBL_REMARK] := substr( cBuffer, 2, 60 )
    axLblDim[LBL_HEIGHT] := bin2I( substr( cBuffer, 62, 2 ) )
    axLblDim[LBL_WIDTH]  := bin2I( substr( cBuffer, 64, 2 ) )
    axLblDim[LBL_MARGIN] := bin2I( substr( cBuffer, 66, 2 ) )
    axLblDim[LBL_LINES]  := bin2I( substr( cBuffer, 68, 2 ) )
    axLblDim[LBL_SPACES] := bin2I( substr( cBuffer, 70, 2 ) )
    axLblDim[LBL_ACROSS] := bin2I( substr( cBuffer, 72, 2 ) )

    // legge il contenuto delle linee della etichetta
    nOffset := 74
    for nI = 1 to 16 // legge tutto
        acLblCont[nI] := substr( cBuffer, nOffset, 60 )
        nOffset += 60
    next

    fclose( nHandle )

    WaitFor()
   
    // ok
    return .T.

//----------------------------------------------------------------
static function lblEdit( axLblDim, acLblCont )
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local bOld_F1               :=;
        setkey( K_F1, {|| Text( LBL_HELP_LABEL ) } ) 
    local bOld_F2               := setkey( K_F2, NIL )
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
   
    local lSave := .F.

    begin sequence

        // Create a kind of window.
        nBottom     := maxrow()
        nTop        := nBottom - 24  // max
        nLeft       := 0
        nRight      := maxcol()
        nWidth      := nRight - nLeft +1
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom-1, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( LBL_WINDOW_LABEL, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_ESC_CANCEL_F1_F2_PGDN, nWidth ) )
      
        while .t.
            setcolor( COLOR_DEFAULT2 )
            scroll( nTop+1, nLeft, nBottom-1, nRight )
            setpos( nTop, nLeft )
            @row()+1,nLeft say LBL_PROMPT_LABEL_REMARK;
                get axLblDim[ LBL_REMARK ];
                   picture "@s"+ltrim(str(nWidth))+"@";
                   when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say LBL_PROMPT_LABEL_HEIGHT; 
                get axLblDim[ LBL_HEIGHT ] picture "999";
                   when trueSetkey( K_F2, NIL );
                   valid ( axLblDim[ LBL_HEIGHT ] > 0 .and.;
                           axLblDim[ LBL_HEIGHT ] <= 16 )
            @row()+1,nLeft say LBL_PROMPT_LABEL_WIDTH; 
                get axLblDim[ LBL_WIDTH ]  picture "999";   
                   when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say LBL_PROMPT_LABEL_MARGIN; 
                get axLblDim[ LBL_MARGIN]  picture "999";   
                   when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say LBL_PROMPT_LABEL_LINES; 
                get axLblDim[ LBL_LINES ]  picture "999";   
                   when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say LBL_PROMPT_LABEL_SPACES; 
                get axLblDim[ LBL_SPACES]  picture "999";   
                   when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say LBL_PROMPT_LABEL_ACROSS; 
                get axLblDim[ LBL_ACROSS]  picture "999";   
                   when trueSetkey( K_F2, NIL )
            @row()+1,nLeft say "   1   "; 
                get acLblCont[1];
                    picture "@s"+ltrim(str(nWidth))+"@";
                    when axLblDim[LBL_HEIGHT] >= 1 .and.;
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   2   "; 
                get acLblCont[2];
                    picture "@s"+ltrim(str(nWidth))+"@";
                    when axLblDim[LBL_HEIGHT] >= 2 .and.;
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   3   "; 
                get acLblCont[3];
                    picture "@s"+ltrim(str(nWidth))+"@";
                    when axLblDim[LBL_HEIGHT] >= 3 .and.;
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   4   "; 
                get acLblCont[4];
                    picture "@s"+ltrim(str(nWidth))+"@";
                    when axLblDim[LBL_HEIGHT] >= 4 .and.;
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   5   "; 
                get acLblCont[5];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 5 .and.;  
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   6   "; 
                get acLblCont[6];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 6 .and.;  
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   7   "; 
                get acLblCont[7];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 7 .and.;  
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   8   "; 
                get acLblCont[8];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 8 .and.;  
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "   9   "; 
                get acLblCont[9];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 9 .and.;  
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "  10   "; 
                get acLblCont[10];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 10 .and.; 
                        trueSetkey( K_F2,;
                        { || keyboard( fieldPicture() ) } )
            @row()+1,nLeft say "  11   "; 
                get acLblCont[11];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 11 .and.; 
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )     
            @row()+1,nLeft say "  12   "; 
                get acLblCont[12];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 12 .and.; 
                        trueSetkey( K_F2,;
                        { || keyboard( fieldPicture() ) } )     
            @row()+1,nLeft say "  13   "; 
                get acLblCont[13];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 13 .and.; 
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )     
            @row()+1,nLeft say "  14   "; 
                get acLblCont[14];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 14 .and.; 
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )     
            @row()+1,nLeft say "  15   "; 
                get acLblCont[15];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 15 .and.; 
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )     
            @row()+1,nLeft say "  16   "; 
                get acLblCont[16];
                    picture "@s"+ltrim(str(nWidth))+"@"; 
                    when axLblDim[LBL_HEIGHT] >= 16 .and.; 
                        trueSetkey( K_F2,;
                            { || keyboard( fieldPicture() ) } )     
            read
         
            do case
            case lastkey() = K_ESC
                lSave := .F.
                exit
            case lastkey() = K_PGDN // conferma la scelta
                lSave := .T.
                exit
            otherwise
                // loop
            end
            // loop
        end

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    // termina la procedura
    return lSave

//---------------------------------------------------------------
static function lblWrite( cFileName, axLblDim, acLblCont ) 

    local cBuffer
    local nHandle   := 0
    local nByteWritten := 0
    local nI

    nHandle := fcreate( cFileName )

    if !ferror() == 0
        alert( cFileName + ";" +;
            LBL_ERROR_CANNOT_CREATE + ;
            str(ferror()) )
        return .F.
    end

    WaitFor( LBL_WAIT_SAVING )
   
    // prepara il buffer
    cBuffer := chr(2)
    cBuffer += axLblDim[LBL_REMARK]
    cBuffer += i2bin(axLblDim[LBL_HEIGHT])
    cBuffer += i2bin(axLblDim[LBL_WIDTH])
    cBuffer += i2bin(axLblDim[LBL_MARGIN])
    cBuffer += i2bin(axLblDim[LBL_LINES])
    cBuffer += i2bin(axLblDim[LBL_SPACES])
    cBuffer += i2bin(axLblDim[LBL_ACROSS])
    for nI = 1 to 16
        cBuffer += acLblCont[nI]
    next
    cBuffer += chr(2)

    nByteWritten := fwrite( nHandle, cBuffer ) 

    if nByteWritten != LBL_SIZE
        alert( LBL_ERROR_SOMETHING_GONE_WRONG )
    end

    fclose( nHandle )

    WaitFor()
   
    // ok
    return .T.

//================================================================
//================================================================
// TEXT DISPLAY
//================================================================
function Text( cText )
//
// Text( <cText> ) --> NIL
//
//
// <cText>      Text to display.
//

    local nOldCursor := setcursor( SC_NONE )
    local cOldScreen := savescreen()
    local cOldColor  := setcolor()
    local bOld_F1    := setkey( K_F1, NIL )   // to avoid recursion when  usign as help
    local nOldRow    := row()
    local nOldCol    := col()

    begin sequence

        if cText == NIL;
            .or. cText == ""
            //
            break
        end

        // full screen
        setcolor( COLOR_DEFAULT1 )
        scroll( maxrow(), 0, maxrow(), maxcol() )
        devpos( maxrow(), 0 )
        devout( padc( TEXT_KEY_REMINDER, maxcol()+1 ) )

        setcolor( COLOR_DEFAULT2 )
        scroll( 0, 0, maxrow()-1, maxcol() )

        memoedit( cText + ENDTEXT, 00,00,maxrow()-1,maxcol(), .F. )

    end

    setcursor( nOldCursor )
    setcolor( cOldColor )
    restscreen( NIL, NIL, NIL, NIL, cOldScreen )
    setkey( K_F1, bOld_F1 )
    setpos( nOldRow, nOldCol )

    return NIL

//================================================================
// WAIT  - These functions must be "opened" and then "closed".
//================================================================
function WaitFileEval()
//
// WaitFileEval() --> .T.
//
// Shows a wait bar calculated with recno()/lastrec(),
//

    if recno() = lastrec()
        WaitProgress(2) // > 100%
    else
        WaitProgress( recno()/lastrec() )
    end

   return .T.

//================================================================
function WaitProgress( nPercent )
//
// WaitProgress ( [<nPercent>] ) --> NIL
//
// <nPercent>   the actual percent value (1 = 100%).
//              When nPercent > 1 the wait bar is closed.
//              When nPercent == 1 the wait bar is displayed
//              and closed.
//
// Shows a wait bar at the display bottom.
//
// This function must be "closed".
//

    static cOldScreen

    local nElements
    local nBar
    local nOldRow    := row()
    local nOldCol    := col()

    begin sequence

        if cOldScreen == NIL
            cOldScreen := savescreen( maxrow(),00,maxrow(),maxcol() )
        end

        if nPercent == NIL .or.;
           nPercent > 1 
       
            restscreen( maxrow(),00,maxrow(),maxcol(), cOldScreen )
            cOldScreen := NIL

            break
        end
   
        nElements := maxcol()+1
        nBar := nElements*nPercent
        if nBar < 0
            nBar := 0
        end

        devpos( maxrow(),00 )
        devout( replicate( "", nBar ) )
        devout( replicate( "", nElements-nBar ) )

        // if 100%
        if nPercent >= 1
            restscreen( maxrow(),00,maxrow(),maxcol(), cOldScreen )
            cOldScreen := NIL
        end

    end

    // restore
    setpos( nOldRow, nOldCol )
    
    return NIL

//================================================================
function WaitFor( cMessage )
//
// This function was originally named "Wait()", but the name is
// now changed as the compiler confuses it with the command
// WAIT
//
// WaitFor( [<cMessage>] ) --> .T.
//
// <cMessage>   Text message to show. If it is NIL, it closes
//              the message.
//
// It shows a wait message at the center of the screen.
//
// WaitFor() without argument, closes the message.
//
// This function must be closed.
//

    static cOldScreen
    static nTop
    static nLeft
    static nBottom
    static nRight
    
    local nOldCursor := setcursor( SC_NONE )
    local cOldColor  := setcolor()
    local nOldRow    := row()
    local nOldCol    := col()

    local nWidth := 0

    local nLines
    local nLine

    begin sequence

        // wait close?
        if valtype( cMessage ) <> "C"
            if cOldScreen <> NIL
                restscreen( nTop,nLeft,nBottom,nRight, cOldScreen )
                cOldScreen := NIL
            end

            break
    
        end

        // already opened?
        if cOldScreen <> NIL
            restscreen( nTop,nLeft,nBottom,nRight, cOldScreen )
            cOldScreen := NIL
        end
    
        // line count
        cMessage := alltrim(cMessage)
        nLines := mlcount( cMessage, maxcol()-1 )
        nLine := 0
        while nLine < nLines
            nLine++
            nWidth := max( nWidth,; 
                len( rtrim( memoline(cMessage, maxcol()-1, nLine ) ) ); 
                         )
        end
   
        nTop        := int( ( (maxrow()+1) - (nLines+2) ) / 2 )
        nLeft       := int( ( (maxcol()+1) - (nWidth+2) ) / 2 ) 
        nBottom     := nTop + nLines+1
        nRight      := nLeft + nWidth+1

        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
    
        setcolor( COLOR_DEFAULT2 )
        dispbox( nTop, nLeft, nBottom, nRight, B_DOUBLE ) 
        scroll( nTop+1, nLeft+1, nBottom-1, nRight-1 )
        nLine := 1
        while nLine <= nLines
            devpos( nTop+nLine, nLeft+1 )
            devout( memoline(cMessage, nWidth, nLine ) )
            nLine++
        end

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    setpos( nOldRow, nOldCol )
    
    return NIL

//================================================================
// PROMPT FUNCTIONS
//================================================================
function Accept( Field, cMessage )
//
// Accept(<Field>, [<cMessage>]) --> Field|NIL
//
// <Field>      the variable to be edited and returned
//
// <cMessage>   the message to show
//
// The function shows <cMessage> and waits for input to place
// into <Field>.
//

    local getlist := {}
    local cOldScreen    := ""
    local cOldColor     := setcolor() 
    local nOldCursor    := setcursor( SC_NORMAL ) 
    local nOldRow       := row()
    local nOldCol       := col()
    local nTop          := 0
    local nLeft         := 0
    local nBottom       := 0
    local nRight        := 0
    local nWidth        := 0
    local nLines
    local nLine
    local ReturnField   := NIL

    default cMessage to ""

    begin sequence

        if Field == NIL
            break
        end

        if valtype( cMessage ) <> "C"
            cMessage := NIL
        end
    
        // calculate field dimention
        nWidth := maxcol()+1

        // message line count
        cMessage := alltrim(cMessage)
        nLines := mlcount( cMessage, nWidth )
        
        // calculate window
        nBottom     := maxrow()
        nTop        := nBottom - ( nLines+1+1 )
        nLeft       := 0
        nRight      := maxcol()
        
        // save window area
        cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
        
        // create window
        setcolor( COLOR_DEFAULT1 )
        scroll( nTop, nLeft, nTop, nRight )
        scroll( nBottom, nLeft, nBottom, nRight )
        devpos( nTop, nLeft )
        devout( padc( ACCEPT_DIALOG_BOX_TOP, nWidth ) )
        devpos( nBottom, nLeft )
        devout( padc( _KEY_ESC_CANCEL_PGDN, nWidth ) )

        setcolor( COLOR_DEFAULT2 )
        scroll( nTop+1, nLeft, nBottom-1, nRight )
        
        nLine := 1
        while nLine <= nLines
            devpos( nTop+nLine, nLeft )
            devout( memoline(cMessage, nWidth, nLine ) )
            nLine++
        end
        
        do case
        case valtype( Field ) == "C" 
            devpos( nTop+nLines+1, nLeft )
            @row(),col() get Field picture "@s"+ltrim(str(nWidth))+"@"
        case valtype( Field ) == "N"
            devpos( nTop+nLines+1, nLeft )
            @row(),col() get Field 
        case valtype( Field ) == "D"
            devpos( nTop+nLines+1, nLeft )
            @row(),col() get Field 
        otherwise
            devpos( nTop+nLines+1, nLeft )
            @row(),col() get Field 
        end
        
        readmodal( getlist )
        getlist := {}
        
        do case
        case lastkey() = K_ESC
            ReturnField := NIL
        otherwise
            ReturnField := Field
        end
        
    end sequence

    // restore
    setcolor( cOldColor )
    setcursor( nOldCursor ) 
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )

    return ReturnField

//================================================================
// STR - STRING PATH/FILE NAME FUNCTIONS
//================================================================
function strTempPath()
//
// strTempPath() --> cTempPath
//
// The function tries to find if a temporary directory is defined
// with enviromental variables.
//

    local cTempPath := ""
   
    do case
    case getenv( "TEMP" ) == "" .and.;
         getenv( "TMP" ) == ""
        cTempPath := "."
    case !( getenv( "TEMP" ) == "" )
        cTempPath := getenv( "TEMP" )
    case !( getenv( "TMP" ) == "" )
        cTempPath := getenv( "TMP" )
    end
    
    return cTempPath

//================================================================
function strAddExtention( cName, cExt )
//
// StrAddExtention( <cName>, <cExt> ) --> cCompleteName
//
// <cName>      the filename or pathname without extention.
//
// <cExt>       the extention to be added.
//
// The function tries to add the extention <cExt> to <cName> if
// <cName> has no extention jet.
//

   local cLastFour
   local cNewName

   default cExt to ""

   cName := alltrim( cName )

   // toglie il puntino dall'estensione
   if left( cExt, 1 ) == "."
      cExt := substr( cExt, 2 )
   end

   cLastFour := right( cName, 4 )

   if at( ".", cLastFour ) == 0 // non c' estensione
      cNewName := cName + "." + cExt
   else
      cNewName := cName
   end

   return cNewName

//================================================================
function strCutExtention( cName )
//
// StrCutExtention( <cName> ) --> cName
//
// <cName>      the file name or pathname with extention.
//
// The funcion tries to cut the extention from <cName>.
//

   local cLastFour
   local cNewName

   cName := alltrim( cName )

   cLastFour := right( cName, 4 )

   if at( ".", cLastFour ) == 0 // non c' estensione
      cNewName := cName
   else
      cNewName := left( cName, rat( ".", cName )-1 )
   end

   return cNewName

//================================================================
function strDrive( cName )
//
// StrDrive( <cName> ) --> cDrive
//
// <cName>      the pathname where is to extract the drive.
//
// The function tries to extract the drive name from <cName>.
//

    local cDrive

    cName := ltrim( cName )

    if subst( cName, 2, 1 ) == ":"
        cDrive := left( cName, 2 )
    else
        cDrive := ""
    end

    return cDrive

//================================================================
function strExtention( cName )
//
// StrExtention( <cName> ) --> cExtension
//
// <cName>      the filename or pathname where is to extract
//              the file extention.
//
// The function tries to extract the file extention from
// <cName>.
//

   local cLastFour
   local cExt

   cName := alltrim( cName )

   cLastFour := right( cName, 4 )

   if at( ".", cLastFour ) == 0 // non c' estensione
      cExt := ""
   else
      cExt := right( cName, ( len(cName) - rat( ".", cName ) ) )
   end

   return cExt

//================================================================
function strFile( cName )
//
// StrFile( <cName> ) --> cFileName
//
// <cName>      the filename or pathname where is to extract
//              the file name.
//
// The function tries to extract the file name without path
// from <cName>.
//

   local nLastSlash
   local nLastColon

   cName := alltrim( cName )

   nLastSlash := rat( "\", cName )
   if nLastSlash > 0
      cName := substr( cName, nLastSlash+1 )
   end
   
   nLastColon := rat( ":", cName )
   if nLastColon > 0
      cName := substr( cName, nLastColon+1 )
   end

   return cName

//================================================================
function strPath( cName )
//
// StrPath( <cName> ) --> cPath
//
// <cName>      the filename or pathname where is to extract
//              the path.
//
// The function tries to extract the path from <cName>.
//

   local cPath
   local nLastSlash
   
   cName := alltrim( cName )

   nLastSlash := rat( "\", cName )
   
   if nLastSlash > 0
       if substr( cName, 2, 1 ) == ":"  // drive
           cPath := strPath( substr( cName, 3 ) )  // recursion
       else
           cPath := substr( cName, 1, nLastSlash )
       end
   else
       cPath := ""
   end

   return cPath

//================================================================
function strParent( cPathName )
//
// StrParent( <cPathName> ) --> cParentPath
//
// <cPath>      the path to transform.
//
// The function tries to return a parent path.
//
   local cDrive := strDrive( cPathName )
   local cPath  := strPath( cPathName )
   local cFile  := strFile( cPathName )

   if len( cPath ) > 0
       do case
       case cPath == "\"
           // root; it remains as it is
       case right( cPath, 1 ) == "\"
           // normal condition
           cPath := strPath( left( cPath, len( cPath )-1 ) )
       otherwise
           // unknown situation
       end
   end

   return cDrive+cPath+cFile

//================================================================
// TB - BROWSE 
//================================================================
function TB( nTop, nLeft, nBottom, nRight,;
                acCol, acColSayPic, acColHead,;
                alColCalc, nFreeze,;
                lModify, lAppend, lDelete,; 
                lAutoSort )
//
// TB( <nTop>, <nLeft>, <nBottom>, <nRight>,
//      <acCol>, <acColSayPic>, [<acColHead>], 
//      [<alColCalc>, [<nFreeze>],
//      [<lModify>], [<lAppend>], [<lDelete>],
//      [<lAutosort>] )  --> NIL
// 
// <nTop>, <nLeft>, <nBottom>, <nRight>         The display area 
//                                              used to browse.
// <acCol>              Column array.
// <acColSayPic>        Picture array.
// <acColHead>          Column head description array.
// <alColCalc>          Calculated column array: .T. menas
//                      calculated, .F. means editable.
// <nFreeze>            Number of columns to be left visible.
// <lModify>            Ability to modify the file.
// <lAppend>            Ability to append data.
// <lDelete>            Ability to delete records.
// <lAutosort>          Ability to refresh record position
//                      when the file is indexed.
//

    local getlist := {}
    local bOldErrorHandler
    local lReadExit
    local nOldCursor
    local nOldRow
    local nOldCol
    local bOldF1
    
    local nALen := 0

    local oBrowse
    local nKey
    local lMore := .T.
    local alColMemo
   
    default nTop     to 0
    default nLeft    to 0
    default nBottom  to maxrow()
    default nRight   to maxcol()
   
    default acColSayPic to; 
        iif( acCol == NIL, NIL, array( len( acCol ) ) )
    default acColHead to; 
        iif( acCol == NIL, NIL, aclone( acCol ) )
    default alColCalc to; 
        iif( acCol == NIL, NIL, afill( array( len( acCol ) ), .F. ) )
    default nFreeze   to 0
    default lModify   to .T.
    default lAppend   to .T.
    default lDelete   to .T.
    default lAutosort to .T.

    if setkey( K_F1 ) == NIL // there is no previous help
        bOldF1 := setkey( K_F1,;
            { || Text( TB_HELP ) } )
    else
        bOldF1 := setkey( K_F1 )
    end    

    if acCol == NIL
        TBDefault( @acCol, @acColSayPic, @alColCalc, @acColHead )
        nFreeze := 1
    end
    if acCol == NIL
       alert( TB_ERROR_NO_COLUMNS )
       return NIL
    end
    
    alColMemo := array( len( acCol ) )

    if acColSayPic == NIL
        alert( TB_ERROR_NO_PICTURES ) 
        return NIL
    end   
    nALen := len( acCol )
    if !( nALen == len( acColSayPic ) .and.;
          nALen == len( acColHead ) .and.;
          nALen == len( alColCalc );
        )
        alert( TB_ERROR_ARRAY_DIFFERENT_DIMENTIONS )
        return NIL
    end

    // if it is not possible to modify...
    if !lModify
        lAppend := .F.
        lDelete := .F.
    end
   
    // if eof() it is better to move the record pointer
    if eof()
        dbgobottom()
    end
    if bof()
        dbgotop()
    end

    lReadExit           := readexit(.T.)          
    nOldCursor          := setcursor()
    nOldRow             := row()
    nOldCol             := col()
    
    bOldErrorHandler    := errorblock( {|e| ErrorHandler(e)} )
    begin sequence
   
        // Make a Tbrowse object for the current workarea
        oBrowse := TBNew( nTop, nLeft, nBottom, nRight,;
                        acCol, acColHead, alColCalc, alColMemo )
   
        TB_APP_MODE_OFF(oBrowse)
        
        // Use a custom 'skipper' to handle append mode (see below)
        oBrowse:skipBlock := { |nSkip| TBSkip( nSkip, oBrowse ) }
   
        // Define pictures
        TBColPicture(oBrowse, acColSayPic)
   
        // turn the cursor off while browsing
        setcursor(SC_NONE)
   
        // freeze columns
        oBrowse:freeze := nFreeze
   
        // Main loop
        while lMore
            // Don't let the cursor move into frozen columns
            if ( oBrowse:colPos <= oBrowse:freeze )
                oBrowse:colPos := ( oBrowse:freeze + 1 )
            end
            // Stabilize the display until it's stable or a key is pressed
            oBrowse:forceStable()
            // check hit top/bottom
            if ( oBrowse:hitBottom .and.;
                 !TB_APP_MODE_ACTIVE( oBrowse ) .and.;
                 lModify .and.; 
                 lAppend; 
               )
                TB_APP_MODE_ON( oBrowse )
                nKey := K_DOWN
            else
                // Make sure that the current record is showing
                // up-to-date data in case we are on a network.
                oBrowse:refreshCurrent():forceStable()
                // Everything's done -- just wait for a key
                nKey := inkey( 0 )
            end
      
            do case
            case ( nKey == K_ESC )
                // Esc means leave
                lMore := .F.
            otherwise // normally
                // Apply the key to the oBrowse
                TBApplyKey( oBrowse, nKey, acCol, alColCalc,; 
                    acColHead, alColMemo, lModify, lDelete,; 
                    lAutosort )
            end
      
        end

    recover
        // nil
    end sequence
    
    errorblock(bOldErrorHandler)
    readexit(lReadExit)          
    setcursor(nOldCursor)
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOldF1 )
    
    return NIL

//----------------------------------------------------------------
static function TBSkip(nSkip, oBrowse)
//
// TBSkip(nSkip, oBrowse) --> nSkippedRecords
//
// Handle record movement requests from the Tbrowse object.
//
   
    local lAppend       := TB_APP_MODE_ACTIVE( oBrowse )
    local nI            := 0
   
    do case
    case ( nSkip == 0 .or. lastrec() == 0 )
        // Skip 0 (significant on a network)
        dbSkip(0)
    case ( nSkip > 0 .and. !eof() )
        // Skip Forward
        while (nI < nSkip)     
            dbskip( 1 )
            if eof()
                iif(lAppend, nI++, dbskip(-1))
                exit
            end
            nI++
        end
    case ( nSkip < 0 )
        // Skip backward
        while ( nI > nSkip )           
            dbskip(-1)
            if bof()
               exit
            end
            nI--
        end
    end
   
   return nI

//----------------------------------------------------------------
static function TBApplyKey( oBrowse, nKey,; 
                    acCol, alColCalc, acColHead, alColMemo,; 
                    lModify, lDelete, lAutosort; 
                          )
//
// Apply one keystroke to the oBrowse.
//
   
    local bOldErrorHandler
    local lMemo := .F.
   
    // determina se si tratti di un campo memo o di
    // campo molto grande
    lMemo := alColMemo[oBrowse:colPos]
        
    do case
    case nKey == K_DOWN
        oBrowse:down()
    case nKey == K_PGDN
        oBrowse:pageDown()
    case nKey == K_CTRL_PGDN
        oBrowse:goBottom()
        TB_APP_MODE_OFF(oBrowse)
    case nKey == K_UP
        oBrowse:up()
        if TB_APP_MODE_ACTIVE( oBrowse )
            TB_APP_MODE_OFF( oBrowse )
            oBrowse:refreshAll()
        end
    case nKey == K_PGUP
        oBrowse:pageUp()
        if TB_APP_MODE_ACTIVE( oBrowse )
            TB_APP_MODE_OFF( oBrowse )
            oBrowse:refreshAll()
        end
    case nKey == K_CTRL_PGUP
        oBrowse:goTop()
        TB_APP_MODE_OFF( oBrowse )
    case nKey == K_RIGHT
        oBrowse:right()
    case nKey == K_LEFT
        oBrowse:left()
    case nKey == K_HOME
        oBrowse:home()
    case nKey == K_END
        oBrowse:end()
    case nKey == K_CTRL_LEFT
        oBrowse:panLeft()
    case nKey == K_CTRL_RIGHT
        oBrowse:panRight()
    case nKey == K_CTRL_HOME
        oBrowse:panHome()
    case nKey == K_CTRL_END
        oBrowse:panEnd()
    case nKey == K_RETURN
        do case
        case !lModify  // read only 
            // no edit allowed
            // (to view a memo, use a calculated column)
        case !alColCalc[oBrowse:colPos]
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock())
                // in caso di read only ed altre eventualit
                bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
                begin sequence
                    TBDoGet(oBrowse, acCol, lMemo, lAutosort)
                    dbcommit()
                recover
                    // nil
                end
                errorblock(bOldErrorHandler)
                dbunlock()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
        otherwise
            // nil
        end
    case nKey == K_CTRL_Y .or.;
         nKey == K_CTRL_DEL     // delete
        do case
        case lModify .and.; 
             lDelete .and.; 
             !deleted()
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock()) .and.;
               alert( TB_PROMPT_DELETE_RECORD,; 
                   { _MENU_NO, _MENU_YES }; 
                    ) == 2
                dbdelete()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
            dbcommit()
            dbunlock()
        case lModify .and.; 
             lDelete .and.; 
             deleted()
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock())
                recall
                dbcommit()
                dbunlock()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
        otherwise
           // nil
        end
    otherwise
        do case
        case !lModify
            // no edit allowed
        case !alColCalc[oBrowse:colPos] .and. !lMemo
            // edit allowed
            keyboard chr(nKey) // repeated the key for the following TBDoGet
            if ( TBAlias(oBrowse, acCol, alColCalc) )->(rlock())
                bOldErrorHandler := errorblock( {|e| ErrorHandler(e)} )
                begin sequence
                    TBDoGet(oBrowse, acCol, lMemo, lAutosort)
                    dbcommit()
                recover
                    // nil
                end
                errorblock(bOldErrorHandler)
                dbunlock()
            else
                alert( TB_ERROR_RECORD_LOCKED )
            end
        otherwise
            // do nothing
        end
    end
   
    return NIL

//----------------------------------------------------------------
static function TBDoGet(oBrowse, acCol, lMemo, lAutosort)
//
// Do a GET for the current column in the browse.
//
   
    local lFlag := .T.
    local oCol
    local getlist
    local nKey
    local nLen
    local lAppend
    local bSavIns
    local nSavRecNo := recno()
    local xNewKey
    local xSavKey
    local cOldScreen
    local cOldColor
    local nSetCursor
    local cMemoOld
    local cMemoNew


    // If we're at EOF we're adding the first record, so turn on append mode
    if eof()
        lAppend := TB_APP_MODE_ON(oBrowse)
    else
        lAppend := TB_APP_MODE_ACTIVE(oBrowse)
    end
   
    // Make sure screen is fully updated, dbf position is correct, etc.
    oBrowse:forceStable()
    if ( lAppend .and. (recno() == lastrec() + 1 ) )
        // dbAppend()
        dbappend()
        // if neterror, ignore
        neterr()
    end
   
    // Save the current record's key value (or NIL)
    xSavKey := iif(empty(indexkey()), NIL, &(indexkey()))
   
    // Get the current column object from the browse
    oCol := oBrowse:getColumn(oBrowse:colPos)
    
    if lMemo

        cOldColor     := setcolor()   
        cOldScreen    := savescreen( maxrow()/2, 00, maxrow(), maxcol() )
        nSetCursor    := setcursor ( SC_NORMAL )
    
        scroll( maxrow()/2, 00, maxrow(), maxcol() )
        setcolor( COLOR_DEFAULT1 )
        devpos( maxrow()/2, 01)
        devout( padc( oCol:heading, maxcol()+1 ) )
        devpos( maxrow(), 01 )
        devout( padc( TB_WINDOW_BOTTOM_MEMO, maxcol()+1 ) )
        setcolor( cOldColor )
   
        cMemoOld := rtrim( &(acCol[oBrowse:colPos]) )
        cMemoNew := memoedit( cMemoOld,; 
                        maxrow()/2+1, 01, maxrow()-1, maxcol()-1,; 
                        .T., NIL, SPACE_LEN; 
                            )
        if ( cMemoOld <> cMemoNew )
            replace &(acCol[oBrowse:colPos]) with cMemoNew
            dbcommit()
        end
   
        // restore
        restscreen( maxrow()/2, 00, maxrow(), maxcol(), cOldScreen )
        setcursor ( nSetCursor )
    
    else
   
        // Get picture len to force scrolling if var is larger than window
        nLen := oBrowse:colWidth(oBrowse:colPos)
   
        // Create a corresponding GET
        getlist := { getnew( row(), col(),     ;
                       oCol:block,       ;
                       oCol:heading,     ;
                       oCol:picture,     ;
                       oBrowse:colorSpec ) }
   
        // Set insert key to toggle insert mode and cursor shape
        bSavIns := setkey(K_INS, { || InsToggle() })
   
        // Set initial cursor shape
        setcursor(iif(ReadInsert(), SC_INSERT, SC_NORMAL))
       
        readmodal(getlist)
        getlist := {}
    
        setcursor(SC_NONE)
        setkey(K_INS, bSavIns)

    end
    
    // Turn append mode off after each new record
    TB_APP_MODE_OFF(oBrowse)
   
    if lAutosort
        // Get the record's key value (or NIL) after the GET
        xNewKey := if(empty(indexkey()), NIL, &(indexkey()))
   
        oBrowse:inValidate()
        oBrowse:refreshAll():forceStable()
   
        // if the key has changed (or if this is a new record)
        if !(xNewKey == xSavKey) .or. (lAppend .and. xNewKey != NIL)
            // do a complete refresh
            oBrowse:refreshAll():forceStable()
            // Make sure we're still on the right record after stabilizing
            while &(indexkey()) > xNewKey .and. !oBrowse:hitTop()
                oBrowse:up():forceStable()
            end
        end
    end
   
    // Check exit key from get
    nKey := lastkey()
    if nKey == K_UP   .or.; 
       nKey == K_DOWN .or. ;
       nKey == K_PGUP .or.; 
       nKey == K_PGDN
        keyboard(chr(nKey))
    end
   
    return NIL

//----------------------------------------------------------------
static function InsToggle()   
//
// InsToggle()
//
// Toggle the global insert mode and the cursor shape.
//
//
    if readinsert()
        readinsert(FALSE)
        setcursor(SC_NORMAL)
    else
        readinsert(TRUE)
        setcursor(SC_INSERT)
    endif

    return NIL
//----------------------------------------------------------------
static function TBNew ( nTop, nLeft, nBottom, nRight,;
                        acCol, acColHead, alColCalc, alColMemo )

    local oBrowse
    local n
    local oColumn
    local cType
   
    // Start with a new browse object from TBrowseDB()
    oBrowse := TBrowseDB(nTop, nLeft, nBottom, nRight)

    // Add columns
    for n := 1 to len( acCol )
      
        // Make a new column 
        cType := valtype( &(acCol[n]) )  // speed up macro evaluation
        do case
        case alColCalc[n]  
            // Make a calculated field
            oColumn := TBAddCalCol( oBrowse, acColHead[n], acCol[n] )
            alColMemo[n] := .F.
        case ( cType == "M")  
            // Make a calculated field to activate the macro window
            oColumn := TBAddCalCol( oBrowse, acColHead[n],;
                           "padr(" + acCol[n] + ", 30 )" ) // shows 30 characters anyway
            alColMemo[n] := .T.
        case ( cType == "C") .and.;
             ( len( &(acCol[n]) ) > SPACE_LEN )
            // Make a calculated field to activate the macro window
            oColumn := TBAddCalCol( oBrowse, acColHead[n],; 
                          "left(" + acCol[n] + ", 30 )" ) // shows only 30 characters
            alColMemo[n] := .T.
        otherwise
            // Make a new normal field column
            oColumn := TBColumnNew( acColHead[n],;
                          TBFieldWBlock( acCol[n] ) ) 
            alColMemo[n] := .F.
        end
      
        // other attribute
        oColumn:headSep( "" )  // no head line separation
        oColumn:colSep( "" )
        oColumn:footSep( "" )  // no foot separation line
        oColumn:footing( "" )  // no foot
        
        oBrowse:addColumn( oColumn )


    next

    return oBrowse

//----------------------------------------------------------------
static function TBColPicture(oBrowse, acColSayPic)
//
// Set up pictures for the column.
//
   local n
   local oColumn
   default acColSayPic to NIL
   
   for n := 1 to oBrowse:colCount
      // Get (a reference to) the column
      oColumn := oBrowse:getColumn(n)
      // define picture
      oColumn:picture := acColSayPic[n]
   next
   
   return NIL

//----------------------------------------------------------------
static function TBFieldWBlock( cFieldName )
//   
// return the column code block
//
    local bFieldWBlock
    local cAlias
    local cField
    local nArrowStart
    local nSelect

    nArrowStart := at( "->", cFieldName )

    do case
    case nArrowStart == 0 // there is no arrow, so, no alias was specified.
        cAlias := NIL
        cField := alltrim( cFieldName )
    case nArrowStart > 0
        cAlias := substr( cFieldName, 1, nArrowStart-1 )
        cField := substr( cFieldName, nArrowStart+2 )
    end

    do case
    case cAlias == NIL
        bFieldWblock := fieldwblock( cField, select() )
    case upper(cAlias) == "FIELD"
        bFieldWblock := fieldwblock( cField, select() )
    case upper(cAlias) == "MEMVAR" .or. ;
        upper(cAlias) == "M"
        bFieldWblock := memvarblock( cField )
    otherwise
        nSelect := select(cAlias)
        bFieldWblock := fieldwblock( cField, nSelect )
    end
   
    return bFieldWBlock

//----------------------------------------------------------------
static function TBAlias( oBrowse, acCol, alColCalc )
//   
// return the alias name for the actual column
//
    local cAlias
    local nArrowStart

    nArrowStart := at( "->", acCol[oBrowse:colPos] )

    do case
    case alColCalc[oBrowse:colPos]
        cAlias := alias()
    case nArrowStart == 0 // there is no arrow, so, no alias was specified.
        cAlias := alias()
    case nArrowStart > 0
        cAlias := substr( acCol[oBrowse:colPos], 1, nArrowStart-1 )
    end

    return alltrim(cAlias)

//----------------------------------------------------------------
static function TBAddCalCol(oBrowse, cColHead, cCol)
//   
// create a calculated column
//
    local oColumn
   
    oColumn := TBColumnNew(cColHead, { || &(cCol) } ) // evaluate the macro each time
   
    return oColumn

//----------------------------------------------------------------
static function TBDefault( acCol, acColSayPic, alColCalc, acColHead )
//

    if ( alias() == "" )
         return NIL
    end
   
    // prepare arrays
    acCol       := {}
    acColSayPic := {}
    alColCalc   := {}
    acColHead   := {}

    // the first default column
    aadd ( acCol, "if( deleted(), '*', ' ') +" +;
                        " str(recno(), 5, 0)" )
    aadd( alColCalc, .T. )
    aadd( acColSayPic, "!!!!!!" )
    
    aadd( acColHead, "Rec. #" )
   
    // add the other columns
    TBDefColAdd( @acCol, @acColSayPic, @alColCalc, @acColHead )

    return NIL

//----------------------------------------------------------------
static function TBDefColAdd( acCol, acColSayPic, alColCalc,; 
                    acColHead )  
//
   
    local aStruct   := {}  
    local nFields   := 0  
    local nColIndex := 0   
    local nRelations := 0  
    local nIndRel := 0    
    local cChildAlias := ""

    default acCol        to {}
    default acColSayPic  to {} 
    default alColCalc    to {}
    default acColHead    to {}

    // determiniamo il numero di relazioni
    nRelations := 0
    while .T.
        nRelations++
        if dbrselect( nRelations ) > 0
            // prosegue
        else
            // deve decrementare nRelations perch l'ultimo non  vero
            nRelations--
            exit
        end
    end
    // adesso nRelations contiene il numero di relazioni con il file principale
   
    // si analizza il file di cui si dovr fare il browse
    aStruct := dbstruct()
    nFields := len( aStruct )

    // si compilano i vettori
    nColIndex := 1
    while nColIndex <= nFields
        aadd( acCol, Alias()+"->"+aStruct[nColIndex, 1] )
        aadd( alColCalc, .F. )
        aadd( acColSayPic, TBDefPicture( aStruct[nColIndex, 1],; 
                                         aStruct[nColIndex, 2],; 
                                         aStruct[nColIndex, 3],; 
                                         aStruct[nColIndex, 4] ) )
        aadd( acColHead, Alias()+"->"+aStruct[nColIndex, 1] )
        nColIndex++
    end
   
    // it adds eventual relations
    if nRelations > 0
        for nIndRel := 1 to nRelations
   
            // it determinates the Alias name for this relation
            cChildAlias := alias( dbrselect( nIndRel ) )

            // it adds the columns from the related Alias
            (cChildAlias)->(TBDefColAdd( @acCol, @acColSayPic,; 
                @alColCalc, @acColHead );
                           )
        next
    end

    return NIL

//----------------------------------------------------------------
static function TBDefPicture( cColumn, cColumnType, nColumnLen,; 
                    nColumnDec ) 

   local nLen := 0
   local cColSayPic := ""

   do case
   case cColumnType == "C"
      nLen := len( &(cColumn) )
      if nLen > 40
         cColSayPic := "@s40"
      else
         cColSayPic := replicate( "x", nLen )
      end
   case cColumnType == "N"
      if nColumnDec > 0
         // con decimale e virgola
         cColSayPic := ;
            replicate ( "9", nColumnLen-1-nColumnDec ) +;
            "." +;
            replicate ( "9", nColumnDec )
      else
         // intero
         cColSayPic := ;
            replicate ( "9", nColumnLen )
      end
   case cColumnType == "D"
      cColSayPic := "99/99/9999"
   case cColumnType == "L"
      cColSayPic := "L"
   case cColumnType == "M"
      cColSayPic := "@s40" // <Memo>
   end

   return cColSayPic

//================================================================
// RPT - FORMATTED TEXT PRINT
//================================================================
function rptMany( cText, bWhileCondition, bForCondition, cSymbol )
//
// rptMany( <cText>, [<bWhileCondition>], [<bForCondition>],
//          [<cSymbol>] ) --> NIL
//
// Prints with the nB report standard the text contained into
// <cText> for every record in the active Alias matching the
// conditions.
//
// <cText>              the text to print.
// <bWhileCondition>    code block WHILE condition for record
//                      to be included.
// <bForCondition>      code block FOR condition for record
//                      to be included.
// <cSymbol>            the symbol for print commands contained
//                      into the text file.
//

    default bWhileCondition to {|| .T. }
    default bForCondition   to {|| .T. }
    default cSymbol         to RPT_DEFAULT_COMMAND_SYMBOL 

    dbgotop()
    while !eof() .and. eval(bWhileCondition)
        WaitFileEval()
        if eval(bForCondition)
            RPT( cText, NIL, NIL, NIL, SPACE_LEN, NIL, NIL, cSymbol )    
        end
        dbskip()
    end
    dbgotop()

    return NIL

//================================================================
function rpt( cText, nTotalLines, lMain, lHeadFoot,; 
              nLineLength, nTabSize, lWrap, cSymbol )
//
// rpt( <cText>, [<nTotalLines>], [<lMain>], [<lHeadFoot>], 
//      [<nLineLength>], [<nTabSize>], [<lWrap>], [<cSymbol>] )
//      --> NIL
//
// Prints with the nB report standard the text contained into
// <cText> once.
//
// As this function is recursive, it is important to transfer the
// exact line dimention to the recursive call.
//
// <cText>              the text to be printed.
// <nTotalLines>        <cText> lines length, used with recursive
//                      call.
// <lMain>              true (.t.) means that it is the first
//                      call and not recursive. True is the
//                      default value.
// <lHeadFoot>          true (.t.) means that <cText> is a header 
//                      or footer.
// <nLineLength>        the maximal line length.
// <nTabSize>           tab length.
// <lWrap>              true (.t.) means that the text will be
//                      wrapped if necessary.
// <cSymbol>            the prefix simbol for report commands.
//

    local nLine := 0
    local cLine := ""
   
    local cSubText := ""
    local nSubTextLines := 0
    local cCommand := ""
    local cCondition := ""
    local nParameter := 0
    local cParameter := ""
    local xParameter
    local nNeeded  := 0

    local nOldRow    := row()
    local nOldCol    := col()

    static nLPP
    static nPrintedLine
    static cHeader
    static nHeadLines
    static cFooter
    static nFootLines
    static cLeft
    static lNewPage
    
    default nTotalLines  to mlcount( cText, nLineLength, nTabSize, lWrap )
    default lMain        to .T. // chiamata principale per l'intero documento
    default lHeadFoot    to .F. // se .T. indica che si sta stampando un head o foot
    default nLineLength  to SPACE_LEN
    default cSymbol      to RPT_DEFAULT_COMMAND_SYMBOL 
    
    begin sequence
    
        // if it is the first call
        if lMain == .T. .or. lMain == NIL
            // static variable initialization
            nLPP         := NIL
            nPrintedLine := NIL
            cHeader      := NIL
            nHeadLines   := NIL
            cFooter      := NIL
            nFootLines   := NIL
            cLeft        := NIL
            lNewPage     := NIL
            
            // public variable creation
            public n_Lines
        end
   
        // static defaults
        if nLPP == NIL
            nLPP := RPT_DEFAULT_LPP
        end
        if nPrintedLine == NIL
            nPrintedLine := 0
        end
        if cHeader == NIL
            cHeader := ""
        end
        if nHeadLines == NIL
            nHeadLines := 0
        end
        if cFooter == NIL
            cFooter := ""
        end
        if nFootLines == NIL
            nFootLines := 0
        end
        if cLeft == NIL
            cLeft := ""
        end
        if lNewPage   == NIL
            lNewPage := .T.
        end

        // if there are no more lines to print
        if nTotalLines <= 0
            break
        end
   
        // if first call, change row() col() position.
        if lMain
           setpos( maxrow(), maxcol()+1 )    // beiond the end 
        end
        
        // main loop
        for nLine = 1 to nTotalLines

            // line conuter correction.
            if nPrintedLine >= nLPP
                nPrintedLine := 0
            end

            // remaining lines calculation.
            Memvar->n_Lines := nLPP - nFootLines - nPrintedLines
      
            // next line
            cLine := memoline( cText, nLineLength, nLine, nTabSize, lWrap )    
      
            // line analisys
            do case
            case upper( alltrim(cLine) ) = cSymbol+"REM"
                // comment, the line is ignored
            case upper( alltrim(cLine) ) = cSymbol+"COMMENT"
                // comment, the line is ignored
            case upper( alltrim(cLine) ) = cSymbol+"END" 
                // if it appears here, it is an error, the line is ignored.
                alert( RPT_ERROR_PRINT_ISOLATED_END )
            case upper( alltrim(cLine) ) = cSymbol+"COMMAND"
                // Start a command line.
                // The command lines are isolated.
                cCommand := rptReadToEnd( cText, @nLine, NIL, nLineLength, nTabSize, lWrap, cSymbol ) 
                EXecute( cCommand, "REPORT COMMAND")
            case upper( alltrim(cLine) ) = cSymbol+"IF"
                cCondition := alltrim( substr( ltrim(cLine), 3+len(cSymbol) ) )
                cSubText := rptReadToEnd( cText, @nLine, @nSubTextLines, nLineLength, nTabSize, lWrap, cSymbol )       
                if cCondition == ""
                    // no condition
                    alert( alltrim(cLine) + ";" +;
                         RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // line ignored
                else
                    if &(cCondition)
                        RPT( cSubText, nSubTextLines, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"WHILE"
                cCondition := alltrim( substr( ltrim(cLine), 6+len(cSymbol) ) )
                cSubText := rptReadToEnd( cText, @nLine, @nSubTextLines, nLineLength, nTabSize, lWrap, cSymbol )       
                if cCondition == ""
                    // no condition
                else
                    while &(cCondition)
                        RPT( cSubText, nSubTextLines, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"DBSKIP"
                // dbskip()
                // the eof() and bof() check MUST be done inside the document!!!
                cParameter := alltrim( substr( ltrim(cLine), 7+len(cSymbol) ) )
                if cParameter == ""
                    dbskip(+1)
                else
                    nParameter := &(cParameter)
                    dbskip( nParameter )
                end
            case upper( alltrim(cLine) ) = cSymbol+"LPP" 
                // lines per page
                cParameter := alltrim( substr( ltrim(cLine), 4+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + ";" +;
                         RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // line ignored
                else
                    nParameter := &(cParameter)
                    nLPP := nParameter
                end
            case upper( alltrim(cLine) ) = cSymbol+"PA"
                if (nPrintedLine > nHeadLines .and. nPrintedLine < nLPP)   //  in mezzo ad una pagina 
                    // allora deve stampare il footer
                    while nPrintedLine < (nLPP - nFootLines)
                        // stampa una riga vuota alla volta
                        // abilitando la stampa di head-foot
                        RPT( NL(1), 1, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                    // il footer dovrebbe essere stampato automaticamente
                end
            case upper( alltrim(cLine) ) = cSymbol+"INSERT"
                cParameter := alltrim( substr( ltrim(cLine), 7+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + ";" +;
                         RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // la riga viene di fatto ignorata
                else
                    // se esiste il file completo di estensione viene incluso
                    if file( cParameter )
                        RPT( memoread(cParameter), NIL, .F., NIL, SPACE_LEN, NIL, NIL, cSymbol )
                    else
                        alert( alltrim(cLine) + ";" +;
                            RPT_ERROR_PRINT_FILE_NOT_FOUND ) 
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"NEED"
                cParameter := alltrim( substr( ltrim(cLine), 5+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + ";" +;
                        RPT_ERROR_PRINT_COMMAND_UNCOMPLETED ) 
                    // la riga viene di fatto ignorata
                else
                    nParameter := &(cParameter)
                    nNeeded := nParameter
                    if nPrintedLine + nNeeded > nLPP - nFootLines  // se uguale va ancora bene
                        // salta pagina in un modo un po' strano
                        RPT( cSymbol+"PA", 1, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
                    end
                end
            case upper( alltrim(cLine) ) = cSymbol+"HEAD"  
                cHeader := rptReadToEnd( cText, @nLine, @nHeadLines, nLineLength, nTabSize, lWrap, cSymbol ) 
            case upper( alltrim(cLine) ) = cSymbol+"FOOT"
                cFooter := rptReadToEnd( cText, @nLine, @nFootLines, nLineLength, nTabSize, lWrap, cSymbol ) 
            case upper( alltrim(cLine) ) = cSymbol+"LEFT"
                cParameter := alltrim( substr( ltrim(cLine), 5+len(cSymbol) ) )
                if cParameter == ""
                    alert( alltrim(cLine) + ";" +;
                        RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                    // the line is ignored
                else   
                    xParameter := &(cParameter)
                    do case
                        case valtype(xParameter) == "N" .and.;
                            xParameter >= 0
                            cLeft := space(xParameter)
                        case valtype(xParameter) == "C"
                            cLeft := xParameter
                         otherwise
                        alert( alltrim(cLine) + ";" +;
                            RPT_ERROR_PRINT_COMMAND_UNCOMPLETED )
                        // the line is ignored
                    end
                end
            otherwise
                // Normal line to be printed.
                // Check if the head must be printed.
                if (nPrintedLine == 0) .and.;
                    !lHeadFoot
                    RPT( cHeader, nHeadLines, .F., .T., nLineLength, nTabSize, lWrap, cSymbol )
                end
                // It prints the line deleting extra characters on the right.
                // The extra space (the line is ever SPACE_LEN long) will 
                // produce extra white lines on the printed text.
                qqout( rtrim( cLeft + RPTLineTrans(cLine) ) )
                qout("")
                nPrintedLine++
                // It checks if the footer must be printed.
                if (nPrintedLine >= ( nLPP - nFootLines ) ) .and.;  
                   (lHeadFoot == .F.)
                 
                    RPT( cFooter, nFootLines, .F., .T., nLineLength, nTabSize, lWrap, cSymbol )
                    // After a footer there is ever a eject.
                    // The Eject command works only with printers
                    qqout( FF )  // form feed - chr(12)
                end
            end
   
        next

        if lMain .and.;    // main call
           (nPrintedLine > 0 .and. nPrintedLine < nLPP)   // in the middle of a page
            // Then the footer must be printed.
            while nPrintedLine < (nLPP - nFootLines)
                // It prints an empty line.
                RPT( NL(1), 1, .F., .F., nLineLength, nTabSize, lWrap, cSymbol )
            end
            // The footer should be automatically printed.
        end

    end sequence
    
    if lMain
        release c_Lines
        devpos( nOldRow, nOldCol )
    end
    
    return NIL

//----------------------------------------------------------------
static function rptReadToEnd( cText, nLine, nLines,; 
                              nLineLength, nTabSize,; 
                              lWrap, cSymbol )
//
// --> cBlock
//
// Returns a block of lines from <cText>.
// The block is delimited with:
//      *...Command
//              ...
//              ...
//      *end
//

    local cLine  := ""
    local cBlock := ""
    local nStartLine := nLine
    local nSubLines := 0
    local lStart    := .T.

    while .T.
        nLine++
        cLine := memoline( cText, nLineLength, nLine, nTabSize, lWrap  )
        do case
        case cLine == "" // end of block
            exit
        case upper( alltrim(cLine) ) = cSymbol+"END"
            exit
        case upper( alltrim(cLine) ) = cSymbol+"COMMAND" .or.;
             upper( alltrim(cLine) ) = cSymbol+"IF" .or.; 
             upper( alltrim(cLine) ) = cSymbol+"WHILE" .or.; 
             upper( alltrim(cLine) ) = cSymbol+"HEAD" .or.; 
             upper( alltrim(cLine) ) = cSymbol+"FOOT" 
            // The next *END must be jumped as it don't belongs
            // to this level.
            if lStart 
                // just started
                cBlock += rtrim(cLine)
            else
                cBlock += NL(1) + rtrim(cLine)
            end
            cBlock += NL(1) +;
                      rptReadToEnd( cText, @nLine, @nSubLines, nLineLength, nTabSize, lWrap, cSymbol ) +;
                      NL(1) + cSymbol+"END"
        otherwise
            if lStart 
                // Just started.
                cBlock += rtrim(cLine)
            else
                cBlock += NL(1) + rtrim(cLine)
            end
        end
        lStart := .F.
    end

    // At now, nLine is pinted to *END
   
    // nLines must count only the content not the commands: *... and *end.
    nLines := (nLine - nStartLine) -1

    // nLine and nLines are sent back
    return cBlock

//----------------------------------------------------------------
static function rptLineTrans( cLine ) 
//
// --> cTranslatedLine
//
// Translates the <cLine> content replacing what is into
//  and , using Memvars and Fields.
//
// This function is recursive.
//

    local bSaveErrorHandler
    local bSav2ErrorHandler
    local nPosLeft   
    local nPosRight
    local cLineLeft
    local cLineCenter
    local cLineRight
    local cString
    local xContent

   
    nPosLeft := at( "", cLine )
    nPosRight := at( "", cLine )

    // if the line contains variables.
    do case
    case nPosLeft == 0 .and.;
         nPosRight == 0
        return cLine
    case nPosLeft == 0 .or.;    // se arriviamo qui non sono tutti e due uguali a zero
         nPosRight == 0
        alert( RPT_ERROR_PARENTESES )
        return cLine
    end

    // A variable string exists.
    cString := substr( cLine, nPosLeft+1 )
    cString := substr( cString, 1, at( "", cString ) - 1 )
    cString := alltrim( cString )

    // Left side determination.
    cLineLeft := substr( cLine, 1, nPosLeft-1 )

    // Variable transfer.
    bSaveErrorHandler := errorblock( {|| break(NIL)} )
    begin sequence
        xContent := &(cString) 
    recover
        // se il contenuto di cString non viene riconosciuto
        // genera un errore.
        bSaveErrorHandler := errorblock( {|| break(NIL)} )  // sta bene cos
        begin sequence
            // maybe, cString is a Memvar not jet defined.
            public &cString. := SPACE_STR
            &cString. := alltrim( Accept( &cString., "Insert " + cString ) )
            xContent := alltrim( &cString. )
        recover
            // But it wasn't a new variable.
            // It is written to help to find the mistake.
            xContent := "" + cString + ""
        end
        errorblock(bSav2ErrorHandler)

    end
    errorblock(bSaveErrorHandler)

    do case
    case valtype(xContent) == "C"
        cLineCenter := xContent     // senza alltrim
    case valtype(xContent) == "N"
        cLineCenter := alltrim( str( xContent ) )
    case valtype(xContent) == "D"
        cLineCenter := dtoc( xContent )
    case valtype(xContent) == "M"
        cLineCenter := xContent     // sentra trim
    end

    // Right extraction with recursion.
    cLineRight := rptLineTrans( substr( cLine, nPosRight+1 ) )

    return cLineLeft + cLineCenter + cLineRight

//================================================================
// EX - MACRO INTERPRETER AND EXECUTOR
//================================================================
function ex( cFileMacro )
//
// ex( <cFileMacro> ) --> nExitCode
//
// execute <cFileMacro>.
//
// <cFileMacro> The macro filename with extension.
//
//

    local cProcedure

    if file( cFileMacro )
        cProcedure := memoread( cFileMacro )
    else
        alert( cFileMacro + ";" +;
                EX_ERROR_MACRO_FILE_NOT_FOUND )
        return EX_EXIT_NO_MACRO_FILE
    end
   
    // start execution and terminate
    return EXecute( cProcedure, cFileMacro )

//----------------------------------------------------------------
static function execute( cCommands, cName ) 
//
// EXecute( <cCommands>, [<cName>] ) --> nExitCode
//
// <cCommands>  Variable containing many commands separated
//              by NL(1) ( = CR+LF).
//
// <cName>      A name to be used for error report duriing program
//              execution.
//

    local aNest         := array( EX_MAX_NEST )
    local nNest         := 0
    local nLine
    local cCommand      := ""
    local cCondition    := ""
    
    local nExitCode     := EX_EXIT_NORMAL

    begin sequence

        nLine := 1
        nNest := 1
        aNest[nNest] := {"MAIN", .T. }
        // [1]MAIN  [2]lContinue
      
        while .T.

            cCommand := EXCommandExtract( cCommands, @nLine ) 
            
            // line analisys
            
            if cCommand == NIL  // end of file
                nExitCode := EX_EXIT_NORMAL
                break
            end

            do case
            case upper( cCommand ) == "END"
                // close one nest level
                do case
                case aNest[nNest][1] == "MAIN"
                    // not allowed
                    if alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;
                           EX_ERROR_END_ALONE,; 
                               { EX_MENU_CHOICE_CONTINUE,; 
                                 EX_MENU_CHOICE_ABORT } ) == 2
                        nExitCode := EX_EXIT_BREAK
                        break
                    else
                        nLine++
                    end
                
                case aNest[nNest][1] == "IF" .or.;
                     aNest[nNest][1] == "THEN" .or.;
                     aNest[nNest][1] == "ELSE"
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    nLine++
                case aNest[nNest][1] == "BEGIN SEQUENCE" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    nLine++
                case aNest[nNest][1] == "WHILE"
                    if &(aNest[nNest][2])
                        // continue - go back
                        nLine := aNest[nNest][3]
                    else
                        // ok - end
                        aNest[nNest] := NIL
                        nNest--
                        nLine++
                    end
                
                case aNest[nNest][1] == "LOOP" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    // Previous level
                    // nLine remain on last END
                    do case
                    case aNest[nNest][1] == "WHILE"
                        if &(aNest[nNest][2])
                            // repeat loop
                            nLine := aNest[nNest][3]
                        else
                            // go to next END starting from next line
                            exJumpIt( cCommands, @nLine,; 
                                {|cLine| cLine=="END"} )
                            // stay, it will end naturally
                        end
                    otherwise
                        aNest[nNest][1] := "LOOP"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                
                case aNest[nNest][1] == "BREAK" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    // nLine remain on last END
                    // previous level
                    do case
                    case aNest[nNest][1] == "BEGIN SEQUENCE"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    otherwise
                        aNest[nNest][1] := "BREAK"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                case aNest[nNest][1] == "EXIT" 
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    // nLine remain on last END
                    // previous level
                    do case
                    case aNest[nNest][1] == "WHILE"  
                        // go to next END starting from next line
                        aNest[nNest][2] := ".F."
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    otherwise
                        aNest[nNest][1] := "BREAK"
                        // go to next END starting from next line
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                case aNest[nNest][1] == "DO CASE" .or.;
                     aNest[nNest][1] == "ENDCASE"
                    // ok
                    aNest[nNest] := NIL
                    nNest--
                    nLine++
                end
            
            case upper( cCommand ) == "BEGIN SEQUENCE"  
                nNest++
                aNest[nNest] := {"BEGIN SEQUENCE"}
                // [1]BEGIN SEQUENCE
                nLine++
            case upper( cCommand ) == "BREAK"
                do case
                case aNest[nNest][1] == "BEGIN SEQUENCE"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                otherwise
                    aNest[nNest][1] := "BREAK"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "IF"  // no condition following
                // no condition
                alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;   
                    EX_ERROR_NO_CONDITION )
                exJumpIt( cCommands, @nLine,; 
                    {|cLine| cLine=="END"} )
                nLine ++
            case upper( left( cCommand, 3 ) ) == "IF "
                nNest++
                aNest[nNest] := {"IF" }
                // [1]IF|THEN|ELSE
                cCondition := alltrim( substr( ltrim(cCommand), 3 ) )
                if &(cCondition)
                    aNest[nNest][1] := "THEN"
                    nLine++
                else
                    aNest[nNest][1] := "ELSE"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="ELSE"}, {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "ELSE"
                if aNest[nNest][1] == "ELSE" // THEN wasn't executed
                    // ok execute ELSE
                    nLine++
                else
                    // THEN was executed and it must jumped
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "DO CASE"  
                nNest++
                aNest[nNest] := {"DO CASE"}
                // [1]DO CASE|ENDCASE
                nLine++
            case upper( cCommand ) == "CASE"  // no condition following
                // no condition
                alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;   
                    EX_ERROR_NO_CONDITION )
                exJumpIt( cCommands, @nLine,; 
                    {|cLine| left(cLine,5)=="CASE "},; 
                    {|cLine| cLine=="OTHERWISE"},; 
                    {|cLine| cLine=="END"} )
            case upper( left( cCommand, 5 ) ) == "CASE "
                do case
                case aNest[nNest][1] == "DO CASE" 
                    // ok try it
                    cCondition := alltrim( substr( ltrim(cCommand), 5 ) )
                    if &(cCondition)
                        aNest[nNest][1] := "ENDCASE"   // no more case
                        nLine++
                    else
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| left(cLine,5)=="CASE "},; 
                            {|cLine| cLine=="OTHERWISE"},; 
                            {|cLine| cLine=="END"} )
                    end
                otherwise // aNest[nNest][1] == "ENDCASE"
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "OTHERWISE"
                do case
                case aNest[nNest][1] == "DO CASE" 
                    // ok do it
                    nLine++
                otherwise // aNest[nNest][1] == "ENDCASE"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "WHILE" // no condition
                // no condition
                alert( cName + "(" + ltrim(str(nLine)) + ")" +";" +;   
                    EX_ERROR_NO_CONDITION )
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                nLine++
            case upper( left( cCommand, 6 ) ) == "WHILE "
                nNest++
                aNest[nNest] := {"WHILE", NIL, NIL }
                // [1]WHILE  [2]cCondition  [3]nLine1stStatement
                // save condition
                aNest[nNest][2] := alltrim( substr( ltrim(cCommand), 6 ) )
                // Save the first statement begin point (it is better).
                aNest[nNest][3] := nLine +1 
                // Check condition
                if &(aNest[nNest][2])
                    // ok, go on
                    nLine++
                else
                    aNest[nNest][2] := ".F."
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "LOOP"
                do case
                case aNest[nNest][1] == "WHILE"
                    if &(aNest[nNest][2])
                        // ok - loop
                        nLine := aNest[nNest][3] 
                    else   
                        // no - exit
                        aNest[nNest][2] := ".F."
                        exJumpIt( cCommands, @nLine,; 
                            {|cLine| cLine=="END"} )
                    end
                otherwise
                    aNest[nNest][1] := "LOOP"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end
            case upper( cCommand ) == "EXIT" 
                do case
                case aNest[nNest][1] == "WHILE"
                    aNest[nNest][2] := ".F."
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                otherwise
                    aNest[nNest][1] := "EXIT"
                    exJumpIt( cCommands, @nLine,; 
                        {|cLine| cLine=="END"} )
                end

            otherwise
                // If we are here, it must be a function
                nExitCode := exOneCommand( cCommand, cName, nLine )
                if nExitCode == EX_EXIT_BREAK
                    break
                end
                nLine++
            end
        
        end
    
    end sequence

    return nExitCode
   
//----------------------------------------------------------------
static function exCommandExtract( cCommands, nLine )
//
// exCommandExtract( <cCommands>, @<nLine> ) --> NIL
//                                           --> cCommand
//
// It returns one complete command or NIL if end of file.
//
// <cCommands>  The variable containing the commands.
//
// <nLine>      The actual line position.
//
   
    local cCommand      := ""
    local nSemicolonPos := 0
   
    // reach next not-empty line.
    while .T.
        cCommand := memoline( cCommands, SPACE_LEN, nLine )
        do case
        case cCommand == ""
            // End of file reached
            return NIL 
        case !( empty( cCommand ) )
            // Cut eventual comment.
            cCommand := EXCommentCut( cCommand )
            // Cut unusefull blanks.
            cCommand := alltrim( cCommand )
            // Check if there is something.
            if !( cCommand == "" )
                // Not empty
                exit
            end
            // Otherwise repeat loop.
        end
        // Loop.
        nLine++
    end

    // At this point there is a command line
    
    // Semicolon
    if right( cCommand, 1 ) == ";"
        // it follows to the next line
        cCommand := left( cCommand, len( cCommand )-1 )
        nLine++
        // recursion
        cCommand += EXCommandExtract( cCommands, @nLine )
    end

    return cCommand

//----------------------------------------------------------------
static function exCommentCut( cCommandLine )
//
// exCommentCut( <cCommandLine> ) --> cCommandLine
//
// The function cuts the command line at the first
// occurence of "//" that is used as only comment
// indicator.
//
// The string "//" cannot be used for other purpose
// as the command line will be truncated there.
//
// <cCommandLine>       a line to be cutted.
//

    local nCommentPos

    nCommentPos := at( "//", cCommandLine )
    if nCommentPos > 0
        cCommandLine := substr( cCommandLine, 1, nCommentPos -1 )
    end
   
    return cCommandLine

//----------------------------------------------------------------
static function exOneCommand( cCommand, cName, nLine )
//
// exOneCommand( <cCommand>, <cName>, <nLine> ) --> nExitCode
//
// This function execute a command after cleaning.
//
// <cCommand>   The command to execute.
//
// <cName>      The procedure name to be used for error
//              documentation.
//
// <nLine>      Actual command line to be used for error
//              documentation.
//

    local bSaveErrorHandler
    local xResult
    local nExitCode := EX_EXIT_NORMAL

    // le righe vuote non devono essere eseguite
    if cCommand == ""
        return EX_EXIT_NORMAL
    end
                                          
    bSaveErrorHandler := errorblock( {|e| ErrorHandler( e, cName, nLine, cCommand )} )
    begin sequence
        xResult := &(cCommand)
    recover
        if alert( cName + " line: " + ltrim( str( nLine ) ) + ";" +;
                cCommand + ";;" +;
                EX_PROMPT_MACRO_EXECUTION_ABORT,; 
                    { EX_MENU_CHOICE_ABORT, EX_MENU_CHOICE_CONTINUE } ) == 1
            nExitCode := EX_EXIT_BREAK
        end
    end
    errorblock(bSaveErrorHandler)

    return nExitCode

//----------------------------------------------------------------
static function exJumpIt( cCommands, nLine,; 
                          bExit1, bExit2, bExit3 )
//
// --> NIL
//
// This function jumps to the statements whitch are not to be 
// executed according with the control structure used.
//

    local cLine
    local nLevel := 1
   
    default bExit1 to { | cLine | cLine == "END" }
    default bExit2 to bExit1
    default bExit3 to bExit2

    begin sequence

        while .T.
            nLine++
            cLine := EXCommandExtract( cCommands, @nLine )
            if cLine == NIL
                alert( EX_ERROR_NO_END )
                break
            end
            cLine := upper( cLine )
            
            if nLevel == 1 // start level
                if eval( bExit1, cLine ) .or.;
                   eval( bExit2, cLine ) .or.;
                   eval( bExit3, cLine )
                    break
                end
            end

            do case
            case cLine == "END"
                nLevel--
                if nLevel == 0
                    alert( EX_ERROR_TOO_MUCH_END )
                    break
                end
            case cLine            == "WHILE"            .or.;  // without condition
                 left( cLine, 6 ) == "WHILE "           .or.;  // with condition
                 cLine            == "IF"               .or.;  // without condition
                 left( cLine, 3 ) == "IF "              .or.;  // with condition
                 cLine            == "BEGIN SEQUENCE"   .or.;
                 cLine            == "DO CASE"

                nLevel++
            end
        end
    
    end sequence

    // nLine is pointed on END or beiond end of file
   
    return NIL

//================================================================
// DTE - DATE/TIME FUNCTIONS
//================================================================
function dteMonth( nMonth, cLanguage )
//
// dteMonth( <nMonth>, <cLanguage> ) --> cMonth
//
// <nMonth>     the month number
// <cLanguage>  the language name
//
// This function translates the <nMonth> number into the month
// name.
//
    local cMonth := ""

    begin sequence

        if valtype( nMonth ) <> "N"
            break
        end
        if valtype( cLanguage ) <> "C" 
            break
        end

        cLanguage := upper( alltrim( cLanguage ) )

        do case
        case cLanguage == "ITALIANO" .or.;
             cLanguage == "ITALIANA" .or.;
             cLanguage == "ITALIA"
            do case
            case nMonth = 1
                cMonth := "gennaio"
            case nMonth = 2
                cMonth := "febbraio"
            case nMonth = 3
                cMonth := "marzo"
            case nMonth = 4
                cMonth := "aprile"
            case nMonth = 5
                cMonth := "maggio"
            case nMonth = 6
                cMonth := "giugno"
            case nMonth = 7
                cMonth := "luglio"
            case nMonth = 8
                cMonth := "agosto"
            case nMonth = 9
                cMonth := "settembre"
            case nMonth = 10
                cMonth := "ottobre"
            case nMonth = 11
                cMonth := "novembre"
            case nMonth = 12
                cMonth := "dicembre"
            end
        end

    end sequence

    return cMonth

//================================================================
function dteWeek( nWeek, cLanguage )
//
// dteWeek( <nWeek>, <cLanguage> ) --> cWeek
//
// <nWeek>      the week number
// <cLanguage>  the language name
//
// This function translates the <nWeek> number into the week
// name.
// nWeek == 1 -> Sunday
// nWeek == 2 -> Monday
// ...
// nWeek == 7 -> Saturday
// 
    local cWeek := ""

    begin sequence

        if valtype( nWeek ) <> "N" 
            break
        end
        if valtype( cLanguage ) <> "C" 
            break
        end

        cLanguage := upper( alltrim( cLanguage ) )
    
        do case
        case cLanguage == "ITALIANO" .or.;
             cLanguage == "ITALIANA" .or.;
             cLanguage == "ITALIA"
            do case
            case nWeek = 1
                cWeek := "domenica"
            case nWeek = 2
                cWeek := "lunedi'"
            case nWeek = 3
                cWeek := "martedi'"
            case nWeek = 4
                cWeek := "mercoledi'"
            case nWeek = 5
                cWeek := "giovedi'"
            case nWeek = 6
                cWeek := "venerdi'"
            case nWeek = 7
                cWeek := "sabato"
            end
        end
    
    end sequence
  
    return cWeek
  
//================================================================
// DOC - TEXT EDITOR
//================================================================
function Doc( cFileName )
//
// Doc( [<cFileName>] ) --> NIL
//
// Text editor for small files, less than 64K.
//
// <cFileName>  the text file name to modify.
//

    local bOldErrorHandler
    local cOldScreen    := savescreen()
    local cOldColor     := setcolor()   
    local nOldCursor    := setcursor( SC_NONE ) 
    local bOld_F1       := setkey( K_F1, NIL ) 
    local bOld_F2       := setkey( K_F2, NIL ) 
    local nOldRow       := row()
    local nOldCol       := col()

    local cOldName
    local lNew
    
    begin sequence

        setcursor( SC_NORMAL )
        set(_SET_BELL, .f.)
        set(_SET_SCOREBOARD, .f.)
        setkey( K_F2, {|| keyboard("" + fieldPicture() + "") } )
      
        if valtype( cFileName ) <> "C"
            cFileName := strAddExtention( _UNTITLED, _EXTENTION_TEXT )
            cOldName := NIL
            lNew := .T.
        else
            cFileName := upper( strAddExtention( cFileName, _EXTENTION_TEXT) )
            if file( cFileName )
                cOldName := cFileName
            else
                cOldName := NIL
                lNew := .T.
            end
        end

        // start the editing
        docEditor( cFileName, cOldName, lNew )

    end sequence

    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( NIL, NIL, NIL, NIL, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function docEditor( cFile, cOld, lNew )

    local nKey

    setkey( K_F1, { || Text( DOC_HELP ) } )
   
    // compilazione dell'array aEdit 
    aEdit                := Array(DOC_LENGTH)
    aEdit[DOC_NEW]       := lNew   
    aEdit[DOC_OLD]       := cOld   // eventualmente gi maiuscolo e alltrim()
    aEdit[DOC_NAME]      := cFile  // eventualmente gi maiuscolo e alltrim()
    do case
    case aEdit[DOC_NEW]
        aEdit[DOC_TEXT]      := ""
    case file( aEdit[DOC_NAME] )
        aEdit[DOC_TEXT]      := MemoRead(cFile)
    otherwise
        aEdit[DOC_TEXT]      := ""
    end

    aEdit[DOC_WRAP]      := .T.
    aEdit[DOC_INS]       := Set(_SET_INSERT)

    aEdit[DOC_ROW]       := 1
    aEdit[DOC_COL]       := 0
    aEdit[DOC_RELROW]    := 0
    aEdit[DOC_RELCOL]    := 0

    aEdit[DOC_CHANGED]   := .F.
    aEdit[DOC_LASTKEY]   := 0

    aEdit[DOC_EXIT]      := .F.

    // clear screen
    scroll()
   
    setcolor( COLOR_DEFAULT1 )
    scroll( 00,00,00,maxcol() )
    @00,05 say alltrim( cFile )
    scroll( maxrow(),00,maxrow(),maxcol() )
    @maxrow(),00 say padc( _KEY_ESC_EXIT_F1_F10, maxcol()+1 )
    setcolor( COLOR_DEFAULT0 )
   
    // editing
    while (!aEdit[DOC_EXIT])
      
        docDoEditing()
        nKey := aEdit[DOC_LASTKEY]
      
        do case
        case (nKey == K_F10) .or.;
             (nKey == K_ALT_M)
	    keyboard( K_ENTER ) // to open automatically the file menu
            docMenu()
            docInfoLine()  // per modificare l'intestazione
        case (nKey == K_ESC)
            DocExit()
        otherwise
        end
    end

    // release data inside the static array
    aEdit := {} 
    
    return NIL

//----------------------------------------------------------------
static function  docDoEditing()

    aEdit[DOC_WRAP] := .T.
    aEdit[DOC_TEXT] := MemoEdit( aEdit[DOC_TEXT],     ;
                                 DOC_TOP,;
                                 DOC_LEFT,;
                                 DOC_BOTTOM,;
                                 DOC_RIGHT,;
                                 .t.,;
                                 "docUDF",;
                                 DOC_WIDTH,;
                                 DOC_TABSIZE,;
                                 aEdit[DOC_ROW],      ;
                                 aEdit[DOC_COL],      ;
                                 aEdit[DOC_RELROW],   ;
                                 aEdit[DOC_RELCOL]    ;
                               )
  
    return NIL

//----------------------------------------------------------------
function docUDF(nMode, nLine, nCol)  
//
// E' la funzione indicata a MemoEdit() come UDF.
// Questa funzione viene interpretata ogni volta con &("docUDF")
// all'interno di MemoEdit() stesso e di conseguenza, questa
// funzione non pu essere static, cio senza un nome richiamabile
// durante il run time.
//

    local nKey

    aEdit[DOC_LASTKEY]   := nKey := LastKey()
    aEdit[DOC_ROW]       := nLine
    aEdit[DOC_COL]       := nCol
    aEdit[DOC_RELROW]    := Row() - DOC_TOP
    aEdit[DOC_RELCOL]    := Col() - DOC_LEFT 

    if (nMode == ME_INIT)
        if (aEdit[DOC_WRAP])
            // turn off word wrap
            aEdit[DOC_WRAP] := .f.
            return (ME_TOGGLEWRAP)  // NOTE
        end
        setcursor( if(aEdit[DOC_INS], SC_INSERT, SC_NORMAL) )
    elseif (nMode == ME_IDLE)
        // mostra la riga delle informazioni
        docInfoLine()
    else
        // keystroke exception
        if (nMode == ME_UNKEYX)
            aEdit[DOC_CHANGED] := .t.
        end
      
        do case
        case (nKey == K_F1)
            Text( DOC_HELP )
        case (nKey == K_F10) .or.;
             (nKey == K_ALT_M)
            // chiama il menu
            return (K_CTRL_W)
        case (nKey == K_INS)
            // ins/ovr
            aEdit[DOC_INS] := !Set(_SET_INSERT)
            setcursor( if(aEdit[DOC_INS], SC_INSERT, SC_NORMAL) )
            docInfoLine()
            return (nKey)
        case (nKey == K_ESC)
            // esc
            return (K_CTRL_W)
        otherwise
            // nix
        end
    end

    return 0

//----------------------------------------------------------------
static function docInfoLine()

    local cLine
    local cCol
    local nCtype
    local nRow
    local nCol

    cLine := padl( alltrim( str( aEdit[DOC_ROW] ) ), 6, "0" )
    cCol  := padl( alltrim( str( aEdit[DOC_COL]+1 ) ), 4, "0" )

    nCtype := SetCursor(0)
    nRow := Row()
    nCol := Col()
    setcolor( COLOR_DEFAULT1 )
    scroll( 00,00,00,maxcol() )
    @00,01 say DOC_FUNCTION_NAME
    @00,col()+5 say alltrim( aEdit[DOC_NAME] )
    @00,col()+5 say cLine + ":" + cCol
    @00,col()+5 say iif( aEdit[DOC_INS], "INS", "OVR" )
    setcolor( COLOR_DEFAULT0 )

    DevPos(nRow, nCol)
   
    SetCursor(nCtype)

    return NIL

//----------------------------------------------------------------
static function docNew()

    local c
    local nCtype
    local lOkDoit := .F.

    if ( aEdit[DOC_CHANGED] )
        nCtype := SetCursor(SC_NORMAL)
        if alert( DOC_PROMPT_ABANDON_CHANGES,; 
                   { _MENU_NO, _MENU_YES } ) == 1
            // nix
        else
            lOkDoit := .T.
        end
        SetCursor(nCtype)
    else
        lOkDoit := .T.
    end

    if lOkDoit
        // nuovo file
        aEdit[DOC_OLD]       := NIL
        aEdit[DOC_NAME]      := _UNTITLED + "." + _EXTENTION_TEXT
        aEdit[DOC_TEXT]      := ""
        aEdit[DOC_CHANGED]   := .F.
        aEdit[DOC_ROW]       := 1
        aEdit[DOC_COL]       := 0
        aEdit[DOC_RELROW]    := 0
        aEdit[DOC_RELCOL]    := 0
    end

    return NIL

//----------------------------------------------------------------
static function docOpen()

    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor( COLOR_DEFAULT0 )
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local bOld_F1               := setkey( K_F1, NIL ) 
    local bOld_F2               := setkey( K_F2, NIL ) 
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
   
    local c
    local nCtype
    local lOkDoit := .F.

    local lGoOn     := .F.
    local cFileName := padr( "*." + _EXTENTION_TEXT, SPACE_LEN )
   
    begin sequence
   
        if ( aEdit[DOC_CHANGED] )
            if alert( DOC_PROMPT_ABANDON_CHANGES,; 
                 { _MENU_NO, _MENU_YES } ) == 1
                // nix
            else
                lOkDoit := .T.
            end
        else
            lOkDoit := .T.
        end
   
        if lOkDoit
            // open file
         
            // create a window
            nBottom     := maxrow()
            nTop        := nBottom - 4
            nLeft       := 0
            nRight      := maxcol()
            nWidth      := nRight - nLeft +1
            cOldScreen := savescreen( nTop, nLeft, nBottom, nRight )
            setcolor( COLOR_DEFAULT1 )
            scroll( nTop, nLeft, nTop, nRight ) 
            scroll( nBottom, nLeft, nBottom, nRight ) 
            devpos( nTop, nLeft )
            devout( padc( DOC_DIALOG_BOX_TOP_OPEN, nWidth ) )
            devpos( nBottom-1, nLeft )
            devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
            devpos( nBottom, nLeft )
            devout( padc( _KEY_F2_PGDN, nWidth ) )

            while .t.
                setcolor( COLOR_DEFAULT2 )
                scroll( nTop+1, nLeft, nBottom-2, nRight ) 
                setpos( nTop, nLeft )
                @row()+1,nLeft;
                    say DOC_PROMPT_OPEN_FILE_DOC
                @row()+1,nLeft;
                    get cFileName;
                    picture "@s"+ ltrim( str( nWidth ) ) +"@" ;
                    when trueSetkey( K_F2, {|| cFileName :=;
                        padr( Dir( cFileName ), SPACE_LEN ) } );
                    valid ( file( cFileName :=;
                        padr(strAddExtention( cFileName,;
                        _EXTENTION_TEXT ), SPACE_LEN) ) );
                        .and. !( "*" $ cFileName);
                        .and. !( "?" $ cFileName)
                read

                do case
                case lastkey() = K_ESC  // exit
                    lGoOn := .F.
                    exit
                case lastkey() = K_PGDN // ok
                    if file( strAddExtention( cFileName, _EXTENTION_TEXT ) ) 
                        lGoOn := .T.
                        exit
                    else
                        // loop
                    end
                otherwise
                    // loop
                end
            end
   
            // delete window
            restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
           
            if lGoOn
                aEdit[DOC_NEW]       := .F.
                aEdit[DOC_OLD]       := upper(alltrim( cFileName ))
                aEdit[DOC_NAME]      := upper(alltrim( cFileName ))
                aEdit[DOC_TEXT]      := MemoRead(cFileName)
                aEdit[DOC_CHANGED]   := .F.
                aEdit[DOC_ROW]       := 1
                aEdit[DOC_COL]       := 0
                aEdit[DOC_RELROW]    := 0
                aEdit[DOC_RELCOL]    := 0
            end

        end

    end sequence
    
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//----------------------------------------------------------------
static function docExit()

    local c
    local lRet
    local nCtype

    lRet = .t.
    if ( aEdit[DOC_CHANGED] )
        nCtype := SetCursor(SC_NORMAL)
        if alert( DOC_PROMPT_ABANDON_CHANGES,; 
                   { _MENU_NO, _MENU_YES } ) <> 1
            lRet := .T.
        else
            lRet := .f.
        end
        setcursor(nCtype)
    else
        nCtype := SetCursor(SC_NORMAL)
        if alert( DOC_PROMPT_EXIT,; 
               { _MENU_YES, _MENU_NO } ) == 1
            lRet := .T.
        else
            lRet := .f.
        end
        setcursor(nCtype)
    end

    aEdit[DOC_EXIT] := lRet
    return (lRet)

//----------------------------------------------------------------
static function docSave()
   
    local lRet

    lRet := .F.
   
    WaitFor( DOC_WRITING + aEdit[DOC_NAME] )
   
    begin sequence
   
        do case 
        case aEdit[DOC_NEW]  // nuovo file
            lRet := docSaveAs()   // chiamata ricorsiva
            break
        case aEdit[DOC_NAME] <> aEdit[DOC_OLD]  // sono gi upper e alltrim()
            if file( aEdit[DOC_NAME] )
                if alert( aEdit[DOC_NAME] + ";" +;
                      _ERROR_PROMPT_ALREADY_EXIST_OVERWRITE,;
                      { _MENU_NO, _MENU_YES } ) == 1
                    // non si vuole sovrascrivere e non si salva
                    lRet := .F.
                    break
                else
                    // ok si salva
                end
            else
                // ok si procede
            end
        end

        // se siamo qui, si procede

        // si aggiorna il vecchio nome
        aEdit[DOC_OLD] := aEdit[DOC_NAME]
   
        if ( MemoWrit(aEdit[DOC_NAME], aEdit[DOC_TEXT]) )
            WaitFor( DOC_WRITE_OK )
            aEdit[DOC_CHANGED] := .F.
            lRet := .T.
        else
            alert( DOC_WRITE_ERROR )
            lRet := .F.
        end
   
    end sequence
   
    WaitFor()

    return lRet

//----------------------------------------------------------------
static function docSaveAs()

    local cNewName := padr( aEdit[DOC_NAME], SPACE_LEN )
    local lRet := .F.

    cNewName := Accept( cNewName, DOC_NEW_NAME )
    
    if cNewName == NIL   // premuto [Esc]
        lRet := .F.
    else
        cNewName := strAddExtention( cNewName, _EXTENTION_TEXT )
        aEdit[DOC_NAME] := upper( alltrim(cNewName) )
        aEdit[DOC_NEW] := .F.
        lRet := docSave()
    end
   
    return lRet

//----------------------------------------------------------------
static function docMenu()
//
    local cOldScreen        := savescreen()
    local cOldColor         := setcolor()
    local nOldCursor        := setcursor( SC_NORMAL )
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )
    local bOld_F2           := setkey( K_F2, NIL )
    local bOld_LEFT         := setkey( K_LEFT, NIL )
    local bOld_RIGHT        := setkey( K_RIGHT, NIL )
    local nOldRow           := row()
    local nOldCol           := col()

    local nMnuChoice := 1

    begin sequence

        while .T.
            setcolor( COLOR_DEFAULT1 )
            scroll( 00, 00, 00, maxcol() )
            scroll( maxrow()-1, 00, maxrow()-1, maxcol() )
            @maxrow(),00 say padc( _KEY_MENU_H, maxcol()+1 )

            // The horizontal menu starts
            @00,01 prompt DOC_MENU_FILE_CHOICE;	 //File
                   message DOC_MENU_FILE_MSG
            menu to nMnuChoice

            // [Left]/[Right] key redirection
            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            setcolor( COLOR_DEFAULT0 )

            // analisi del menu
            do case
            case nMnuChoice == 0 // ESC
                break
            case nMnuChoice == 1
                if docMenuFile( 0, 0+len(DOC_MENU_FILE_CHOICE)+1 )
		    break
		end
	    end

            // arrow keys are released
            setkey( K_LEFT, NIL )
            setkey( K_RIGHT, NIL )

        end

    end sequence

    // restore
    setcolor( cOldColor )
    setcursor( nOldCursor )
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
    setkey( K_LEFT, bOld_LEFT )
    setkey( K_RIGHT, bOld_RIGHT )

    return NIL

//----------------------------------------------------------------
static function docMenuFile( nHLeft, nHRight )

    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nTB               := 1            // top border
    local nLB               :=;
        iif( nHLeft+_MENU_TENDINA_WIDTH > maxcol(),;
        nHRight - _MENU_TENDINA_WIDTH,;
        nHLeft )                            // left border
    local nBB               := nTB +12       // bottom border
    local nRB               :=;
        nLB +_MENU_TENDINA_WIDTH -1         // right border
    local n

    local cSaveScreen
    local bSaveF1

    local nMnuChoice := 0

    local lRetBreak := .F.

    begin sequence

        while .T.  // do until Esc
            setcolor( COLOR_DEFAULT1 )
            dispbox( nTB,nLB, nBB,nRB )
            @maxrow(),00 say padc( _KEY_MENU_V, maxcol()+1 )
            n := 0
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_NEW_CHOICE;
                             message DOC_MENU_FILE_NEW_MSG
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_OPEN_CHOICE;
                             message DOC_MENU_FILE_OPEN_MSG
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_SAVE_CHOICE;
                             message DOC_MENU_FILE_SAVE_MSG
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_SAVE_AS_CHOICE;
                             message DOC_MENU_FILE_SAVE_AS_MSG
            @nTB+(++n),nLB+1 say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_SET_DEFAULT_TO_CHOICE;
                             message DOC_MENU_FILE_SET_DEFAULT_TO_MSG
            @nTB+(++n),nLB+1 say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_PRINT_AS_IS_CHOICE;
                             message DOC_MENU_FILE_PRINT_AS_IS_MSG
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_PRINT_CHOICE;
                             message DOC_MENU_FILE_PRINT_MSG
            @nTB+(++n),nLB+1 say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1 prompt DOC_MENU_FILE_EXIT_CHOICE;
                             message DOC_MENU_FILE_EXIT_MSG
            menu to nMnuChoice

            // Extra security screen save.
            cSaveScreen := savescreen()

            // arrow keys are released
            setkey( K_LEFT, NIL )
            setkey( K_RIGHT, NIL )

            setcolor( COLOR_DEFAULT0 )

            bSaveF1 := setkey( K_F1, NIL )

            do case
            case (nMnuChoice == 0)
                lRetBreak := .F.
                exit
            case (nMnuChoice == 1)
                lRetBreak := .T.
                docNew()
                exit
            case (nMnuChoice == 2)
                lRetBreak := .T.
                docOpen()
                exit
            case (nMnuChoice == 3)
                lRetBreak := .T.
                docSave()
                exit
            case (nMnuChoice == 4)
                lRetBreak := .T.
                docSaveAs()
                exit
            case (nMnuChoice == 5)
                set( _SET_ALTERNATE, .T. )
                set( _SET_ALTFILE,;
                    accept( padr( set( _SET_ALTFILE ),;
                    SPACE_LEN ),;
                    ASSIST_PROMPT_SET_ALTERNATE ) )
            case (nMnuChoice == 6)
                lRetBreak := .T.
                QOut( aEdit[DOC_TEXT] )
                exit
            case (nMnuChoice == 7)
                lRetBreak := .T.
                docPrintCondition( aEdit[DOC_TEXT] ) 
                exit
            case (nMnuChoice == 8)
                lRetBreak := .T.
                aEdit[DOC_EXIT] := .T.
                exit
            end

            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            // Screen restore.
            restscreen(,,,,cSaveScreen)

            setkey( K_F1, bSaveF1 )

        end

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return lRetBreak

//----------------------------------------------------------------
static function docPrintCondition( cText )  
//
// Richiede le informazioni necessarie per stampare il testo.
//
// <cText>       il testo da stampare
//
   
    local getlist               := {}
    local bOldErrorHandler      
    local cOldScreen            := ""
    local cOldColor             := setcolor()
    local nOldCursor            := setcursor( SC_NORMAL ) 
    local bOld_F1               :=;
        setkey( K_F1, { || Text( DOC_HELP_PRINT ) } ) 
    local bOld_F2               := setkey( K_F2, NIL ) 
    local nOldRow               := row()
    local nOldCol               := col()
    local nTop                  := 0
    local nLeft                 := 0
    local nBottom               := 0
    local nRight                := 0
    local nWidth                := 0
   
    local lGoPrint := .T.
   
    local cWhileCondition := padr( ".T.", SPACE_LEN )
    local cForCondition   := padr( ".T.", SPACE_LEN )
   
    begin sequence

        // analisi dei parametri forniti
        if valtype( cText ) <> "C" 
            // non  stato fornito alcun testo da stampare
            break
        end
      
        // si fanno le domande se esiste un alias attivo
        if !(alias() == "")
         
            // create a window
            nBottom     := maxrow()
            nTop        := nBottom - 6
            nLeft       := 0
            nRight      := maxcol()
            nWidth      := nRight - nLeft +1
            cOldScreen  :=;
                savescreen( nTop, nLeft, nBottom, nRight )
            setcolor( COLOR_DEFAULT1 )
            devpos( nTop, nLeft )
            devout( padc( DOC_DIALOG_BOX_TOP_PRINT, nWidth ) )
            devpos( nBottom-1, nLeft )
            devout( padc( _KEY_ESC_CANCEL_F1, nWidth ) )
            devpos( nBottom, nLeft )
            devout( padc( _KEY_F2_PGDN, nWidth ) )

            while .t.
                setcolor( COLOR_DEFAULT2 )
                scroll( nTop+1, nLeft, nBottom-2, nRight )

                setpos( nTop, nLeft )
                @row()+1,nLeft;
                    say DOC_PROMPT_WHILE_EXPRESSION
                @row()+1,nLeft;
                    get cWhileCondition;
                    picture "@s"+ltrim(str(nWidth))+"@";  
                    when trueSetkey( K_F2, NIL );
                    valid ( ( cWhileCondition :=;
                        iif( cWhileCondition <> SPACE_STR,;
                        cWhileCondition, padr( ".T.",;
                        SPACE_LEN ) ) ) <> NIL )
                @row()+1,nLeft;
                    say DOC_PROMPT_FOR_EXPRESSION
                @row()+1,nLeft;
                    get cForCondition;
                    picture "@s"+ltrim(str(nWidth))+"@";    
                    when trueSetkey( K_F2, {|| cForCondition :=;
                        padr( fieldNormal(),;
                        SPACE_LEN ) } );     
                    valid ( ( cForCondition :=;
                        iif( cForCondition <> SPACE_STR,;
                        cForCondition, padr( ".T.",;
                        SPACE_LEN ) ) ) <> NIL )
                read

                do case
                case lastkey() = K_ESC
                    lGoPrint := .F.
                    exit
                case lastkey() = K_PGDN // conferma la scelta
                    // verifica la presenza di dati corretti
                    if cWhileCondition <> SPACE_STR .and.;
                        cForCondition <> SPACE_STR 
                        lGoPrint := .T.
                        exit
                    else
                        // ripete il loop
                    end
                otherwise
                    // ripete il loop
                end
            end

            // delete window
            restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
         
      
        else
            // se non c' un alias attivo stampa ugualmente
            lGoPrint := .T.
        end
   
        if lGoPrint
            if alias() == ""
                rpt( cText )
            else
                rptMany( cText, { || &(alltrim(cWhileCondition)) },;
                    { || &(alltrim(cForCondition)) }, NIL )
            end
            WaitProgress(2)
        end

    end sequence
   
    // restore
    setcursor( nOldCursor ) 
    setcolor( cOldColor )
    restscreen( nTop, nLeft, nBottom, nRight, cOldScreen )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
   
    return NIL

//================================================================
// ASSIST - nB MENU
//================================================================
function Assist()
//

    local cOldScreen        := savescreen()
    local cOldColor         := setcolor()
    local nOldCursor        := setcursor( SC_NORMAL )
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )
    local bOld_F2           := setkey( K_F2, NIL )
    local bOld_F3           :=;
        setkey( K_F3, { || Text( dbiStatus() ) } )
    local bOld_F5           :=;
        setkey( K_F5, { || set( _SET_ALTERNATE, .T. ),;
            set( _SET_ALTFILE,;
            accept( padr( set( _SET_ALTFILE ), SPACE_LEN ),;
            ASSIST_PROMPT_SET_ALTERNATE ) ) } )
    local bOld_LEFT         := setkey( K_LEFT, NIL )
    local bOld_RIGHT        := setkey( K_RIGHT, NIL )
    local nOldRow           := row()
    local nOldCol           := col()

    local nMnuChoice := 1

    begin sequence

        while .T.
            setcolor( COLOR_DEFAULT1 )
            scroll( 00, 00, 00, maxcol() )
            scroll( maxrow()-1, 00, maxrow()-1, maxcol() )
            @maxrow(),00 say padc( _KEY_MENU_H, maxcol()+1 )

            // The horizontal menu starts
            @00,01 prompt ASSIST_MENU_FILE_CHOICE;
                message ASSIST_MENU_FILE_MSG
            @row(),col()+1 prompt ASSIST_MENU_EDIT_CHOICE;
                message ASSIST_MENU_EDIT_MSG
            @row(),col()+1 prompt ASSIST_MENU_REPORT_CHOICE;
                message ASSIST_MENU_REPORT_MSG
            @row(),col()+1 prompt ASSIST_MENU_INFO_CHOICE;
                message ASSIST_MENU_INFO_MSG
            menu to nMnuChoice

            // [Left]/[Right] key redirection
            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            setcolor( COLOR_DEFAULT0 )

            // analisi del menu
            do case
            case nMnuChoice == 0 // ESC
                break
            case nMnuChoice == 1
                mfile( 0, 0+len(ASSIST_MENU_FILE_CHOICE)+1 )
            case nMnuChoice == 2
                mEdit( 5, 5+len(ASSIST_MENU_FILE_CHOICE)+1 )
            case nMnuChoice == 3
                mReport( 10, 10+len(ASSIST_MENU_FILE_CHOICE)+1 )
            case nMnuChoice == 4
                mInfo( 17, 17+len(ASSIST_MENU_FILE_CHOICE)+1 )
            end

            // arrow keys are released
            setkey( K_LEFT, NIL )
            setkey( K_RIGHT, NIL )

        end

    end sequence

    // restore
    setcolor( cOldColor )
    setcursor( nOldCursor ) 
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setpos( nOldRow, nOldCol )
    setkey( K_F1, bOld_F1 )
    setkey( K_F2, bOld_F2 )
    setkey( K_F3, bOld_F3 )
    setkey( K_F5, bOld_F5 )
    setkey( K_LEFT, bOld_LEFT )
    setkey( K_RIGHT, bOld_RIGHT )

    return NIL

//----------------------------------------------------------------
static function mFile( nHLeft, nHRight )

    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nTB               := 1            // top border
    local nLB               :=;
        iif( nHLeft+_MENU_TENDINA_WIDTH > maxcol(),;
        nHRight - _MENU_TENDINA_WIDTH,;
        nHLeft )                            // left border
    local nBB               := nTB +9       // bottom border
    local nRB               :=;
        nLB +_MENU_TENDINA_WIDTH -1         // right border
    local n

    local cSaveScreen
    local bSaveF1

    local nMnuChoice := 0

    begin sequence

        while .T.  // do until Esc
            setcolor( COLOR_DEFAULT1 )
            dispbox( nTB,nLB, nBB,nRB )
            @maxrow(),00 say padc( _KEY_MENU_V, maxcol()+1 )
            n := 0
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_NEWDIR_CHOICE;
                message ASSIST_MENU_FILE_NEWDIR_MSG
            @nTB+(++n),nLB+1 say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_DBF_CHOICE; 
                message ASSIST_MENU_FILE_DBF_MSG 
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_NTX_CHOICE;
                message ASSIST_MENU_FILE_NTX_MSG
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_ALIAS_CHOICE;
                message ASSIST_MENU_FILE_ALIAS_MSG
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_INDEX_CHOICE;
                message ASSIST_MENU_FILE_INDEX_MSG
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_RELATION_CHOICE;
                message ASSIST_MENU_FILE_RELATION_MSG
            @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_RDD_DEFAULT_CHOICE;
                message ASSIST_MENU_FILE_RDD_DEFAULT_MSG
            menu to nMnuChoice

            // Extra security screen save.
            cSaveScreen := savescreen()

            // arrow keys are released
            setkey( K_LEFT, NIL ) 
            setkey( K_RIGHT, NIL )

            setcolor( COLOR_DEFAULT0 )

            bSaveF1 := setkey( K_F1, NIL )

            do case
            case nMnuChoice == 0  // ESC
                exit
            case nMnuChoice == 1
                run( "cd " + Dir( "*.*", .f., .t., .f. ) )
            case nMnuChoice == 2
                pFileDbf(4,nRB)
            case nMnuChoice == 3
                pFileNtx(5,nRB)
            case nMnuChoice == 4
                pFileAlias(6,nRB)
            case nMnuChoice == 5
                pFileIndex(7,nRB)
            case nMnuChoice == 6
                pFileRelation(8,nRB)
            case nMnuChoice == 7
                pFileRDD(9,nRB)
            end

            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            // Screen restore.
            restscreen(,,,,cSaveScreen)

            setkey( K_F1, bSaveF1 )

        end

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//----------------------------------------------------------------
static function pFileDbf(nTB,nLB)


    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nMenuHeight       := 5    // menu lines +2
    local nBB                       // bottom border
    local nRB                       // right border
    local n

    local nMnuChoice := 0

    local bSaveF1

    if (nTB + nMenuHeight -1) > maxrow()
        nBB := maxrow()
        nTB := maxrow() - nMenuHeight +1
    else
        nBB := nTB + nMenuHeight -1
    end
    if (nLB + _MENU_TENDINA_WIDTH -1) > maxcol()
        nLB := nLB - 2*_MENU_TENDINA_WIDTH +1
    end
    nRB := nLB + _MENU_TENDINA_WIDTH -1

    begin sequence

        setcolor( COLOR_DEFAULT1 )
        dispbox( nTB,nLB, nBB,nRB )
        @maxrow(),00 say;
            padc( _KEY_MENU_P,;
                maxcol()+1 )
        n := 0
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_DBF_NEW_CHOICE;
            message ASSIST_MENU_FILE_DBF_NEW_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_DBF_MODSTRU_CHOICE;
            message ASSIST_MENU_FILE_DBF_MODSTRU_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_DBF_OPEN_CHOICE;
            message ASSIST_MENU_FILE_DBF_OPEN_MSG
        menu to nMnuChoice

        bSaveF1 := setkey( K_F1, NIL )

        do case
        case nMnuChoice == 1
            idbStructure()
        case nMnuChoice == 2
            idbModStructure()
        case nMnuChoice == 3
            idbUseArea()
        end

        setkey( K_F1, bSaveF1 )

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//----------------------------------------------------------------
static function pFileNtx(nTB,nLB)


    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nMenuHeight       := 4    // menu lines +2
    local nBB                       // bottom border
    local nRB                       // right border
    local n

    local nMnuChoice := 0

    local bSaveF1

    if (nTB + nMenuHeight -1) > maxrow()
        nBB := maxrow()
        nTB := maxrow() - nMenuHeight +1
    else
        nBB := nTB + nMenuHeight -1
    end
    if (nLB + _MENU_TENDINA_WIDTH -1) > maxcol()
        nLB := nLB - 2*_MENU_TENDINA_WIDTH +1
    end
    nRB := nLB + _MENU_TENDINA_WIDTH -1

    begin sequence

        setcolor( COLOR_DEFAULT1 )
        dispbox( nTB,nLB, nBB,nRB )
        @maxrow(),00 say padc( _KEY_MENU_P, maxcol()+1 )
        n := 0

        @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_NTX_NEW_CHOICE;
            message ASSIST_MENU_FILE_NTX_NEW_MSG
        @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_NTX_OPEN_CHOICE;
            message ASSIST_MENU_FILE_NTX_OPEN_MSG
        menu to nMnuChoice

        bSaveF1 := setkey( K_F1, NIL )

        do case
        case (nMnuChoice == 1)
            idbOrdCreate()
        case (nMnuChoice == 2)
            idbSetIndex()
        end

        setkey( K_F1, bSaveF1 )	    

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//----------------------------------------------------------------
static function pFileAlias(nTB,nLB)


    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nMenuHeight       := 6    // menu lines +2
    local nBB                       // bottom border
    local nRB                       // right border
    local n

    local dummy

    local nMnuChoice := 0

    local bSaveF1

    if (nTB + nMenuHeight -1) > maxrow()
        nBB := maxrow()
        nTB := maxrow() - nMenuHeight +1
    else
        nBB := nTB + nMenuHeight -1
    end
    if (nLB + _MENU_TENDINA_WIDTH -1) > maxcol()
        nLB := nLB - 2*_MENU_TENDINA_WIDTH +1
    end
    nRB := nLB + _MENU_TENDINA_WIDTH -1

    begin sequence

        setcolor( COLOR_DEFAULT1 )
        dispbox( nTB,nLB, nBB,nRB )
        @maxrow(),00 say padc( _KEY_MENU_P, maxcol()+1 )
        n := 0

        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_ALIAS_SELECT_CHOICE;
            message ASSIST_MENU_FILE_ALIAS_SELECT_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_ALIAS_DISPSTRU_CHOICE;
            message ASSIST_MENU_FILE_ALIAS_DISPSTRU_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_ALIAS_CLOSE_CHOICE;
            message ASSIST_MENU_FILE_ALIAS_CLOSE_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_ALIAS_CLOSEALL_CHOICE;
            message ASSIST_MENU_FILE_ALIAS_CLOSEALL_MSG
        menu to nMnuChoice

        bSaveF1 := setkey( K_F1, NIL )

        do case
        case nMnuChoice == 1
            dummy := selectArray()[1]
            if valtype(dummy) == "N"
                dbselectarea(dummy)
            end
        case nMnuChoice == 2
            dummy := dbiStructure()
            if valtype(dummy) == "C"
                text( dummy )
            end
        case nMnuChoice == 3
            dbclosearea()
        case nMnuChoice == 4
            dbcloseall()
        end

        setkey( K_F1, bSaveF1 )

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL


//----------------------------------------------------------------
static function pFileIndex(nTB,nLB)


    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nMenuHeight       := 5    // menu lines +2
    local nBB                       // bottom border
    local nRB                       // right border
    local n

    local dummy

    local nMnuChoice := 0

    local bSaveF1

    if (nTB + nMenuHeight -1) > maxrow()
        nBB := maxrow()
        nTB := maxrow() - nMenuHeight +1
    else
        nBB := nTB + nMenuHeight -1
    end
    if (nLB + _MENU_TENDINA_WIDTH -1) > maxcol()
        nLB := nLB - 2*_MENU_TENDINA_WIDTH +1
    end
    nRB := nLB + _MENU_TENDINA_WIDTH -1

    begin sequence

        setcolor( COLOR_DEFAULT1 )
        dispbox( nTB,nLB, nBB,nRB )
        @maxrow(),00 say padc( _KEY_MENU_P, maxcol()+1 )
        n := 0

        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_INDEX_REINDEX_CHOICE;
            message ASSIST_MENU_FILE_INDEX_REINDEX_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_INDEX_SET_ORDER_CHOICE;
            message ASSIST_MENU_FILE_INDEX_SET_ORDER_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_INDEX_CLOSENTX_CHOICE;
            message ASSIST_MENU_FILE_INDEX_CLOSENTX_MSG
        menu to nMnuChoice

        bSaveF1 := setkey( K_F1, NIL )

        do case
        case nMnuChoice == 1
            idbReindex()
        case nMnuChoice == 2
    	    dummy := selectOrder()
            if valtype( dummy ) == "N"
                ordsetfocus( dummy )
            end
        case nMnuChoice == 3
            idbClearIndex()
        end

        setkey( K_F1, bSaveF1 )

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL


//----------------------------------------------------------------
static function pFileRelation(nTB,nLB)


    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nMenuHeight       := 4    // menu lines +2
    local nBB                       // bottom border
    local nRB                       // right border
    local n

    local dummy

    local nMnuChoice := 0

    local bSaveF1

    if (nTB + nMenuHeight -1) > maxrow()
        nBB := maxrow()
        nTB := maxrow() - nMenuHeight +1
    else
        nBB := nTB + nMenuHeight -1
    end
    if (nLB + _MENU_TENDINA_WIDTH -1) > maxcol()
        nLB := nLB - 2*_MENU_TENDINA_WIDTH +1
    end
    nRB := nLB + _MENU_TENDINA_WIDTH -1

    begin sequence

        setcolor( COLOR_DEFAULT1 )
        dispbox( nTB,nLB, nBB,nRB )
        @maxrow(),00 say padc( _KEY_MENU_P, maxcol()+1 )
        n := 0

        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_RELATION_SET_CHOICE;
            message ASSIST_MENU_FILE_RELATION_SET_MSG
        @nTB+(++n),nLB+1 prompt ASSIST_MENU_FILE_RELATION_CLOSE_CHOICE;
            message ASSIST_MENU_FILE_RELATION_CLOSE_MSG
        menu to nMnuChoice

        bSaveF1 := setkey( K_F1, NIL )

        do case
        case nMnuChoice == 1
            idbRelation()
        case nMnuChoice == 2
            dbclearrelation()
        end

        setkey( K_F1, bSaveF1 )

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL


//----------------------------------------------------------------
static function pFileRDD(nTB,nLB)


    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nMenuHeight       := 4    // menu lines +2
    local nBB                       // bottom border
    local nRB                       // right border
    local n

    local dummy

    local nMnuChoice := 0

    local bSaveF1

    if (nTB + nMenuHeight -1) > maxrow()
        nBB := maxrow()
        nTB := maxrow() - nMenuHeight +1
    else
        nBB := nTB + nMenuHeight -1
    end
    if (nLB + _MENU_TENDINA_WIDTH -1) > maxcol()
        nLB := nLB - 2*_MENU_TENDINA_WIDTH +1
    end
    nRB := nLB + _MENU_TENDINA_WIDTH -1

    begin sequence

        setcolor( COLOR_DEFAULT1 )
        dispbox( nTB,nLB, nBB,nRB )
        @maxrow(),00 say padc( _KEY_MENU_P, maxcol()+1 )
        n := 0

        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_RDD_SHOW_CHOICE;
            message ASSIST_MENU_FILE_RDD_SHOW_MSG
        @nTB+(++n),nLB+1;
            prompt ASSIST_MENU_FILE_RDD_SET_DEFAULT_CHOICE;
            message ASSIST_MENU_FILE_RDD_SET_DEFAULT_MSG
        menu to nMnuChoice

        bSaveF1 := setkey( K_F1, NIL )

        do case
        case nMnuChoice == 1
            alert( rddsetdefault() )
        case nMnuChoice == 2
    	    dummy := selectRdd()
	        if valtype(dummy) == "C"
                rddsetdefault( dummy )
	        end
        end
        
        setkey( K_F1, bSaveF1 )	    

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//----------------------------------------------------------------
static function mEdit( nHLeft, nHRight )

    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nTB               := 1            // top border
    local nLB               :=;
        iif( nHLeft+_MENU_TENDINA_WIDTH > maxcol(),;
        nHRight - _MENU_TENDINA_WIDTH,;
        nHLeft )                            // left border
    local nBB               := nTB +8       // bottom border
    local nRB               :=;
        nLB +_MENU_TENDINA_WIDTH -1         // right border
    local n  

    local cSaveScreen
    local bSaveF1
    
    local nMnuChoice := 0

    begin sequence

        while .T.  // do until Esc
            setcolor( COLOR_DEFAULT1 )
            dispbox( nTB,nLB, nBB,nRB )
            @maxrow(),00 say padc( _KEY_MENU_V, maxcol()+1 )
            n := 0
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_EDIT_VIEW_CHOICE;
                message ASSIST_MENU_EDIT_VIEW_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_EDIT_BROWSE_CHOICE;
                message ASSIST_MENU_EDIT_BROWSE_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_EDIT_REPLACE_CHOICE;
                message ASSIST_MENU_EDIT_REPLACE_MSG
            @nTB+(++n),nLB+1;
                say replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_EDIT_RECALL_CHOICE;
                message ASSIST_MENU_EDIT_RECALL_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_EDIT_DELETE_CHOICE;
                message ASSIST_MENU_EDIT_DELETE_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_EDIT_PACK_CHOICE;
                message ASSIST_MENU_EDIT_PACK_MSG
            menu to nMnuChoice

            // Extra security screen save.
            cSaveScreen := savescreen()

            // arrow keys are released
            setkey( K_LEFT, NIL )
            setkey( K_RIGHT, NIL )

            setcolor( COLOR_DEFAULT0 )

            bSaveF1 := setkey( K_F1, NIL )

            do case
            case nMnuChoice == 0  // ESC
                exit
            case nMnuChoice == 1
                tb(,,,,,,,,, .f.,,,)
            case nMnuChoice == 2
                tb()
            case nMnuChoice == 3
                idbReplace()
            case nMnuChoice == 4
                idbRecall()
            case nMnuChoice == 5
                idbDelete()
            case nMnuChoice == 6
                idbPack()
            end

            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            // Screen restore.
            restscreen(,,,,cSaveScreen)

            setkey( K_F1, bSaveF1 )

        end

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//----------------------------------------------------------------
static function mReport( nHLeft, nHRight )

    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nTB               := 1            // top border
    local nLB               :=;
        iif( nHLeft+_MENU_TENDINA_WIDTH > maxcol(),;
        nHRight - _MENU_TENDINA_WIDTH,;
        nHLeft )                            // left border
    local nBB               := nTB +10      // bottom border
    local nRB               :=;
        nLB +_MENU_TENDINA_WIDTH -1         // right border
    local n

    local cSaveScreen
    local bSaveF1

    local nMnuChoice := 0

    begin sequence

        while .T.  // do until Esc
            setcolor( COLOR_DEFAULT1 )
            dispbox( nTB,nLB, nBB,nRB )
            @maxrow(),00 say padc( _KEY_MENU_V, maxcol()+1 )
            n := 0
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_LABEL_NEW_CHOICE;
                message ASSIST_MENU_REPORT_LABEL_NEW_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_LABEL_MODIFY_CHOICE;
                message ASSIST_MENU_REPORT_LABEL_MODIFY_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_LABEL_FORM_CHOICE;
                message ASSIST_MENU_REPORT_LABEL_FORM_MSG
            @nTB+(++n),nLB+1;
                say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_REPORT_NEW_CHOICE;
                message ASSIST_MENU_REPORT_REPORT_NEW_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_REPORT_MODIFY_CHOICE;
                message ASSIST_MENU_REPORT_REPORT_MODIFY_MSG
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_REPORT_FORM_CHOICE;
                message ASSIST_MENU_REPORT_REPORT_FORM_MSG
            @nTB+(++n),nLB+1;
                say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_REPORT_DOCUMENT_NEW_CHOICE;
                message ASSIST_MENU_REPORT_DOCUMENT_NEW_MSG
            menu to nMnuChoice

            // Extra security screen save.
            cSaveScreen := savescreen()

            // arrow keys are released
            setkey( K_LEFT, NIL )
            setkey( K_RIGHT, NIL )

            setcolor( COLOR_DEFAULT0 )

            bSaveF1 := setkey( K_F1, NIL )

            do case
            case nMnuChoice == 0  // ESC
                exit
            case nMnuChoice == 1
                lbl()
            case nMnuChoice == 2
                lblModify()
            case nMnuChoice == 3
                lblPrint()
            case nMnuChoice == 4
                frm()
            case nMnuChoice == 5
                frmModify()
            case nMnuChoice == 6
                frmPrint()
            case nMnuChoice == 7
                doc()
            end

            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            // Screen restore.
            restscreen(,,,,cSaveScreen)

            setkey( K_F1, bSaveF1 )

        end

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//----------------------------------------------------------------
static function mInfo( nHLeft, nHRight )

    local cOldScreen        := savescreen()
    local lOldSetWrap       := set(_SET_WRAP, .T.)
    local nOldSetMessage    := set(_SET_MESSAGE, maxrow()-1)
    local lOldSetCenter     := set(_SET_MCENTER, .T.)
    local bOld_F1           :=;
        setkey( K_F1, {|| Text(_MENU_HELP)} )

    local nTB               := 1            // top border
    local nLB               :=;
        iif( nHLeft+_MENU_TENDINA_WIDTH > maxcol(),;
        nHRight - _MENU_TENDINA_WIDTH,;
        nHLeft )                            // left border
    local nBB               := nTB +6       // bottom border
    local nRB               :=;
        nLB +_MENU_TENDINA_WIDTH -1         // right border
    local n  

    local cSaveScreen
    local bSaveF1    

    local nMnuChoice := 0

    begin sequence

        while .T.  // do until Esc
            setcolor( COLOR_DEFAULT1 )
            dispbox( nTB,nLB, nBB,nRB )
            @maxrow(),00 say padc( _KEY_MENU_V, maxcol()+1 )
            n := 0
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_INFO_ABOUT_CHOICE;
                message ASSIST_MENU_INFO_ABOUT_MSG
            @nTB+(++n),nLB+1;
                say    replicate("", _MENU_TENDINA_WIDTH-2)
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_INFO_F1_CHOICE
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_INFO_F3_CHOICE
            @nTB+(++n),nLB+1;
                prompt ASSIST_MENU_INFO_F5_CHOICE
            menu to nMnuChoice

            // Extra security screen save.
            cSaveScreen := savescreen()

            // arrow keys are released
            setkey( K_LEFT, NIL )
            setkey( K_RIGHT, NIL )

            setcolor( COLOR_DEFAULT0 )

            bSaveF1 := setkey( K_F1, NIL )

            do case
            case nMnuChoice == 0  // ESC
                exit
            case nMnuChoice == 1
                // Waitfor() is used as new lines are made with CL+LF
                WaitFor( PROGRAM_COPYRIGHT )
                inkey(0)
                WaitFor()
            case nMnuChoice == 2
		Text( _MENU_HELP )
            case nMnuChoice == 3
                Text( dbiStatus() )
            case nMnuChoice == 4
                set( _SET_ALTERNATE, .T. )
                set( _SET_ALTFILE,;
                    accept( padr( set( _SET_ALTFILE ),;
                    SPACE_LEN ),;
                    ASSIST_PROMPT_SET_ALTERNATE ) )
            end

            setkey( K_LEFT, { || MenuLeft() } )
            setkey( K_RIGHT, { || MenuRight() } )

            // Screen restore.
            restscreen(,,,,cSaveScreen)

            setkey( K_F1, bSaveF1 )

        end

    end sequence

    // restore
    restscreen( ,,,, cOldScreen )
    set(_SET_WRAP, lOldSetWrap )
    set(_SET_MESSAGE, nOldSetMessage )
    set(_SET_MCENTER, lOldSetCenter )
    setkey( K_F1, bOld_F1 )

    return NIL

//================================================================
//================================================================
//================================================================
// COMMAND SUBSTITUTION FUNCTIONS
//================================================================
function bCompile( cBlockString )
//
// bCompile( <cBlockString> ) --> bBlock
//
// <bBlockString>       Block in string form.
//
// Compiles the string <cBlockString> and result a code block.
//

   if cBlockString <> NIL
      return &(cBlockString)
   else
      return NIL
   end

   return NIL

//================================================================
function CopyFile( cSource, cDestination )
//
// CopyFile( <cSource>, <cDestination> ) --> NIL
//
// Copy File command substitute.
//

    copy file (cSource) to (cDestination)

    return NIL

//================================================================
function dbApp( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )
//
// dbApp( <cFileName>, 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>], [<cDriver>] ) --> NIL
//
// Import records from a (.dbf) file.
//
// <cFileName>          the (.dbf) file name to use to 
//                      import data.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <cDriver>            the RDD name to use.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbApp( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )

    return NIL

//================================================================
function dbContinue()
//
// dbContinue() --> NIL
//
// Continue command substitute.
//

    continue
    
    return NIL

//================================================================
function dbCopy( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )
//
// dbCopy( <cFileName>, 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>], [<cDriver>] ) --> NIL
//
// Export records to a new (.dbf) file.
//
// <cFileName>          the (.dbf) file name to create.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <cDriver>            the RDD name to use.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbCopy( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest, cDriver )

    return NIL

//================================================================
function dbCopyStruct( cFileName, acFields )
//
// dbCopyStruct( <cFileName>, [<acFields>] )  --> NIL
//
// Copy the current (.dbf) structure to a new (.dbf) file.
//
    
    // change the following line if a new release uses 
    // a different internal function. 
    __dbCopyStruct( cFileName, acFields )
    
    return NIL

//================================================================
function dbCopyXStruct( cFileStructure )
//
// dbCopyXStruct( <cFileStructure> )  --> NIL
//
// Copy field definitions to a (.dbf) file.
//

    copy structure extended to (cFileStructure) 

    return NIL

//================================================================
function dbOldCreate( cFileName, cFileStructure,; 
                      cDriver, lNew, cAlias )
//
// dbOldCreate( <cFileName>, <cFileStructure>, 
//      [<cDriver>], [<lNew>], [<cAlias>] ) --> NIL
//
// Creates a new database file with structure defined by
// the specified extended file <cFileStructure>.
//
// The name "dbOldCreate" is so to make no confusion with the
// CA-Clipper function dbCreate that works with an Array for
// the data structue.
//
// <cFileName>          contains the file name to create.
// <cFileStructure>     contains the file containing the
//                      structure.
// <cDriver>            contains the driver name to be used.
// <lNew>               if true (.T.) it opend the new file
//                      into the next free work area, else
//                      it will be opened into the actura
//                      area.
// <cAlias>             it contains the alias to be used
//                      to indentify the new file.
//
    
    default lNew to .F.

    if lNew
        create (cFileName);
            from (cFileStructure);
            via cDriver;
            alias (cAlias);
            new
    else
        create (cFileName);
            from (cFileStructure);
            via cDriver;
            alias (cAlias)
    end

    return NIL

//================================================================
function dbDelim( lCopy, cFileName, cDelimiter, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbDelim( <lCopy>, <cFileName>, [<cDelimiter>], 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Export records to a new ASCII file.
// Import records from a ASCII file.
//
// <lCopy>              if true (.T.) it exports, else it 
//                      imports data.
// <cFileName>          the ASCII file name to create or the 
//                      ASCII file name to use to import data.
// <cDelimiter>         the delimiter used.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbDelim( lCopy, cFileName, cDelimiter, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbJoin( cAlias, cFileName, acFields, bForCondition )
//
// dbJoin( <cAlias>, <cFileName>, [<acFields>], 
//      [<bForCondition>] ) --> NIL
//
// Generates a new (.dbf) file by joining the active Alias
// with a second alias <cAlias> and eliminating records that
// do not meet the condition stated into the code block
// <bForCondition>.
//
// <cAlias>             the other Alias containing the data
//                      to use to join.
// <cFileName>          the (.dbf) file to create.
// <acFields>           fields to be involved.
// <bForCondition>      code block FOR.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbJoin( cAlias, cFileName, acFields, bForCondition )

    return NIL

//================================================================
function dbLabelForm( cLabel, lToPrinter, cFileName,;
             lNoConsole, bForCondition, bWhileCondition,; 
             nNextRecords, nRecord, lRest,; 
             lSample )
//
// dbLabelForm( <cLabel>, [<lToPrinter>], [<cFileName>],
//     [<lNoConsole>], [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>], 
//     [<lSample> ) --> NIL
//
// Displays or prints lables form a definition held in a 
// label (.lbl) file for a range of records in the active 
// Alias.
//
// <cLabel>             the label (.lbl) file name.
// <lToPrinter>         if true (.T.) the labels will be
//                      printed.
// <cFileName>          the text file name to create 
//                      containing the lables.
// <lNoConsole>         if true (.T.) the labels will not
//                      displayed on the console.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <lSample>            if true (.T.), only test labels
//                      of asterisks are printed.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __LabelForm( cLabel, lToPrinter, cFileName,;
        lNoConsole, bForCondition, bWhileCondition,; 
        nNextRecords, nRecord, lRest,; 
        lSample )
    
    return NIL

//================================================================
function dbList( lToDisplay, abListColumns, lAll,;
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest,; 
                lToPrinter, cFileName )
//
// dbList( [<lToDisplay>], <abListColumns>, [<lAll>],
//     [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>], 
//     [<lToPrinter>], [<cFileName>] ) --> NIL
//
// Displays or prints the result of one or more expressions
// listed into <abListColumns> for a range of records in the
// active Alias.
//
// <lToDisplay>         if true (.T.) it displays the list.
// <abListColumns>      every element contains the expression
//                      (in code block form) to be listed.
// <lAll>               if true (.T.) it lists all records.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <lToPrinter>         if true (.T.) it prints the list.
// <cFileName>          the text file to create containing
//                      the list.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbList( lToDisplay, abListColumns, lAll,;
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest,; 
                lToPrinter, cFileName )

    return NIL

//================================================================
function dbLocate( bForCondition, bWhileCondition,; 
    nNextRecords, nRecord, lRest; 
                 )
//
// dbLocate( [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>] ) --> NIL
//
// Positions the record pointer to the first record
// in the active Alias that matches the specified condition
// within the given scope.
//
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//
    
    default bForCondition to { || .T. }
    default bWhileCondition to { || .T. }
    
    // change the following line if a new release uses 
    // a different internal function. 
    __dbLocate( bForCondition, bWhileCondition,; 
        nNextRecords, nRecord, lRest; 
              )

    return NIL

//================================================================
function dbPack()
//
// dbPack() --> NIL
//
// Pack command substitute.
//

    pack
    
    return NIL

//================================================================
function dbReportForm( cForm, lToPrinter, cFileName,;
             lNoConsole, bForCondition, bWhileCondition,; 
             nNextRecords, nRecord, lRest,; 
             lPlain, cHeading,;
             lNoEject, lSummary )
//
// dbReportForm( <cForm>, [<lToPrinter>], [<cFileName>],
//     [<lNoConsole>], [<bForCondition>], [<bWhileCondition>], 
//     [<nNextRecords>], [<nRecord>], [<lRest>], 
//     [<lPlain>], [<cHeading>],
//     [<lNoEject>, [<lSummary>] ) --> NIL
//
// Displays or prints a tabular and optionally grouped 
// report with page and column headings for a range of
// records in the active Alias.
// The report definition is held in a report (.frm) file.
//
// <cForm>              the report (.frm) file name.
// <lToPrinter>         if true (.T.) the report will be
//                      printed.
// <cFileName>          the text file name to create 
//                      containing the report.
// <lNoConsole>         if true (.T.) the report will not
//                      displayed on the console.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
// <lPlain>             if true (.T.), the report will
//                      not have page breaks.
// <cHeading>           contains the header printed each
//                      page.
// <lNoEject>           if true (.T.) it suppresses the
//                      initial page eject.
// <lSummary>           if true (.T.), only group, subgroup, 
//                      and grand total lines are printed.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __ReportForm( cForm, lToPrinter, cFileName,;
        lNoConsole, bForCondition, bWhileCondition,; 
        nNextRecords, nRecord, lRest,; 
        lPlain, cHeading,;
        lNoEject, lSummary )
    
    return NIL

//================================================================
function dbSDF( lCopy, cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbSDF( <lCopy>, <cFileName>, 
//              [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Export records to a new ASCII file.
// Import records from a ASCII file.
//
// <lCopy>              if true (.T.) it exports, else it 
//                      imports data.
// <cFileName>          the ASCII file name to create or the 
//                      ASCII file name to use to import data.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbSDF( lCopy, cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbSort( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbSort( <cFileName>, [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Copies records within the specified scope and condition
// from the current work area to another database file
// sorted according to the specified fields.
//
// <cFileName>          the (.dbf) file name to create. 
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbSort( cFileName, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbTotal( cFileName, bKey, acFields,;
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )
//
// dbTotal( <cFileName>, <bKey>, [<acFields>], [<bForCondition>], 
//              [<bWhileCondition>], [<nNextRecords>], [<nRecord>], 
//              [<lRest>] ) --> NIL
//
// Summerizes records in the current work area by key value,
// summing the specified numeric fields and then copying
// summary records to a second database file.
//
// <cFileName>          the (.dbf) file name to create. 
// <bKey>               code block Key expression.
// <acFields>           the fields to be involved.
// <bForCondition>      code block FOR condition.
// <bWhileCondition>    code block WHILE condition.
// <nNextRecords>       equivalent to NEXT ...
// <nRecord>            equivalent to RECORD ...
// <lRest>              if true (.T.), then REST records.
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbTotal( cFileName, bKey, acFields,; 
                bForCondition, bWhileCondition,; 
                nNextRecords, nRecord, lRest )

    return NIL

//================================================================
function dbUpdate( cAlias, bKey, lRandom, bReplacement )
//
// dbUpdate( <cAlias>, <bKey>, [<lRandom>], 
//              [<bReplacement>] ) --> NIL
//
// Updates the active Alias with the data contained into
// <cAlias> using the code block <bReplacement>.
//
// <cAlias>             the other Alias containing the data
//                      to use to update the active Alias.
// <bKey>               code block Key expression.
// <lRandom>            if true (.T.), then RANDOM.
// <bReplacement>       code block to be executed for the
//                      records with a matching key.
//
// Example:
//      dbUpdate( "INVOICE", {|| LAST}, .T.,;
//              {|| FIELD->TOTAL1 := INVOICE->SUM1,;
//                  FIELD->TOTAL2 := INVOICE->SUM2 } )
//

    // change the following line if a new release uses 
    // a different internal function. 
    __dbUpdate( cAlias, bKey, lRandom, bReplacement )

    return NIL

//================================================================
function dbZap()
//
// dbZap() --> NIL
//
// Zap command substitute.
//

    zap
    
    return NIL

//================================================================
function Get( nRow, nCol, cVar,; 
    cGetPicture, cColorString, bPreExpression, bValid; 
            )
//
// like @...get
//
   memvar getlist
   
   default cColorString   to setcolor()
   default bPreExpression to { || .T. }
   default bValid         to { || .T. }

   setpos( nRow, nCol ) 
   aadd( GetList, _GET_(; 
       &cVar., cVar, cGetPicture, bValid, bPreExpression; 
                       ):display(); 
       )
   atail(GetList):colorDisp(cColorString)
   
   return NIL

//================================================================
function Keyboard( cString )
//
// Keyboard( [<cString>] ) --> NIL
//
// Keyboard command substitute.
//
    default cString to ""

    keyboard cString

    return NIL

//================================================================
function MemPublic( cVarName )
//
// MemPublic( <cVarName> ) --> NIL
//
// <cVarName>   Contains the variable name.
//
// Creates the public variable named into <cVarName>.
//
   
   if cVarName <> NIL
      public &cVarName.
   else
      // nil
   end

   return NIL

//================================================================
function MemRelease( cMemvar )
//
// Release( <cMemvar> ) --> NIL
//
// Release command substitute.
//

    release &cMemvar.

    return NIL

//================================================================
function MemRestore( cMemFileName, lAdditive )
//
// MemRestore( <cMemFileName>, [<lAdditive>] ) --> NIL
//
// Restore From command substitute.
//

    default lAdditive to .F.

    // change the following line if a new release uses 
    // a different internal function. 
    __MRestore( cMemFileName, lAdditive )

    return NIL

//================================================================
function MemSave( cMemFileName, cSkeleton, lLike )
//
// MemSave( <cMemFileName>, [<cSkeleton>], [<lLike>] ) --> NIL
//
// Save To command substitute.
//
// <cMemFileName>       (.mem) file name to create.
// <cSkeleton>          memory names skeleton.
// <lLike>              true (.T.) means LIKE SKELETON,
//                      false (.F.) means EXCEPT SKELETON.
//

    default cSkeleton to "*"
    default lLike to .T.

    if lLike
        save to &cMemFileName. all like &cSkeleton.
    else
        save to &cMemFileName. all except &cSkeleton.
    end

    return NIL

//================================================================
function Quit()
//
// Quit() --> NIL
//
// Quit command substitute.
//

    quit
    
    return NIL

//================================================================
function Run( cCommand )
//
// Run( <cCommand> ) --> NIL
//
// <cCommand>   String to execute by OS.
//
// Run command substitute.
//

   if cCommand <> NIL
      run (cCommand)
   else   
      // nil
   end

   return NIL

//================================================================
function Say( nRow, nCol, cVar, cSayPicture, cColorString )
//
// like @...say
//
   memvar getlist
   
   default cColorString   to setcolor()

   
   devpos( nRow, nCol ) 
   if cSayPicture == NIL
      devout( cVar, cColorString )
   else
      devoutpict( cVar, cSayPicture, cColorString )
   end

   return NIL

//================================================================
function SetFunction( nFunctionKey, cString )
//
// SetFunction( <nFunctionKey>, [<cString>] ) --> NIL
//
// Set Function command substitute.
//
    default cString to ""

    set function nFunctionKey to cString

    return NIL

//================================================================
// END
//================================================================
